///--------------------------------------------------------------------------------------
// Scrolling Demo.c
//
// By Vern Jensen. Created in August of 1995
//
// Modified by Patrick Edson for BeOS May 1998.
///--------------------------------------------------------------------------------------

#include "Scrolling Demo.h"

#define	kFullScreenWindow			false		// Try me full screen!
#define kWorldRectInset				0			// Make the SpriteWorld smaller?
#define	kInterlacedMode				false		// Turns Interlaced mode on/off
#define kSyncToVBL					false		// Sync SpriteWorld to VBL?
#define kMaxFPS						30			// Set to 0 for unrestricted speed

#define kBallHeight					44			// Diamond sprite height
#define kBallWidth					44			// Diamond sprite width
#define kBallHorizOffset			0
#define kBallVertOffset				0
#define kBallFrames					1

#define	kSpriteMoveDelta			20			// Try 5, 10, 20, or 40
#define kDiamondSpace				8			// How far apart the diamonds are spaced
												// (try 1!)

#define kSpriteMoveDistance			80			// How far the sprite can move from
												// the center of the screen, in pixels.
												// Try making this value higher!

// Number of ticks to wait before changing tile image; 0 = change every frame
#define kDiamondFrameRate			132780		// How often diamond tiles change frames
#define kWallFrameRate				1991701		// How often wall changes frames

#define kBWPictResIDOffset			100
#define kTileWidth					40
#define kTileHeight					40
#define kTileMapRows				100
#define kTileMapCols				150

#define kStartRow					5			// Starting position of sprite
#define kStartCol					5			// in tile col and row


#define	kLeftArrowKey				0x7B
#define	kRightArrowKey				0x7C
#define	kDownArrowKey				0x7D
#define	kUpArrowKey					0x7E

#define	kLeftKeyPad					0x56
#define	kRightKeyPad				0x58
#define	kDownKeyPad					0x54
#define	kUpKeyPad					0x5B

#define kEscKey						0x35


#define kNoKey						0
#define kLeftKey					1
#define kUpKey						2
#define kRightKey					3
#define kDownKey					4


enum tileIDs
{
	kWallTile,
	kLastWallTile,
	kGrassTile,
	kBlackTile,
	kDiamondTile,
	kDiamondTile2,
	kLastDiamondTile,
	
	kTunnelTile1,
	kTunnelTile2,
	kTunnelTile3,
	kTunnelTile4,
	kTunnelTile5,
	kTunnelTile6,
	kWireTile1,
	kWireTile2,
	
	kMaxNumTiles
};

/*******************************/
/* SpriteWorld related Globals */
/*******************************/

SpriteWorldPtr		gSpriteWorldP;
SpriteLayerPtr		gSpriteLayerP;
TileMapStructPtr	gTileMapStructP;
TileMapPtr			gTileMap;
SpritePtr			gSimpleSpriteP, gDiamondMeterSpriteP;
DrawProcPtr			gSpriteDrawProc, gScreenDrawProc;
DoubleDrawProcPtr	gDoubleRectDrawProc;

Rect				gScreenMidRect;

long				gNumDiamondsInMap = 0;	// Number of diamonds in the TileMap
long				gNumDiamonds = 0;		// Number of diamonds sprite has collected

moveKeys 			gKeys;			// Keeps track of which keys are up and which are down

/************************/
/* BeOS related Globals */
/************************/

extern BPath		*gAppPath;
extern SimpleView	*gViewP;
extern SimpleWindow	*gWindowP;

///--------------------------------------------------------------------------------------
// CreateSpriteWorld
///--------------------------------------------------------------------------------------

void	CreateSpriteWorld( void )
{
	Rect		offscreenRect, worldRect, windRect;
	RgnHandle	mBarUpdateRgn;
	OSErr		err;
	
	err = SWEnterSpriteWorld();
	FatalError(err);
	
//--------------------------------------------------------------------------------------	
//	worldRect = gWindowP->portRect;
	worldRect = gViewP->Bounds();
//--------------------------------------------------------------------------------------

	InsetRect(&worldRect, kWorldRectInset, kWorldRectInset);
	
	
		// Set size of offscreen area
	offscreenRect = worldRect;
	OffsetRect(&offscreenRect, -offscreenRect.left, -offscreenRect.top);
	
//--------------------------------------------------------------------------------------
		// Make offscreen area evenly divisible by tile width & height
//	if ( (offscreenRect.right/kTileWidth)*kTileWidth != offscreenRect.right)
//		offscreenRect.right = (offscreenRect.right/kTileWidth)*kTileWidth + kTileWidth;
	if ((((int32)offscreenRect.right + 1)/kTileWidth) * kTileWidth != offscreenRect.right + 1)
	{
		offscreenRect.right = (((int32)offscreenRect.right + 1)/kTileWidth) * kTileWidth
			+ kTileWidth - 1;
	}
//	if ( (offscreenRect.bottom/kTileHeight)*kTileHeight != offscreenRect.bottom)
//		offscreenRect.bottom = (offscreenRect.bottom/kTileHeight)*kTileHeight + kTileHeight;
	if ((((int32)offscreenRect.bottom + 1)/kTileHeight) * kTileHeight != offscreenRect.bottom + 1)
	{
		offscreenRect.bottom = (((int32)offscreenRect.bottom + 1)/kTileHeight)*kTileHeight
			+ kTileHeight - 1;
	}
//--------------------------------------------------------------------------------------
	
		// Create the scrolling sprite world
	err = SWCreateSpriteWorldFromWindow(&gSpriteWorldP, gWindowP, gViewP, 
			&worldRect, &offscreenRect, 0);
	FatalError(err);
	
			// Create the sprite layer
	err = SWCreateSpriteLayer(&gSpriteLayerP);
	FatalError(err);
	
		// Add it to the world
	SWAddSpriteLayer(gSpriteWorldP, gSpriteLayerP);
	
	gSpriteDrawProc = SWStdSpriteDrawProc;
	gScreenDrawProc = SWStdWorldDrawProc;
	gDoubleRectDrawProc = NULL;
}
	
	
///--------------------------------------------------------------------------------------
// SetUpTiling
///--------------------------------------------------------------------------------------

void	SetUpTiling( void )	
{
	short		resIDOffset;
	short		row, col;
	OSErr		err;

	BPath	*tilePath = new BPath;
	BPath	*maskPath = new BPath;

	err = SWInitTiling(gSpriteWorldP, kTileHeight, kTileWidth, kMaxNumTiles);
	FatalError(err);
	
	err = SWCreateTileMap(&gTileMapStructP, kTileMapRows, kTileMapCols);
	FatalError(err);
	
	SWInstallTileMap(gSpriteWorldP, gTileMapStructP);
	gTileMap = gTileMapStructP->tileMap;
	
	tilePath->SetTo(gAppPath->Path(), "FullTiles.tiff");	
	
	err = SWLoadTilesFromFileXY(
		gSpriteWorldP, 
		kWallTile,				// startTileID 
		kLastDiamondTile,		// endTileID
		tilePath->Path(),		// tile image file
		NULL,					// mask image file
		1,						// horizBorderWidth
		1,						// vertBorderHeight
		kNoMask);				// maskType
	FatalError(err);

	tilePath->SetTo(gAppPath->Path(), "MaskTiles.tiff");	
	maskPath->SetTo(gAppPath->Path(), "Masks.tiff");
	
	err = SWLoadTilesFromFileXY(
		gSpriteWorldP, 
		kTunnelTile1,			// startTileID 
		kWireTile2,				// endTileID
		tilePath->Path(),		// tile image file
		maskPath->Path(),		// mask image file
		1,						// horizBorderWidth
		1,						// vertBorderHeight
		kPixelMask);			// maskType
	FatalError(err);

		// Set up tileMap
	for (row = 0; row < kTileMapRows; row++)
	{
		for (col = 0; col < kTileMapCols; col++)
		{
			if (row == 0 || col == 0 || row == kTileMapRows-1 || col == kTileMapCols-1)
				gTileMap[row][col] = kWallTile;
			else if (row > kTileMapRows / 2)
				gTileMap[row][col] = kWireTile1;
			else if ((row/kDiamondSpace)*kDiamondSpace == row && 
					 (col/kDiamondSpace)*kDiamondSpace == col)
			{
				gTileMap[row][col] = kDiamondTile;
				gNumDiamondsInMap++;
			}
			else
				gTileMap[row][col] = kGrassTile;
		}
	}
	

		// Add tunnel to tileMap
	row = kTileMapRows / 2;

	for (col = 1; col < kTileMapCols-1; col += 2)
	{
		gTileMap[row][col] = kTunnelTile1;
		gTileMap[row][col+1] = kTunnelTile2;
		gTileMap[row+1][col] = kTunnelTile4;
		gTileMap[row+1][col+1] = kTunnelTile5;
	}
}


///--------------------------------------------------------------------------------------
// CreateBallSprite
///--------------------------------------------------------------------------------------

void	CreateBallSprite( void )
{
	OSErr	err;
	
	BPath	*spritePath = new BPath;
	spritePath->SetTo(gAppPath->Path(), "Ball.tiff");	
	
	err = SWCreateSpriteFromSingleFileXY(gSpriteWorldP, &gSimpleSpriteP, NULL,
		spritePath->Path(), spritePath->Path(), kBallWidth, kBallHeight,
		kBallHorizOffset, kBallVertOffset, kBallFrames, kNoMask);
	FatalError(err);

		// Set up the ball sprite
	SWAddSprite(gSpriteLayerP, gSimpleSpriteP);
	SWSetSpriteUnderTiles(gSimpleSpriteP, true);
	SWSetSpriteMoveProc(gSimpleSpriteP, KeySpriteMoveProc);
	SWSetSpriteLocation(gSimpleSpriteP, kStartCol * kTileWidth, kStartRow * kTileHeight);
	SWSetSpriteMoveDelta(gSimpleSpriteP, 0, 0);
	SWSetSpriteDrawProc(gSimpleSpriteP, gSpriteDrawProc);
}


///--------------------------------------------------------------------------------------
// CreateDiamondMeterSprite
///--------------------------------------------------------------------------------------

void	CreateDiamondMeterSprite( void )
{
	short		spriteResID;
	OSErr		err;
	
	BPath	*spritePath = new BPath;
	spritePath->SetTo(gAppPath->Path(), "DiamondMeter.tiff");	

	err = SWCreateSpriteFromSingleFileXY(gSpriteWorldP, &gDiamondMeterSpriteP, NULL,
		spritePath->Path(), NULL, 200, 20, 0, 0, 1, kNoMask);
	FatalError(err);

	gDiamondMeterSpriteP->curFrameP->framePort
		= new BView(gDiamondMeterSpriteP->curFrameP->frameRect,
		"", B_FOLLOW_NONE, B_WILL_DRAW);
	gDiamondMeterSpriteP->curFrameP->framePix->AddChild(gDiamondMeterSpriteP->curFrameP->framePort);
	
	SWAddSprite(gSpriteLayerP, gDiamondMeterSpriteP);
	SWSetSpriteDrawProc(gDiamondMeterSpriteP, gSpriteDrawProc);
}


///--------------------------------------------------------------------------------------
// SetUpAnimation
///--------------------------------------------------------------------------------------

void	SetUpAnimation( void )
{
	Rect		moveBoundsRect;
	
	SWLockSpriteWorld(gSpriteWorldP);
	
		// Set up data used by the SmoothScrollingWorldMoveProc
	gScreenMidRect = gSimpleSpriteP->curFrameP->frameRect;
	CenterRect( &gScreenMidRect, &gSpriteWorldP->backRect );
	
	SWSetSpriteWorldMaxFPS(gSpriteWorldP, kMaxFPS);
//	SWSyncSpriteWorldToVBL(gSpriteWorldP, kSyncToVBL);
	
		// movement boundary = size of tileMap
	SetRect(&moveBoundsRect,0,0, kTileMapCols * kTileWidth - 1, kTileMapRows * kTileHeight - 1);
	
	SWSetScrollingWorldMoveBounds(gSpriteWorldP, &moveBoundsRect);
	SWSetScrollingWorldMoveProc(gSpriteWorldP, SmoothScrollingWorldMoveProc, gSimpleSpriteP);
	
	SWSetTileChangeProc(gSpriteWorldP, TileChangeProc);
	
		// Move visScrollRect to starting sprite position
	SWMoveVisScrollRect(gSpriteWorldP, 
		gSpriteWorldP->followSpriteP->destFrameRect.left - gSpriteWorldP->backRect.right/2,
		gSpriteWorldP->followSpriteP->destFrameRect.top - gSpriteWorldP->backRect.bottom/2);
	
		// Set starting position of diamond meter sprite
	DiamondMeterSpriteMoveProc(gDiamondMeterSpriteP);

	SWSetSpriteWorldScreenDrawProc(gSpriteWorldP, gScreenDrawProc);
	SWSetSpriteWorldOffscreenDrawProc(gSpriteWorldP, gScreenDrawProc);
	SWSetDoubleRectDrawProc(gSpriteWorldP, gDoubleRectDrawProc);	
	
		// Make sure CopyBits, if used, doesn't try to colorize things
//	SWSetPortToWindow(gSpriteWorldP);
//	ForeColor(blackColor);
//	BackColor(whiteColor);

	SWDrawTilesInBackground(gSpriteWorldP);
	SWUpdateScrollingSpriteWorld(gSpriteWorldP, true);
}


///--------------------------------------------------------------------------------------
//  RunAnimation
///--------------------------------------------------------------------------------------

int32	RunAnimation( void *arg )
{
	unsigned long		frames;
	
	frames = 0;
//	StartTimer();


//	while (!Button())
	while (true)
	{
		SWProcessScrollingSpriteWorld(gSpriteWorldP);
		
			// We call this in case the Control Strip or something changes our window's visRgn
//		KeepMenuBarHidden(gWindowP);

			// Move diamond meter sprite to new visScrollRect location
		DiamondMeterSpriteMoveProc(gDiamondMeterSpriteP);
	
		SWAnimateScrollingSpriteWorld(gSpriteWorldP);
		
		if (gSpriteWorldP->frameHasOccurred)
			frames++;
	}
	
	
//	ShowMenuBar(gWindowP);
//	ShowResults(frames);
}


///--------------------------------------------------------------------------------------
//  ShutDown (clean up and dispose of the SpriteWorld)
///--------------------------------------------------------------------------------------

void	ShutDown( void )
{
	SWDisposeSpriteWorld(gSpriteWorldP);
	SWExitSpriteWorld();
	
//	FlushEvents(everyEvent, 0);
//---------------------------------------------------------------------------------------
//	ShowCursor();
	be_app->ShowCursor();
//---------------------------------------------------------------------------------------
}


///--------------------------------------------------------------------------------------
//  TileChangeProc
///--------------------------------------------------------------------------------------

SW_FUNC void TileChangeProc(
	SpriteWorldPtr spriteWorldP)
{
	short			curImage;
	static uint64	wallDelay = 0, diamondDelay = 0;
	static uint64	oldTicks = 0;
	uint64			ticksPassed, ticks;
	
		// Initialize oldTicks the first time this function is called
	if (oldTicks == 0)
		oldTicks = TickCount();
	
	ticks = TickCount();
	ticksPassed = ticks - oldTicks;		// Number of ticks passed since last call
	oldTicks = ticks;
	
		// kWallTile
	wallDelay += ticksPassed;
	if (wallDelay >= kWallFrameRate)
	{
		curImage = spriteWorldP->curTileImage[kWallTile];
		if (curImage < kLastWallTile)
			curImage++;
		else
			curImage = kWallTile;
		
		SWChangeTileImage(spriteWorldP, kWallTile, curImage);
		wallDelay = 0;
	}
	
	
		// kDiamondTile
	diamondDelay += ticksPassed;
	if (diamondDelay >= kDiamondFrameRate)
	{
		curImage = spriteWorldP->curTileImage[kDiamondTile];
		if (curImage < kLastDiamondTile)
			curImage++;
		else
			curImage = kDiamondTile;
		
		SWChangeTileImage(spriteWorldP, kDiamondTile, curImage);
		diamondDelay = 0;
	}
}


///--------------------------------------------------------------------------------------
//  KeySpriteMoveProc
///--------------------------------------------------------------------------------------

SW_FUNC void KeySpriteMoveProc(SpritePtr srcSpriteP)
{
	short	row, col;
	short	rowDelta, colDelta;
	short	tile;
	
//	UpdateKeys();	// Put the latest key values in the keys structure
	
	row = ((int32)srcSpriteP->destFrameRect.top) / kTileHeight;
	col = ((int32)srcSpriteP->destFrameRect.left) / kTileWidth;

	if (row * kTileHeight == srcSpriteP->destFrameRect.top &&
		col * kTileWidth == srcSpriteP->destFrameRect.left)
	{
		tile = gTileMap[row][col];
		
			// Leave black trail behind sprite
		if (tile == kGrassTile || tile == kDiamondTile)
			SWDrawTile(gSpriteWorldP, row, col, kBlackTile);
		else if (tile == kTunnelTile2 || tile == kTunnelTile5 || tile == kWireTile1)
			SWDrawTile(gSpriteWorldP, row, col, tile+1);
			
		if (tile == kDiamondTile)
		{
			gNumDiamonds++;			// Increase number of diamonds sprite has collected
			UpdateDiamondMeter();	// Update meter
		}

		rowDelta = 0;
		colDelta = 0;
		
		if (gKeys.left)
		{
			colDelta = -1;
			gKeys.left = false;
		}
		else if (gKeys.right)
		{
			colDelta = 1;
			gKeys.right = false;
		}
		else if (gKeys.up)
		{
			rowDelta = -1;
			gKeys.up = false;
		}
		else if (gKeys.down)
		{
			rowDelta = 1;
			gKeys.down = false;
		}
				
		tile = gTileMap[row + rowDelta][col + colDelta];
		
		if (tile != kWallTile && tile != kTunnelTile1 && tile != kTunnelTile4)
		{
			srcSpriteP->vertMoveDelta = rowDelta * kSpriteMoveDelta;
			srcSpriteP->horizMoveDelta = colDelta * kSpriteMoveDelta;
		}
		else
		{
			srcSpriteP->vertMoveDelta = 0;
			srcSpriteP->horizMoveDelta = 0;
		}
	}
	
	SWOffsetSprite(srcSpriteP, srcSpriteP->horizMoveDelta, srcSpriteP->vertMoveDelta);
}


///--------------------------------------------------------------------------------------
//  DiamondMeterSpriteMoveProc - not installed as a MoveProc, but called directly
///--------------------------------------------------------------------------------------

SW_FUNC void DiamondMeterSpriteMoveProc(SpritePtr srcSpriteP)
{
		// Move sprite to top-left corner of current visScrollRect location
	SWMoveSprite(srcSpriteP, 
		gSpriteWorldP->visScrollRect.left + 10, 
		gSpriteWorldP->visScrollRect.top + 10);
}


///--------------------------------------------------------------------------------------
//  UpdateDiamondMeter - change the meter Sprite's image and mask
///--------------------------------------------------------------------------------------

SW_FUNC void UpdateDiamondMeter( void )
{
	double	percent;
	Rect	meterRect;
	
	percent = (double)gNumDiamonds / gNumDiamondsInMap;
	
		// 108 = length of meter; 17 = offset from start of meter
	meterRect.Set(17, 2, (108 * percent) + 17, 13);

	gDiamondMeterSpriteP->curFrameP->framePix->Lock();
	gDiamondMeterSpriteP->curFrameP->framePort->SetHighColor(255, 0, 255);
	gDiamondMeterSpriteP->curFrameP->framePort->FillRect(meterRect);
	gDiamondMeterSpriteP->curFrameP->framePort->Sync();
	gDiamondMeterSpriteP->curFrameP->framePix->Unlock();

			
/*		// Set port to our sprite's framePort GWorld
	SetGWorld(gDiamondMeterSpriteP->curFrameP->framePort, nil);
	
	PaintRect(&meterRect);
			
		// Set port to our sprite's pixel mask GWorld
	SetGWorld(gDiamondMeterSpriteP->curFrameP->maskPort, nil);
	
		// Mask image is inverted when in 8-bit or less
	if (gSpriteWorldP->pixelDepth <= 8)
		ForeColor(whiteColor);
	else
		ForeColor(blackColor);
	
	PaintRect(&meterRect);
	ForeColor(blackColor);
*/	
		// Set sprite to be redrawn, since we've changed its image
	gDiamondMeterSpriteP->needsToBeDrawn = true;
}


///--------------------------------------------------------------------------------------
//  SmoothScrollingWorldMoveProc - our scrolling WorldMoveProc
///--------------------------------------------------------------------------------------

SW_FUNC void SmoothScrollingWorldMoveProc(
	SpriteWorldPtr spriteWorldP,
	SpritePtr followSpriteP)
{	
	short	screenMidRectTop, screenMidRectLeft;
	
	screenMidRectTop = gScreenMidRect.top + spriteWorldP->visScrollRect.top;
	screenMidRectLeft = gScreenMidRect.left + spriteWorldP->visScrollRect.left;
	
	
	spriteWorldP->horizScrollDelta = (kSpriteMoveDelta * 
		(int32)(followSpriteP->destFrameRect.left - screenMidRectLeft) ) / kSpriteMoveDistance;
	
	spriteWorldP->vertScrollDelta = (kSpriteMoveDelta * 
		(int32)(followSpriteP->destFrameRect.top - screenMidRectTop) ) / kSpriteMoveDistance;
	
//	if (kInterlacedMode)
//		spriteWorldP->vertScrollDelta = spriteWorldP->vertScrollDelta>>1<<1;
}


///--------------------------------------------------------------------------------------
//  UpdateKeys (Put the latest key values in the keys structure)
///--------------------------------------------------------------------------------------

void	UpdateKeys( void )
{
//	EventRecord		event;
	short			theKey;
	Boolean			isDown;
	
//	while ( GetOSEvent( (keyUpMask | keyDownMask), &event ) )
	{
//		theKey = (event.message & keyCodeMask) >> 8;
//		isDown = (event.what != keyUp);
		
		if ( (theKey == kLeftArrowKey) || (theKey == kLeftKeyPad) )
			gKeys.left = isDown;
		else if ( (theKey == kRightArrowKey) || (theKey == kRightKeyPad) )
			gKeys.right = isDown;
		else if ( (theKey == kDownArrowKey) || (theKey == kDownKeyPad) )
			gKeys.down = isDown;
		else if ( (theKey == kUpArrowKey) || (theKey == kUpKeyPad) )
			gKeys.up = isDown;
	}
}

///--------------------------------------------------------------------------------------
//	FatalError
///--------------------------------------------------------------------------------------
void FatalError(OSErr err)
{
}

///--------------------------------------------------------------------------------------
//	CenterRect - centers rectA in rectB, returning result in rectA
///--------------------------------------------------------------------------------------

void CenterRect(Rect* rectA, Rect* rectB)
{
	short width = (rectA->right - rectA->left);
	short height = (rectA->bottom - rectA->top);

	rectA->left = rectB->left + (((rectB->right - rectB->left) / 2) - (width / 2));
	rectA->top = rectB->top + (((rectB->bottom - rectB->top) / 2) - (height / 2));
	rectA->right = rectA->left + width;
	rectA->bottom = rectA->top + height;
}

