///--------------------------------------------------------------------------------------
// Tiling Demo.c
///--------------------------------------------------------------------------------------

#include <SWFPSReport.h>
#include <SWApplication.h>
#include <SWGameUtils.h>

#include "Tiling Demo.h"

#define	kFullScreenWindow			true		// Makes the window fill the screen
#define	kInterlacedMode				false		// Turns Interlaced mode on/off
#define kSyncToVBL					false		// Syncs animation to VBL
#define kWorldRectInset				0			// Makes SpriteWorld smaller than window
#define kMaxFPS						30			// Set to 0 for unrestricted speed

#define kNumDiamonds				10			// Number of diamonds
#define kDiamondHeight				40			// Diamond sprite height
#define kDiamondWidth				40			// Diamond sprite width
#define kDiamondHorizOffset			0
#define kDiamondVertOffset			0
#define kDiamondFrames				1
#define kMaxDiamondMoveDelta		5			// Max speed for the diamonds

#define kBallHeight					44			// Diamond sprite height
#define kBallWidth					44			// Diamond sprite width
#define kBallHorizOffset			0
#define kBallVertOffset				0
#define kBallFrames					1

#define kTileWidth					40			// Don't change these unless you
#define kTileHeight					40			// change the resource as well.

#define kMaxSpriteMoveDelta			8


enum tileIDs			// Meaningful values for the tileIDs
{
	kWallTile,
	kBoardTile,
	kMaxNumTiles
};

/*******************************/
/* SpriteWorld related Globals */
/*******************************/

SpriteWorldPtr		gSpriteWorldP;

SpriteLayerPtr		gTopDiamondSpriteLayerP;
SpriteLayerPtr		gBottomDiamondSpriteLayerP;
	
TileMapStructPtr	gTileMapStructP;
SpritePtr			gBallSpriteP;

DrawProcPtr			gSpriteDrawProc;
short				gNumTileMapRows, gNumTileMapCols;

uint32				gFrames;

/************************/
/* BeOS related Globals */
/************************/

extern SimpleWindow	*gWindowP;
extern SimpleView	*gViewP;
extern BPath		*gAppPath;

///--------------------------------------------------------------------------------------
// CreateSpriteWorld
///--------------------------------------------------------------------------------------

void	CreateSpriteWorld( void )
{
	Rect		offscreenRect, worldRect, windRect;
	OSErr		err;
	
	err = SWEnterSpriteWorld();
	FatalError(err);

	
	worldRect = gViewP->Bounds();
	worldRect.InsetBy(kWorldRectInset, kWorldRectInset);
	
	
		// Set size of offscreen area
	offscreenRect = worldRect;
	OffsetRect(&offscreenRect, -offscreenRect.left, -offscreenRect.top);
	

		// Make offscreen area evenly divisible by tile width & height
//	if ( (offscreenRect.right/kTileWidth)*kTileWidth != offscreenRect.right)
//		offscreenRect.right = (offscreenRect.right/kTileWidth)*kTileWidth + kTileWidth;
	
//	if ( (offscreenRect.bottom/kTileHeight)*kTileHeight != offscreenRect.bottom)
//		offscreenRect.bottom = (offscreenRect.bottom/kTileHeight)*kTileHeight + kTileHeight;
	
	
	gNumTileMapRows = (offscreenRect.bottom + 1) / kTileHeight;
	gNumTileMapCols = (offscreenRect.right + 1) / kTileWidth;
	

	err = SWCreateSpriteWorldFromWindow(&gSpriteWorldP, gWindowP, gViewP, 
			&worldRect, &offscreenRect, 8);
	FatalError(err);
	
	gSpriteDrawProc = SWStdSpriteDrawProc;
}


///--------------------------------------------------------------------------------------
// CreateSprites
///--------------------------------------------------------------------------------------

void	CreateSprites( void )
{
	SpritePtr			diamondSpriteP;
	SpritePtr			tempSpriteP;
	short				spriteNum;
	Point				moveDelta, mousePoint;
	Rect				moveBounds;
	OSErr				err;
	
	BPath	*spritePath = new BPath;
		
		// Create the sprite layers
	SWCreateSpriteLayer(&gTopDiamondSpriteLayerP);
	SWCreateSpriteLayer(&gBottomDiamondSpriteLayerP);
	FatalError(SWStickyError());
	
	
	moveBounds = gSpriteWorldP->backRect;
	moveBounds.InsetBy(40, 40);

	spritePath->SetTo(gAppPath->Path(), "GreenDiamond.tiff");	

		// create and set up the diamond sprites
	for (spriteNum = 0; spriteNum < kNumDiamonds; spriteNum++)
	{
		if (spriteNum == 0)
		{

			err = SWCreateSpriteFromSingleFileXY(gSpriteWorldP, &diamondSpriteP, NULL,
				spritePath->Path(), NULL, kDiamondWidth, kDiamondHeight,
				kDiamondHorizOffset, kDiamondVertOffset, kDiamondFrames, kNoMask);
			FatalError(err);
			tempSpriteP = diamondSpriteP;
		}
		else
		{
			err = SWCloneSprite(diamondSpriteP, &tempSpriteP, NULL);
			FatalError(err);
		}
		

		SWSetSpriteMoveBounds(tempSpriteP, &moveBounds);
		SWSetSpriteMoveProc(tempSpriteP, DiamondSpriteMoveProc);
		
		SWSetSpriteLocation(tempSpriteP, 
				GetRandom(40, gSpriteWorldP->backRect.right-80), 
				GetRandom(40, gSpriteWorldP->backRect.bottom-80));
		
		do
		{
			moveDelta.x = GetRandom(-kMaxDiamondMoveDelta, kMaxDiamondMoveDelta);
			moveDelta.y = GetRandom(-kMaxDiamondMoveDelta, kMaxDiamondMoveDelta);
		} while (!moveDelta.y || !moveDelta.x);
		
		SWSetSpriteMoveDelta(tempSpriteP, moveDelta.x, moveDelta.y);
		SWSetSpriteDrawProc(tempSpriteP, gSpriteDrawProc);
		SWAddSprite(gTopDiamondSpriteLayerP, tempSpriteP);
	}
	
	
		// Create the ball sprite
	spritePath->SetTo(gAppPath->Path(), "ball.tiff");	

	err = SWCreateSpriteFromSingleFileXY(gSpriteWorldP, &gBallSpriteP, NULL,
		spritePath->Path(), spritePath->Path(), kBallWidth, kBallHeight,
		kBallHorizOffset, kBallVertOffset, kBallFrames, kNoMask);
		FatalError(err);
	FatalError(err);
	
	
		// Set up the ball sprite
	SWAddSprite(gTopDiamondSpriteLayerP, gBallSpriteP);
	SWSetSpriteMoveProc(gBallSpriteP, BallSpriteMoveProc);
	SWSetSpriteDrawProc(gBallSpriteP, gSpriteDrawProc);
	SWSetFrameHotSpot(gBallSpriteP->curFrameP, 20, 20);

	uint32	modifiers;

	mousePoint.x = 50; mousePoint.y = 50;

		SWSetSpriteMoveDelta(gBallSpriteP,
					GetRandom(-kMaxSpriteMoveDelta, kMaxSpriteMoveDelta), 
					GetRandom(-kMaxSpriteMoveDelta, kMaxSpriteMoveDelta));
	SWSetSpriteMoveBounds(gBallSpriteP, &gSpriteWorldP->backRect);
	
	SWSetSpriteLocation(gBallSpriteP, mousePoint.x, mousePoint.y);
	
	SWAddSpriteLayer(gSpriteWorldP, gBottomDiamondSpriteLayerP);	// Add bottom first, so it is drawn first
	SWAddSpriteLayer(gSpriteWorldP, gTopDiamondSpriteLayerP);		// Add top last, so it is drawn last

	SWLockSpriteWorld(gSpriteWorldP);
}

///--------------------------------------------------------------------------------------
// SetUpTiling
///--------------------------------------------------------------------------------------

void	SetUpTiling( void )
{
	short		row, col;
	OSErr		err;
	
	BPath	*spritePath = new BPath;
	BPath	*maskPath = new BPath;
		
		// Must be done before we can load the tiles.
	err = SWInitTiling(gSpriteWorldP, kTileHeight, kTileWidth, kMaxNumTiles);
	FatalError(err);
	
	err = SWCreateTileMap(&gTileMapStructP, gNumTileMapRows, gNumTileMapCols);
	FatalError(err);
	
	SWInstallTileMap(gSpriteWorldP, gTileMapStructP);
	
	// For demonstration purposes, we've decided to load the tiles from CICN
	// resources here, and load them from PICT resources in the Scrolling Demo.

	spritePath->SetTo(gAppPath->Path(), "RedSquare.tiff");

	err = SWLoadTileFromFile(
		gSpriteWorldP, 
		kWallTile,				// tileID
		spritePath->Path(), 	// tile image file
		NULL,					// no mask or solid mask
		kSolidMask);			// maskType
	FatalError(err);
	
		// Load the board ("wire") tile
	spritePath->SetTo(gAppPath->Path(), "BlueCross.tiff");
	maskPath->SetTo(gAppPath->Path(), "Cross.tiff");
	
	err = SWLoadTileFromFile(
		gSpriteWorldP, 
		kBoardTile,				// tileID
		spritePath->Path(),		// tile image file
		maskPath->Path(),		// mask image file
		kPixelMask);			// maskType
	FatalError(err);

	
		// Initialize the values in the tileMap
	for (row = 0; row < gNumTileMapRows; row++)
	{
		for (col = 0; col < gNumTileMapCols; col++)
		{
			if (row == 0 || col == 0 || row == gNumTileMapRows-1 || col == gNumTileMapCols-1)
				gTileMapStructP->tileMap[row][col] = kWallTile;
			else
				gTileMapStructP->tileMap[row][col] = kBoardTile;
		}
	}
}

///--------------------------------------------------------------------------------------
// SetUpAnimation
///--------------------------------------------------------------------------------------

void	SetUpAnimation( void )
{
	SWSetSpriteWorldMaxFPS(gSpriteWorldP, kMaxFPS);
//	SWSyncSpriteWorldToVBL(gSpriteWorldP, kSyncToVBL);	

	SWDrawTilesInBackground(gSpriteWorldP);
	SWUpdateSpriteWorld(gSpriteWorldP, true);
}


///--------------------------------------------------------------------------------------
//  RunAnimation
///--------------------------------------------------------------------------------------

	
int32	RunAnimation( void *arg )
{
	SpriteWorldPtr		theWorld = (SpriteWorldPtr)arg;
	
	gFrames = 0;
	StartTimer();
	
//	while (!Button())
	while(true)
	{
		SWProcessSpriteWorld(theWorld);
		SWAnimateSpriteWorld(theWorld);
		
		if (theWorld->frameHasOccurred)
			gFrames++;
	}

//	ShowResults(frames);
}


///--------------------------------------------------------------------------------------
//  ShutDown (clean up and dispose of the SpriteWorld)
///--------------------------------------------------------------------------------------


void	ShutDown( void )
{
	SWUnlockSpriteWorld(gSpriteWorldP);
	SWDisposeSpriteWorld(gSpriteWorldP);
	SWExitSpriteWorld();
}

///--------------------------------------------------------------------------------------
//  BallSpriteMoveProc - makes the ball follow the mouse
///--------------------------------------------------------------------------------------

SW_FUNC void BallSpriteMoveProc(SpritePtr srcSpriteP)
{
	Point				mousePoint;
	SpritePtr			tempSpriteP;
//	EventRecord			event;
	short				theChar;
	
	
		// Move the ball sprite
//---
	uint32	modifiers;
	if (gViewP->Window()->Lock())
	{
		gViewP->GetMouse(&mousePoint, &modifiers);
		gViewP->Window()->Unlock();
	}
//---
	
		// Subtract any offset from worldRect to window if a worldRect is used
	mousePoint.x -= gSpriteWorldP->windRect.left;
	mousePoint.y -= gSpriteWorldP->windRect.top;
	
		// This sprite uses a hotPoint of 20,20, so all we have to do is move its
		// hotPoint to the current mouse location, and it will appear centered properly
	SWMoveSprite(srcSpriteP, mousePoint.x, mousePoint.y);
		
	// Check for collision with diamonds in same layer as ball //

	if (srcSpriteP->isUnderTiles)
		tempSpriteP = gBottomDiamondSpriteLayerP->headSpriteP;
	else
		tempSpriteP = gTopDiamondSpriteLayerP->headSpriteP;
	

	while (tempSpriteP != NULL)
	{
			// are the sprite's rectangles overlapping?
		if (tempSpriteP != gBallSpriteP &&
			(tempSpriteP->destFrameRect.top < gBallSpriteP->destFrameRect.bottom) &&
			(tempSpriteP->destFrameRect.bottom > gBallSpriteP->destFrameRect.top) &&
			(tempSpriteP->destFrameRect.left < gBallSpriteP->destFrameRect.right) &&
			(tempSpriteP->destFrameRect.right > gBallSpriteP->destFrameRect.left))
		{
				// Check to see if the masks overlap
			if ( SWRadiusCollision(tempSpriteP, gBallSpriteP)) 
			{
					// Switch diamond position above/under tiles
				if (tempSpriteP->isUnderTiles)
				{
					SWSetSpriteUnderTiles(tempSpriteP, false);
					SWRemoveSprite(gBottomDiamondSpriteLayerP, tempSpriteP);
					SWAddSprite(gTopDiamondSpriteLayerP, tempSpriteP);
				}
				else
				{
					SWSetSpriteUnderTiles(tempSpriteP, true);
					SWRemoveSprite(gTopDiamondSpriteLayerP, tempSpriteP);
					SWAddSprite(gBottomDiamondSpriteLayerP, tempSpriteP);
				}
			}
		}
		
		tempSpriteP = tempSpriteP->nextSpriteP;
	}
}

///--------------------------------------------------------------------------------------
//  DiamondSpriteMoveProc
///--------------------------------------------------------------------------------------
SW_FUNC void DiamondSpriteMoveProc(SpritePtr ballSpriteP)
{	
	SWOffsetSprite(ballSpriteP, ballSpriteP->horizMoveDelta, ballSpriteP->vertMoveDelta);
	(void)SWBounceSprite(ballSpriteP);
}
