///--------------------------------------------------------------------------------------
//	MacOSGlue.h
///--------------------------------------------------------------------------------------

#pragma once

typedef BWindow		*CWindowPtr;
typedef BWindow		*WindowPtr;
typedef BWindow		*GrafPtr;
typedef BView		*GWorldPtr;
typedef uint32		*PixMapPtr;
typedef uint32		**PixMapHandle;
typedef uint32		**CIconHandle;
typedef rgb_color	RGBColor;
typedef BRegion		*RgnHandle;
typedef BRect		Rect;
typedef BPoint		Point;

typedef uint64	UnsignedWide;
typedef size_t	Size;
typedef uint32	*Handle;
typedef uint32	VBLTask;
typedef bool	Boolean;
typedef uint32	*Ptr;
typedef status_t	OSErr;
//typedef uint32	*GDHandle;

#define HLock		HLockHi

enum {
srcCopy,
notSrcCopy,
srcOr,
notSrcOr,
SrcXor,
notSrcXor,
srcBic,
notSrcBic};

const status_t	noErr					= B_NO_ERROR;
const int32		nil						= NULL;
const status_t	resNotFound				= -1;
const int32		gestaltSystemVersion	= 1;
const status_t	kMemoryError			= B_NO_MEMORY;
const status_t	kLockError				= 1000;
const status_t	kColorSpaceNotSupported	= 1001;

// Machine independent 32 bit color union
typedef union Color32 Color32;

union Color32
{
	uint8	bytes[4]; 
	uint32	word; 
};

// Pic stucture
typedef struct Pic Pic;
typedef Pic *PicPtr, **PicHandle;

struct Pic
{
	BRect	picFrame;
};

// GDevice stucture
typedef struct GDevice GDevice;
typedef GDevice *GDPtr, **GDHandle;

struct GDevice
{
	int32			gdRefNum;
	int32			gdID;
	int32			gdType;
	int32   		gdFlags;
	PixMapHandle	gdPMap;
};

color_space	ColorSpaceFromDepth(int16 inDepth);
int16		DepthFromColorSpace(color_space inSpace);
short		DepthFromWindow( WindowPtr theWindow );

// Miscellaneous functions
char		GetMMUMode();
OSErr		Gestalt(int32 inParameter, int32 *ioResult);
status_t	MemError();
Handle		Get1Resource(int32 type, int32 id);
PicHandle	GetPicture(int32 id);
void		ReleaseResource(Handle inHandle);
uint64		TickCount();

// Memory Functions
void		*NewPtrClear(size_t inSize);
void		*NewPtr(size_t inSize);
void		DisposePtr(uint32 *inPtr);
void		HLock(Handle inHandle);
void		HUnlock(Handle inHandle);
uint32		*NewHandle(size_t inSize);
uint32		*NewHandleClear(size_t inSize);
void		DisposeHandle(Handle inHandle);

///--------------------------------------------------------------------------------------
//	Quickdraw related
///--------------------------------------------------------------------------------------
void		SetPort(WindowPtr window);
void		SetGWorld(BView *inView, GDHandle inHandle);
void		GetGWorld(BView **ioView, GDHandle *ioHandle);
void		DisposeGWorld(GWorldPtr inWorld);
void		CopyBits( BBitmap *srcBitmap,
				BView	*dstView,
				BRect	*srcRect,
				BRect	*dstRect,
				short	mode,
				BRegion	*copyMask);
void		OffsetRect(BRect *inRect, int32 inHoriz, int32 inVert);
void		InsetRect(BRect *inRect, int32 inHoriz, int32 inVert);
void		SetRect(BRect *inRect, int32 left, int32 top, int32 right, int32 bottom);
BRegion	*	NewRgn();
void		DisposeRgn(BRegion *inRegion);
void		OffsetRgn( BRegion *inRegion,
				long offsetX,
				long offsetY);
bool		EmptyRgn(BRegion *inRegion);
OSErr		BitmapToRegion(BRegion *ioRegion,
				BBitmap	*inBitmap);
#define		DetachResource	DisposeHandle
#define		BlockMoveData	memcpy

enum {
	B_SMALL_ICON		= 1296646990,
	B_BIG_ICON			= 1229147982,
};
