///--------------------------------------------------------------------------------------
//	Tiling.h
//
//	By: Vern Jensen 10/27/95
///--------------------------------------------------------------------------------------


#ifndef __TILING__
#define __TILING__

#ifndef __SWCOMMON__
#include "SWCommonHeaders.h"
#endif

#ifndef __SPRITE__
#include "Sprite.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

#if PRAGMA_ALIGN_SUPPORTED
#pragma options align=mac68k
#endif


///--------------------------------------------------------------------------------------
//	Tiling definitions
///--------------------------------------------------------------------------------------

typedef enum	// Used to SWCheckSpriteWithTiles
{
	kSWEntireSprite = 0,
	kSWTopSide,
	kSWRightSide,
	kSWBottomSide,
	kSWLeftSide
} SWTileSearchType;


typedef short** TileMapPtr;

typedef struct TileMapStruct
{
	short		numRows;
	short		numCols;
	TileMapPtr	tileMap;
//	Handle		tileMapDataH;
//	Handle		arrayOfPointersH;
	short		*tileMapDataH;
	short		**arrayOfPointersH;
	Boolean		isLocked;
} TileMapStruct, *TileMapStructPtr;

typedef SW_FUNC void (*TileChangeProcPtr)(
	SpriteWorldPtr spriteWorldP);


///--------------------------------------------------------------------------------------
//	tile function prototypes
///--------------------------------------------------------------------------------------

SW_FUNC OSErr SWInitTiling(
	SpriteWorldPtr	spriteWorldP,
	short 			tileHeight,
	short			tileWidth,
	short			maxNumTiles);

SW_FUNC void SWExitTiling(
	SpriteWorldPtr	spriteWorldP);

SW_FUNC OSErr SWCreateTileMap(
	TileMapStructPtr	*tileMapStructPP,
	short 				numTileMapRows,
	short				numTileMapCols);

SW_FUNC void SWDisposeTileMap(
	TileMapStructPtr	*tileMapStructPP);

SW_FUNC void SWLockTileMap(
	TileMapStructPtr	tileMapStructP);

SW_FUNC void SWUnlockTileMap(
	TileMapStructPtr	tileMapStructP);

SW_FUNC void SWInstallTileMap(
	SpriteWorldPtr		spriteWorldP,
	TileMapStructPtr	tileMapStructP);

SW_FUNC OSErr SWLoadTileMap(
	TileMapStructPtr	*tileMapStructPP,
	short 				resourceID);

SW_FUNC OSErr SWSaveTileMap(
	TileMapStructPtr	tileMapStructP,
	short 				destResID);

SW_FUNC OSErr SWResizeTileMap(
	TileMapStructPtr	tileMapStructP,
	short 				numNewTileMapRows,
	short				numNewTileMapCols);

SW_FUNC void SWLockTiles(
	SpriteWorldPtr	spriteWorldP);

SW_FUNC void SWUnlockTiles(
	SpriteWorldPtr	spriteWorldP);

SW_FUNC void SWSetTilingOn(
	SpriteWorldPtr	spriteWorldP,
	Boolean			tilingIsOn);

SW_FUNC void SWSetSpriteUnderTiles(
	SpritePtr	srcSpriteP,
	Boolean		isUnder);

SW_FUNC OSErr SWSetTileMaskDrawProc(
	SpriteWorldPtr	spriteWorldP,
	DrawProcPtr		drawProc);

SW_FUNC OSErr SWLoadTileFromCicnResource(
	SpriteWorldPtr	spriteWorldP, 
	short			tileID,
	short			cicnID, 
	MaskType		maskType);

SW_FUNC OSErr SWLoadTilesFromPictResource(
	SpriteWorldPtr	spriteWorldP, 
	short			startTileID,
	short			endTileID, 
	short			pictResID, 
	short			maskResID,
	MaskType		maskType,
	short			horizBorderWidth,
	short			vertBorderHeight);

SW_FUNC void SWDisposeTile(
	SpriteWorldPtr	spriteWorldP,
	short			tileID);

SW_FUNC OSErr SWDrawTilesInBackground(
	SpriteWorldPtr	spriteWorldP);
	
SW_FUNC void SWResetTilingCache(
	SpriteWorldPtr	spriteWorldP);

SW_FUNC void SWDrawTile(
	SpriteWorldPtr	spriteWorldP,
	short			tileRow,
	short			tileCol,
	short			tileID);

SW_FUNC void SWSetTileChangeProc(
	SpriteWorldPtr		spriteWorldP,
	TileChangeProcPtr	tileChangeProc);

SW_FUNC void SWChangeTileImage(
	SpriteWorldPtr	spriteWorldP,
	short			tileID,
	short			currentFrame);

SW_FUNC void SWUpdateTileOnScreen(
	SpriteWorldPtr	spriteWorldP,
	short			tileID);

SW_FUNC void SWResetCurrentTileImages(
	SpriteWorldPtr	spriteWorldP);

SW_FUNC void SWDrawTilesInRect(
	SpriteWorldPtr	spriteWorldP,
	Rect*			updateRectP,
	Boolean			faster);

SW_FUNC static void SWAddChangedRect(
	SpriteWorldPtr	spriteWorldP, 
	Rect			*changedRectP);

SW_FUNC void SWDrawTilesAboveSprite(
	SpriteWorldPtr	spriteWorldP,
	Rect*			updateRectP);

SW_FUNC short SWReturnTileUnderPixel(
	SpriteWorldPtr	spriteWorldP,
	short pixelCol,
	short pixelRow);

SW_FUNC Boolean SWCheckSpriteWithTiles(
	SpriteWorldPtr	spriteWorldP,
	SpritePtr		srcSpriteP, 
	SWTileSearchType searchType,
	Rect			*insetRectP,
	short			firstTileID,
	short			lastTileID,
	Boolean			fixPosition);
	
SW_FUNC void SWWrapRectToWorkArea(
	SpriteWorldPtr	spriteWorldP,
	Rect*			dstRectP);

SW_FUNC OSErr SWLoadTileFromBitmap(
	SpriteWorldPtr	spriteWorldP, 
	short			tileID,
	BBitmap			*bitmap, 
	BBitmap			*overBitmap,
	MaskType		maskType);
	
SW_FUNC void SWStdTileMaskDrawProc(
	FramePtr srcFrameP,
	FramePtr dstFrameP,
	Rect* srcRect,
	Rect* dstRect);

SW_FUNC OSErr SWLoadTilesFromBitmapTwo(
	SpriteWorldPtr	spriteWorldP, 
	short			startTileID,
	short			endTileID, 
	BBitmap			*tileBitmap, 
	BBitmap			*maskBitmap,
	MaskType		maskType,
	short			horizBorderWidth,
	short			vertBorderHeight);

SW_FUNC void SWTileMaskDrawProc(
	FramePtr srcFrameP,
	FramePtr dstFrameP,
	Rect* srcRect,
	Rect* dstRect);

SW_FUNC OSErr SWLoadTileFromFile(
	SpriteWorldPtr	spriteWorldP, 
	short			tileID,
	const char 		*tilePath,
	const char 		*maskPath, 
	MaskType		maskType);

SW_FUNC OSErr SWLoadTilesFromFileXY(
	SpriteWorldPtr	spriteWorldP, 
	short			startTileID,
	short			endTileID, 
	const char 		*tileBitmap, 
	const char 		*maskBitmap,
	short			borderWidth,
	short			borderHeight,
	MaskType		maskType);

#if PRAGMA_ALIGN_SUPPORTED
#pragma options align=reset
#endif

#ifdef __cplusplus
}
#endif

#endif /* __TILING__ */