///--------------------------------------------------------------------------------------
//	MacOSGlue.c
///--------------------------------------------------------------------------------------

#include "MacOSGlue.h"

///--------------------------------------------------------------------------------------
//	ColorSpaceFromDepth
///--------------------------------------------------------------------------------------
color_space
ColorSpaceFromDepth(int16 inDepth)
{
	color_space	cs;
		
	switch(inDepth)
	{
		case 8:
			cs = B_CMAP8;
		break;
		
		case 16:
			cs = B_RGB16;
		break;
		
		case 32:
			cs = B_RGB32;
		break;
		
		default:
			cs = B_CMAP8;
	}
	
	return cs;
}

///--------------------------------------------------------------------------------------
//	DepthFromColorSpace
///--------------------------------------------------------------------------------------
int16
DepthFromColorSpace(color_space inSpace)
{
	int16	depth;
		
	switch(inSpace)
	{
		case B_CMAP8:
			depth = 8;
		break;
		
		case B_RGB16:
			depth = 16;
		break;
		
		case B_RGB32:
			depth = 32;
		break;
		default:
			depth = 8;
	}
	
	return depth;
}

///--------------------------------------------------------------------------------------
//	DepthFromWindow
///--------------------------------------------------------------------------------------

short DepthFromWindow( WindowPtr theWindow )
{
	short			depth;

	BScreen		*theScreen = new BScreen(theWindow);
	
	if (theScreen->IsValid())
	{
		depth = DepthFromColorSpace(theScreen->ColorSpace());
	}
	
	return depth;
}

///--------------------------------------------------------------------------------------
//	GetMMUMode
///--------------------------------------------------------------------------------------
char
GetMMUMode()
{
	return 'a';
}

///--------------------------------------------------------------------------------------
//	Gestalt
///--------------------------------------------------------------------------------------
OSErr
Gestalt(int32 inParamter, int32 *ioResult)
{
	*ioResult = 0x0700;
	
	return noErr;
}

///--------------------------------------------------------------------------------------
//	MemError
///--------------------------------------------------------------------------------------
status_t
MemError()
{
	return kMemoryError;
}

///--------------------------------------------------------------------------------------
//	Get1Resource
///--------------------------------------------------------------------------------------
Handle
Get1Resource(int32 type, int32 id)
{
	return NULL;
}

///--------------------------------------------------------------------------------------
//	ReleaseResource
///--------------------------------------------------------------------------------------
void
ReleaseResource(Handle inHandle)
{
}

///--------------------------------------------------------------------------------------
//	GetPicture
///--------------------------------------------------------------------------------------
PicHandle
GetPicture(int32 id)
{
	return NULL;
}

///--------------------------------------------------------------------------------------
//	TickCount
///--------------------------------------------------------------------------------------
uint64 TickCount()
{
	return system_time();
}

#pragma mark -
#pragma mark *** Memory functions ***
#pragma mark -

///--------------------------------------------------------------------------------------
//	NewPtr
///--------------------------------------------------------------------------------------
void *
NewPtr(size_t inSize)
{
	void	*mem;
	
	mem =  malloc(inSize);
	
	return mem;
}

///--------------------------------------------------------------------------------------
//	NewPtrClear
///--------------------------------------------------------------------------------------
void *
NewPtrClear(size_t inSize)
{
	void	*mem;
	
	mem =  malloc(inSize);
	
	memset(mem, 0, inSize);
	
	return mem;
}

///--------------------------------------------------------------------------------------
//	DisposePtr
///--------------------------------------------------------------------------------------
void
DisposePtr(uint32 *inPtr)
{
	free(inPtr);
}

///--------------------------------------------------------------------------------------
//	NewHandle
///--------------------------------------------------------------------------------------
uint32 *
NewHandle(size_t inSize)
{
	uint32	*mem;
	
	mem =  (uint32 *)malloc(inSize);

	return mem;
}

///--------------------------------------------------------------------------------------
//	NewHandleClear
///--------------------------------------------------------------------------------------
uint32 *
NewHandleClear(size_t inSize)
{
	uint32	*mem;
	
	mem =  (uint32 *)malloc(inSize);
	
	memset(mem, 0, inSize);

	return mem;
}

///--------------------------------------------------------------------------------------
//	DisposeHandle
///--------------------------------------------------------------------------------------
void
DisposeHandle(Handle inHandle)
{	
	free(inHandle);
}

///--------------------------------------------------------------------------------------
//	HLock
///--------------------------------------------------------------------------------------
void
HLock(Handle inHandle)
{
}

///--------------------------------------------------------------------------------------
//	HUnlock
///--------------------------------------------------------------------------------------
void
HUnlock(Handle inHandle)
{
}

#pragma mark -
#pragma mark *** Quickdraw related ***
#pragma mark -

///--------------------------------------------------------------------------------------
//	SetPort
///--------------------------------------------------------------------------------------
void
SetPort(WindowPtr window)
{
	
}

///--------------------------------------------------------------------------------------
//	SetGWorld
///--------------------------------------------------------------------------------------
void
SetGWorld(BView *inView, GDHandle inHandle)
{
	
}

///--------------------------------------------------------------------------------------
//	GetGWorld
///--------------------------------------------------------------------------------------
void
GetGWorld(BView **ioView, GDHandle *ioHandle)
{
	
}

///--------------------------------------------------------------------------------------
//	DisposeGWorld
///--------------------------------------------------------------------------------------
void
DisposeGWorld(GWorldPtr inWorld)
{
	
}

///--------------------------------------------------------------------------------------
//	CopyBits
///--------------------------------------------------------------------------------------
void
CopyBits(BBitmap *srcBitmap,
	BView	*dstView,
	BRect	*srcRect,
	BRect	*dstRect,
	short	mode,
	BRegion	*copyMask)
{
	if (dstView->Window()->Lock())
	{
		dstView->DrawBitmapAsync(srcBitmap, *srcRect, *dstRect);	
		dstView->Sync();
		
		dstView->Window()->Unlock();	
	}	
}

///--------------------------------------------------------------------------------------
//	OffsetRect
///--------------------------------------------------------------------------------------
void
OffsetRect(BRect *inRect, int32 inHoriz, int32 inVert)
{
	inRect->OffsetBy(inHoriz, inVert);
}

///--------------------------------------------------------------------------------------
//	InsetRect
///--------------------------------------------------------------------------------------
void
InsetRect(BRect *inRect, int32 inHoriz, int32 inVert)
{
	inRect->InsetBy(inHoriz, inVert);
}

///--------------------------------------------------------------------------------------
//	SetRect
///--------------------------------------------------------------------------------------
void
SetRect(BRect *inRect, int32 left, int32 top, int32 right, int32 bottom)
{
	inRect->Set(left, top, right, bottom);
}

///--------------------------------------------------------------------------------------
//	NewRgn
///--------------------------------------------------------------------------------------
BRegion *
NewRgn()
{
	return (new BRegion());
}

///--------------------------------------------------------------------------------------
//	DisposeRgn
///--------------------------------------------------------------------------------------
void
DisposeRgn(BRegion *inRegion)
{
	delete inRegion;
}

///--------------------------------------------------------------------------------------
//	OffsetRgn
///--------------------------------------------------------------------------------------
void
OffsetRgn(BRegion *inRegion, long offsetX, long offsetY)
{
	inRegion->OffsetBy(offsetX, offsetY);
}

///--------------------------------------------------------------------------------------
//	EmptyRegion
///--------------------------------------------------------------------------------------
bool
EmptyRgn(BRegion *inRegion)
{
	return (!inRegion->Frame().IsValid());
}

///--------------------------------------------------------------------------------------
//	BitmapToRegion
///--------------------------------------------------------------------------------------
OSErr
BitmapToRegion(BRegion *ioRegion, BBitmap *inBitmap)
{
	int32	dispWidth;
	int32	dispHeight;
	uint32	*bits;
	
	if (inBitmap == NULL)
		return kMemoryError;
	
	dispWidth	= inBitmap->BytesPerRow()/4;
	dispHeight	= inBitmap->BitsLength()/inBitmap->BytesPerRow();
	bits		= (uint32 *) inBitmap->Bits();
	
	Color32	theColor;
	
	color_space thecs	= inBitmap->ColorSpace();
	
	if (thecs != B_RGB32 && thecs != B_RGB32_BIG && thecs != B_RGBA32)
		return kColorSpaceNotSupported;
	
	ioRegion->MakeEmpty();
	
	for (int y = 0; y < dispHeight; y++)
	{
		int32 lastj = y * dispWidth + dispWidth;
		
		for (int32 j = y * dispWidth; j < lastj; j++)
		{
			theColor.word = bits[j];
			
				// If not transparent make black
			if (theColor.bytes[0] != B_TRANSPARENT_32_BIT.blue ||
				theColor.bytes[1] != B_TRANSPARENT_32_BIT.green ||
				theColor.bytes[2] != B_TRANSPARENT_32_BIT.red)
			{
				int32	x = j - y * dispWidth;
				
				ioRegion->Include(BRect(x, y, x, y));
			}
		}
	}	

	return noErr;
}

