///--------------------------------------------------------------------------------------
//	Sprite.c
//
//	Portions are copyright:  1991-94 Tony Myles, All rights reserved worldwide.
//
//	Description:	implementation of the sprites
///--------------------------------------------------------------------------------------


#ifndef __SWCOMMON__
#include "SWCommonHeaders.h"
#endif

#ifndef __SPRITEWORLDUTILS__
#include "SpriteWorldUtils.h"
#endif

#ifndef __SPRITEWORLD__
#include "SpriteWorld.h"
#endif

#ifndef __SPRITE__
#include "Sprite.h"
#endif

#ifndef __SPRITEFRAME__
#include "SpriteFrame.h"
#endif

#ifndef __BLITPIXIE__
#include "BlitPixie.h"
#endif

extern RgnHandle 		gCollisionSectRgn;
extern RgnHandle		gCollisionSpareRgn;
extern Boolean 			gSWmmuMode;


///--------------------------------------------------------------------------------------
//	CalcMaxWidthHeight
///--------------------------------------------------------------------------------------
static void CalcMaxWidthHeight( 
	RectArrayPtr newRectArrayP, 
	short *maxWidthP, 
	short *maxHeightP )
{
	short 		frame;
	Rect 		*rectP;
	short 		maxWidth, maxHeight;
	
	
	maxWidth = 0;
	maxHeight = 0;
	
	rectP = &newRectArrayP->frameRects[0];
	for( frame = 0; frame < newRectArrayP->numFrames; frame++ ) 
	{
		if ( rectP->right - rectP->left > maxWidth ) 
		{
			maxWidth = rectP->right - rectP->left;
		}
		if ( rectP->bottom - rectP->top > maxHeight ) 
		{
			maxHeight = rectP->bottom - rectP->top;
		}
		rectP++;
	}
	*maxWidthP = maxWidth;
	*maxHeightP =maxHeight;
}

///--------------------------------------------------------------------------------------
//	SWCreateSprite
///--------------------------------------------------------------------------------------

SW_FUNC OSErr SWCreateSprite(
	SpritePtr* newSpriteP,
	void* spriteStorageP,
	short maxFrames)
{
	OSErr err = noErr;
	SpritePtr tempSpriteP;
	Size frameArraySize;

	*newSpriteP = NULL;

	tempSpriteP =  (spriteStorageP == NULL) ? (SpritePtr)NewPtr(sizeof(SpriteRec))
		: (SpritePtr)spriteStorageP;

	if (tempSpriteP != NULL)
	{
		frameArraySize = (Size)maxFrames * sizeof(FramePtr);

		tempSpriteP->frameArray = (FramePtr *)NewPtrClear(frameArraySize);

		if (tempSpriteP->frameArray != NULL)
		{
			tempSpriteP->nextSpriteP = NULL;
			tempSpriteP->prevSpriteP = NULL;

				// initialize drawing fields
			tempSpriteP->isVisible = true;
			tempSpriteP->needsToBeDrawn = true;
			tempSpriteP->needsToBeErased = false;
			tempSpriteP->isUnderTiles = false;
			tempSpriteP->destFrameRect.left = 0;
			tempSpriteP->destFrameRect.top = 0;
			tempSpriteP->destFrameRect.right = 0;
			tempSpriteP->destFrameRect.bottom = 0;
			tempSpriteP->oldFrameRect.left = 0;
			tempSpriteP->oldFrameRect.top = 0;
			tempSpriteP->oldFrameRect.right = 0;
			tempSpriteP->oldFrameRect.bottom = 0;
			tempSpriteP->deltaFrameRect.left = 0;
			tempSpriteP->deltaFrameRect.top = 0;
			tempSpriteP->deltaFrameRect.right = 0;
			tempSpriteP->deltaFrameRect.bottom = 0;
			tempSpriteP->frameDrawProc = SWStdSpriteDrawProc;
			tempSpriteP->scaledFrameRect.left = -1;
			tempSpriteP->scaledFrameRect.top = -1;
			tempSpriteP->scaledFrameRect.right = -1;
			tempSpriteP->scaledFrameRect.bottom = -1;
				
				// initialize drawing fields for scrolling routines
			tempSpriteP->destOffscreenRect.left = 0;
			tempSpriteP->destOffscreenRect.top = 0;
			tempSpriteP->destOffscreenRect.right = 0;
			tempSpriteP->destOffscreenRect.bottom = 0;
			tempSpriteP->oldOffscreenRect.left = 0;
			tempSpriteP->oldOffscreenRect.top = 0;
			tempSpriteP->oldOffscreenRect.right = 0;
			tempSpriteP->oldOffscreenRect.bottom = 0;
			tempSpriteP->oldRectIsVisible = false;

				// initialize frame fields
			tempSpriteP->frameArrayIsShared = false;
			tempSpriteP->curFrameP = NULL;
			tempSpriteP->numFrames = 0;
			tempSpriteP->maxFrames = maxFrames;
			tempSpriteP->frameTimeInterval = -1;
			tempSpriteP->timeOfLastFrameChange = 0;
			tempSpriteP->frameAdvance = 1;
			tempSpriteP->curFrameIndex = 0;
			tempSpriteP->firstFrameIndex = 0;
			tempSpriteP->lastFrameIndex = 0;
			tempSpriteP->frameChangeProc = NULL;

				// initialize movement fields
			tempSpriteP->moveTimeInterval = 0;
			tempSpriteP->timeOfLastMove = 0;
			tempSpriteP->horizMoveDelta = 0;
			tempSpriteP->vertMoveDelta = 0;
			tempSpriteP->moveBoundsRect.left = 0;
			tempSpriteP->moveBoundsRect.top = 0;
			tempSpriteP->moveBoundsRect.right = 0;
			tempSpriteP->moveBoundsRect.bottom = 0;
			tempSpriteP->spriteMoveProc = NULL;

			tempSpriteP->spriteCollideProc = NULL;
			
			tempSpriteP->sharedPictGWorld = NULL;
			tempSpriteP->sharedMaskGWorld = NULL;
			tempSpriteP->spriteRemoval = kSWDontRemoveSprite;
			tempSpriteP->frameAdvanceMode = kSWWrapAroundMode;
			tempSpriteP->userData = 0;

				// the sprite has been successfully created
			*newSpriteP = tempSpriteP;
		}
		else
		{
			err = MemError();

			DisposePtr((Ptr)tempSpriteP);
		}
	}
	else
	{
		err = MemError();
	}

	SWSetStickyIfError( err );
	return err;
}

///--------------------------------------------------------------------------------------
//	SWCloneSprite
///--------------------------------------------------------------------------------------

SW_FUNC OSErr SWCloneSprite(
	SpritePtr	cloneSpriteP,
	SpritePtr*	newSpriteP,
	void*		spriteStorageP)
{
	SpritePtr 		tempSpriteP;
	short 			index;
	Size			frameArraySize;
	OSErr 			err = noErr;

		// Allocate memory for the new Sprite
	tempSpriteP = (spriteStorageP == NULL) ? (SpritePtr)NewPtr(sizeof(SpriteRec)) : (SpritePtr)spriteStorageP;
	if (tempSpriteP == NULL)
		err = MemError();

	if (err == noErr)
	{
			// copy the clone sprite into the temp sprite
		*tempSpriteP = *cloneSpriteP;

			// clear the next and prev fields, just in case
		tempSpriteP->nextSpriteP = NULL;
		tempSpriteP->prevSpriteP = NULL;
		
		tempSpriteP->frameArrayIsShared = false;
	}
		
	if (err == noErr)
	{
			// Create a new frameArray for this sprite
		frameArraySize = (Size)cloneSpriteP->maxFrames * sizeof(FramePtr);
		tempSpriteP->frameArray = (FramePtr *)NewPtrClear(frameArraySize);

		if (tempSpriteP->frameArray != NULL)
		{
				// copy the frame array, and increment useCount for each frame
			for (index = 0; index < cloneSpriteP->maxFrames; index++)
			{
				tempSpriteP->frameArray[index] = cloneSpriteP->frameArray[index];
				tempSpriteP->frameArray[index]->useCount++;
			}
		}
		else
		{
			err = MemError();
			DisposePtr((Ptr)tempSpriteP);
		}
	}
	
	if (err == noErr)
		*newSpriteP = tempSpriteP;

	SWSetStickyIfError( err );
	return err;
}


///--------------------------------------------------------------------------------------
//	SWFastCloneSprite
///--------------------------------------------------------------------------------------

SW_FUNC OSErr SWFastCloneSprite(
	SpritePtr	cloneSpriteP,
	SpritePtr*	newSpriteP,
	void*		spriteStorageP)
{
	SpritePtr 		tempSpriteP;
	OSErr 			err = noErr;

		// Allocate memory for the new Sprite
	tempSpriteP = (spriteStorageP == NULL) ? (SpritePtr)NewPtr(sizeof(SpriteRec)) : (SpritePtr)spriteStorageP;
	if (tempSpriteP == NULL)
		err = MemError();

	if (err == noErr)
	{
			// copy the clone sprite into the temp sprite
		*tempSpriteP = *cloneSpriteP;

			// clear the next and prev fields, just in case
		tempSpriteP->nextSpriteP = NULL;
		tempSpriteP->prevSpriteP = NULL;
		
		tempSpriteP->frameArrayIsShared = true;
		*newSpriteP = tempSpriteP;
	}

	SWSetStickyIfError( err );
	return err;
}


///--------------------------------------------------------------------------------------
//	SWMarkSpriteForRemoval
///--------------------------------------------------------------------------------------

SW_FUNC void  SWMarkSpriteForRemoval(
	SpriteWorldPtr	spriteWorldP,
	SpritePtr		spriteP,
	Boolean			disposeOfSprite)
{
	
	spriteWorldP->anySpritesNeedRemoving = true;
	SWSetSpriteVisible( spriteP, false );
	
	if ( disposeOfSprite )
	{
		spriteP->spriteRemoval = kSWRemoveAndDisposeSprite;
	}
	else
	{
		spriteP->spriteRemoval = kSWRemoveSprite;
	}
}
	

///--------------------------------------------------------------------------------------
//	SWDisposeSprite
///--------------------------------------------------------------------------------------

SW_FUNC void SWDisposeSprite(
	SpritePtr deadSpriteP)
{
	SWCloseSprite(deadSpriteP);

	DisposePtr((Ptr)deadSpriteP);
}


///--------------------------------------------------------------------------------------
//	SWCloseSprite
///--------------------------------------------------------------------------------------

SW_FUNC void SWCloseSprite(
	SpritePtr deadSpriteP)
{
	short 			frame;
	Boolean			framesDisposed;
	

	if (deadSpriteP != NULL)
	{
			// Dispose the frames if they aren't shared, or if this is the master sprite
		if (!deadSpriteP->frameArrayIsShared)
		{
			for (frame = 0; frame < deadSpriteP->numFrames; frame++)
			{
				framesDisposed = SWDisposeFrame(deadSpriteP->frameArray[frame]);
			}
			
			if ( framesDisposed )
			{
				if ( deadSpriteP->sharedPictGWorld != NULL )
				{
//					DisposeGWorld(deadSpriteP->sharedPictGWorld );
					delete deadSpriteP->sharedPictGWorld;
					deadSpriteP->sharedPictGWorld = NULL;
				}
			
				if ( deadSpriteP->sharedMaskGWorld != NULL )
				{
//					DisposeGWorld(deadSpriteP->sharedMaskGWorld );
					delete deadSpriteP->sharedPictGWorld;
					deadSpriteP->sharedMaskGWorld = NULL;
				}
			}
			
			DisposePtr((Ptr)deadSpriteP->frameArray);
		}
	}
}


///--------------------------------------------------------------------------------------
//	SWLockSprite
///--------------------------------------------------------------------------------------

SW_FUNC void SWLockSprite(
	SpritePtr srcSpriteP)
{
	register long numFrames;

	numFrames = srcSpriteP->numFrames;

	while (numFrames--)
	{
		SWLockFrame(srcSpriteP->frameArray[numFrames]);
	}
}


///--------------------------------------------------------------------------------------
//	SWUnlockSprite
///--------------------------------------------------------------------------------------

SW_FUNC void SWUnlockSprite(
	SpritePtr srcSpriteP)
{
	register long numFrames;

	if ( srcSpriteP != NULL )
	{
		numFrames = srcSpriteP->numFrames;
	
		while (numFrames--)
		{
			SWUnlockFrame(srcSpriteP->frameArray[numFrames]);
		}
	}
}


///--------------------------------------------------------------------------------------
//	SWSetSpriteDrawProc
///--------------------------------------------------------------------------------------

SW_FUNC OSErr SWSetSpriteDrawProc(
	SpritePtr srcSpriteP,
	DrawProcPtr drawProc)
{
	OSErr		err = noErr;
	FramePtr	headFrameP = NULL;
	short		framePixelDepth = 0;
	
	if (drawProc == NULL)
		err = kNilParameterErr;

	/*
		Deleted a bunch of checking about a valid drawproc for the current depth.
		Not applicable for BeOS.
	*/
		
	if ( err == noErr )
		srcSpriteP->frameDrawProc = drawProc;
		
	SWSetStickyIfError( err );
	return err;
}


///--------------------------------------------------------------------------------------
//	SWStdSpriteDrawProc - for drawing sprites
///--------------------------------------------------------------------------------------

SW_FUNC void SWStdSpriteDrawProc(
	FramePtr srcFrameP,
	FramePtr dstFrameP,
	Rect* srcRect,
	Rect* dstRect)
{

	if (srcFrameP->maskRgn != NULL)
	{
		/*
			Some modest changes for region rects and point h-v to x-y.
		*/
		Rect rgnRect = srcFrameP->maskRgn->Frame();

			// move the mask region to the new sprite location
		OffsetRgn(srcFrameP->maskRgn,
			(dstRect->left - (srcRect->left - srcFrameP->frameRect.left) - rgnRect.left) 
			+ srcFrameP->offsetPoint.x,
			(dstRect->top - (srcRect->top - srcFrameP->frameRect.top) - rgnRect.top)
			+ srcFrameP->offsetPoint.y);
	}

		// CopyBits with a mask region is generally faster than CopyMask
	CopyBits(srcFrameP->framePix, dstFrameP->framePort, srcRect, dstRect,
		srcCopy, srcFrameP->maskRgn);
}


///--------------------------------------------------------------------------------------
//	SWAddFrame
///--------------------------------------------------------------------------------------

SW_FUNC OSErr SWAddFrame(
	SpritePtr srcSpriteP,
	FramePtr newFrameP)
{
	OSErr err = noErr;

		// don't exceed maximum number of frames
	if (srcSpriteP->numFrames < srcSpriteP->maxFrames)
	{
		srcSpriteP->frameArray[srcSpriteP->numFrames] = newFrameP;

			// increment the number of frames
		srcSpriteP->numFrames++;
		newFrameP->useCount++;
	}
	else
	{
		err = kMaxFramesErr;
	}

	SWSetStickyIfError( err );
	return err;
}


///--------------------------------------------------------------------------------------
//	SWInsertFrame
///--------------------------------------------------------------------------------------

SW_FUNC OSErr SWInsertFrame(
	SpritePtr	srcSpriteP,
	FramePtr	newFrameP,
	long		frameIndex)
{
	register long		index;
	register FramePtr	*frameArray;
	OSErr				err = noErr;
	
	frameArray = srcSpriteP->frameArray;

		// don't exceed maximum number of frames
	if (frameIndex < srcSpriteP->maxFrames && frameIndex >= 0 &&
		srcSpriteP->numFrames < srcSpriteP->maxFrames)
	{
			// move frames above frameIndex up
		for (index = srcSpriteP->numFrames; index > frameIndex; index--)
			frameArray[index] = frameArray[index - 1];
		
			// fill the hole
		frameArray[frameIndex] = newFrameP;

		srcSpriteP->numFrames++;
		newFrameP->useCount++;
	}
	else
	{
		err = kMaxFramesErr;
	}

	SWSetStickyIfError( err );
	return err;
}


///--------------------------------------------------------------------------------------
//	SWRemoveFrame
///--------------------------------------------------------------------------------------

SW_FUNC void SWRemoveFrame(
	SpritePtr srcSpriteP,
	FramePtr oldFrameP)
{
	register long curFrame;
	register FramePtr *frameArray;

	curFrame = srcSpriteP->numFrames;
	frameArray = srcSpriteP->frameArray;

		// find the frame to be removed
	while (curFrame--)
	{
		if (frameArray[curFrame] == oldFrameP)
		{
			srcSpriteP->numFrames--;
			oldFrameP->useCount--;
			
				// move the rest of the frames down
			while (curFrame <= (srcSpriteP->numFrames - 1L))
			{
				frameArray[curFrame] = frameArray[curFrame + 1L];
				curFrame++;
			}

			break;
		}
	}
}


///--------------------------------------------------------------------------------------
//	SWSetCurrentFrame
///--------------------------------------------------------------------------------------

SW_FUNC void SWSetCurrentFrame(
	SpritePtr srcSpriteP,
	FramePtr newFrameP)
{
	short horizOffset, vertOffset;
	short frameIndex = 0;

	if (srcSpriteP->curFrameP != newFrameP)
	{
		horizOffset = (srcSpriteP->destFrameRect.left - newFrameP->frameRect.left) +
					(srcSpriteP->curFrameP->hotSpotH - newFrameP->hotSpotH);
		vertOffset = (srcSpriteP->destFrameRect.top - newFrameP->frameRect.top) +
					(srcSpriteP->curFrameP->hotSpotV - newFrameP->hotSpotV);
		
		srcSpriteP->curFrameP = newFrameP;
		srcSpriteP->destFrameRect = newFrameP->frameRect;

		if (srcSpriteP->scaledFrameRect.right > 0)
			srcSpriteP->destFrameRect.right = srcSpriteP->destFrameRect.left + 
												srcSpriteP->scaledFrameRect.right;
		if (srcSpriteP->scaledFrameRect.bottom > 0)
			srcSpriteP->destFrameRect.bottom = srcSpriteP->destFrameRect.top + 
												srcSpriteP->scaledFrameRect.bottom;
		
		srcSpriteP->destFrameRect.left += horizOffset;
		srcSpriteP->destFrameRect.right += horizOffset;
		srcSpriteP->destFrameRect.top += vertOffset;
		srcSpriteP->destFrameRect.bottom += vertOffset;

			// this really lame search is performed so we set the
			// frame index of the sprite properly, which could
			// prevent strange behavior later on. The lesson to be learned
			// here is to call SWSetCurrentFrameIndex instead.
		for (frameIndex = 0; frameIndex < srcSpriteP->numFrames; frameIndex++)
		{
			if (srcSpriteP->frameArray[frameIndex] == newFrameP)
			{
				srcSpriteP->curFrameIndex = frameIndex;
				break;
			}
		}

		srcSpriteP->needsToBeDrawn = true;
	}
}


///--------------------------------------------------------------------------------------
//	SWSetCurrentFrameIndex
///--------------------------------------------------------------------------------------

SW_FUNC void SWSetCurrentFrameIndex(
	SpritePtr srcSpriteP,
	short frameIndex)
{
	register FramePtr newFrameP;
	short horizOffset, vertOffset;

	if (frameIndex < srcSpriteP->numFrames)
	{
		newFrameP = srcSpriteP->frameArray[frameIndex];
		
			// If this is a brand new sprite, it won't have a current frame.
			// (This is used when calculating the hotSpot offset from old to new frame.)
		if (srcSpriteP->curFrameP == NULL)
			srcSpriteP->curFrameP = newFrameP;

		horizOffset = (srcSpriteP->destFrameRect.left - newFrameP->frameRect.left) +
					(srcSpriteP->curFrameP->hotSpotH - newFrameP->hotSpotH);
		vertOffset = (srcSpriteP->destFrameRect.top - newFrameP->frameRect.top) +
					(srcSpriteP->curFrameP->hotSpotV - newFrameP->hotSpotV);
		
		srcSpriteP->curFrameP = newFrameP;
		srcSpriteP->destFrameRect = newFrameP->frameRect;

		if (srcSpriteP->scaledFrameRect.right > 0)
			srcSpriteP->destFrameRect.right = (srcSpriteP->destFrameRect).left + 
												(srcSpriteP->scaledFrameRect).right;
		if (srcSpriteP->scaledFrameRect.bottom > 0)
			srcSpriteP->destFrameRect.bottom = srcSpriteP->destFrameRect.top + 
												srcSpriteP->scaledFrameRect.bottom;
		
		srcSpriteP->destFrameRect.left += horizOffset;
		srcSpriteP->destFrameRect.right += horizOffset;
		srcSpriteP->destFrameRect.top += vertOffset;
		srcSpriteP->destFrameRect.bottom += vertOffset;

		srcSpriteP->curFrameIndex = frameIndex;

		srcSpriteP->needsToBeDrawn = true;
	}
}


///--------------------------------------------------------------------------------------
//	SWSetSpriteFrameAdvance
///--------------------------------------------------------------------------------------

SW_FUNC void SWSetSpriteFrameAdvance(
	SpritePtr srcSpriteP,
	short frameAdvance)
{
	srcSpriteP->frameAdvance = frameAdvance;
}


///--------------------------------------------------------------------------------------
//	SWSetSpriteFrameAdvanceMode
///--------------------------------------------------------------------------------------

SW_FUNC void SWSetSpriteFrameAdvanceMode(
	SpritePtr srcSpriteP,
	AdvanceType advanceMode)
{
	srcSpriteP->frameAdvanceMode = advanceMode;
}


///--------------------------------------------------------------------------------------
//	SWSetSpriteFrameRange
///--------------------------------------------------------------------------------------

SW_FUNC void SWSetSpriteFrameRange(
	SpritePtr srcSpriteP,
	short firstFrameIndex,
	short lastFrameIndex)
{
	if ( firstFrameIndex < 0 )
		firstFrameIndex = 0;
	
	if ( lastFrameIndex > srcSpriteP->maxFrames - 1 )
		lastFrameIndex = srcSpriteP->maxFrames - 1;
	
	srcSpriteP->firstFrameIndex = firstFrameIndex;
	srcSpriteP->lastFrameIndex = lastFrameIndex;
	
		// Make sure the sprite's curFrameIndex is within the new frame range
	if (srcSpriteP->curFrameIndex < firstFrameIndex)
		SWSetCurrentFrameIndex(srcSpriteP, firstFrameIndex);
	else if (srcSpriteP->curFrameIndex > lastFrameIndex)
		SWSetCurrentFrameIndex(srcSpriteP, lastFrameIndex);
}


///--------------------------------------------------------------------------------------
//	SWSetSpriteFrameTime
///--------------------------------------------------------------------------------------

SW_FUNC void SWSetSpriteFrameTime(
	SpritePtr srcSpriteP,
	long timeInterval)
{
	srcSpriteP->frameTimeInterval = timeInterval;
}


///--------------------------------------------------------------------------------------
//	SWSetSpriteFrameProc
///--------------------------------------------------------------------------------------

SW_FUNC void SWSetSpriteFrameProc(
	SpritePtr srcSpriteP,
	FrameProcPtr frameProc)
{
	srcSpriteP->frameChangeProc = frameProc;
}


///--------------------------------------------------------------------------------------
//	SWMoveSprite
///--------------------------------------------------------------------------------------

SW_FUNC void SWMoveSprite(
	SpritePtr srcSpriteP,
	short horizLoc,
	short vertLoc)
{
	horizLoc -= srcSpriteP->curFrameP->hotSpotH;
	vertLoc -= srcSpriteP->curFrameP->hotSpotV;
	
	if ((horizLoc != srcSpriteP->destFrameRect.left) || 
		(vertLoc != srcSpriteP->destFrameRect.top) )
	{
		srcSpriteP->destFrameRect.bottom = vertLoc + (srcSpriteP->destFrameRect.bottom - 
													srcSpriteP->destFrameRect.top);
		srcSpriteP->destFrameRect.right = horizLoc + (srcSpriteP->destFrameRect.right - 
													srcSpriteP->destFrameRect.left);
		srcSpriteP->destFrameRect.top = vertLoc;
		srcSpriteP->destFrameRect.left = horizLoc;
	
		srcSpriteP->needsToBeDrawn = true;
	}
}


///--------------------------------------------------------------------------------------
//	SWOffsetSprite
///--------------------------------------------------------------------------------------

SW_FUNC void SWOffsetSprite(
	SpritePtr srcSpriteP,
	short horizOffset,
	short vertOffset)
{
	if ((horizOffset != 0) || (vertOffset != 0))
	{
		srcSpriteP->destFrameRect.right += horizOffset;
		srcSpriteP->destFrameRect.bottom += vertOffset;
		srcSpriteP->destFrameRect.left += horizOffset;
		srcSpriteP->destFrameRect.top += vertOffset;
		
		srcSpriteP->needsToBeDrawn = true;
	}
}


///--------------------------------------------------------------------------------------
//	SWSetSpriteLocation
///--------------------------------------------------------------------------------------

SW_FUNC void SWSetSpriteLocation(
	SpritePtr srcSpriteP,
	short horizLoc,
	short vertLoc)
{
	horizLoc -= srcSpriteP->curFrameP->hotSpotH;
	vertLoc -= srcSpriteP->curFrameP->hotSpotV;
	
	srcSpriteP->destFrameRect.bottom = vertLoc + (srcSpriteP->destFrameRect.bottom - 
												srcSpriteP->destFrameRect.top);
	srcSpriteP->destFrameRect.right = horizLoc + (srcSpriteP->destFrameRect.right - 
												srcSpriteP->destFrameRect.left);
	srcSpriteP->destFrameRect.top = vertLoc;
	srcSpriteP->destFrameRect.left = horizLoc;

	srcSpriteP->deltaFrameRect = srcSpriteP->destFrameRect;
	srcSpriteP->oldFrameRect = srcSpriteP->destFrameRect;
	srcSpriteP->needsToBeDrawn = true;
}


///--------------------------------------------------------------------------------------
//	SWSetSpriteMoveBounds
///--------------------------------------------------------------------------------------

SW_FUNC void SWSetSpriteMoveBounds(
	SpritePtr srcSpriteP,
	Rect* moveBoundsRect)
{
	srcSpriteP->moveBoundsRect = *moveBoundsRect;
}


///--------------------------------------------------------------------------------------
//	SWSetSpriteMoveDelta
///--------------------------------------------------------------------------------------

SW_FUNC void SWSetSpriteMoveDelta(
	SpritePtr srcSpriteP,
	short horizDelta,
	short vertDelta)
{
		// set sprite's velocity
	srcSpriteP->horizMoveDelta = horizDelta;
	srcSpriteP->vertMoveDelta = vertDelta;
}


///--------------------------------------------------------------------------------------
//	SWSetSpriteMoveTime
///--------------------------------------------------------------------------------------

SW_FUNC void SWSetSpriteMoveTime(
	SpritePtr srcSpriteP,
	long timeInterval)
{
	srcSpriteP->moveTimeInterval = timeInterval;
}


///--------------------------------------------------------------------------------------
//	SWSetSpriteMoveProc
///--------------------------------------------------------------------------------------

SW_FUNC void SWSetSpriteMoveProc(
	SpritePtr srcSpriteP,
	MoveProcPtr moveProc)
{
	srcSpriteP->spriteMoveProc = moveProc;
}


///--------------------------------------------------------------------------------------
//	SWBounceSprite
///--------------------------------------------------------------------------------------

SW_FUNC Boolean SWBounceSprite(
	SpritePtr srcSpriteP)
{
	short	horizOffset, vertOffset;

		// Hit left side or right side
	if (srcSpriteP->destFrameRect.left < srcSpriteP->moveBoundsRect.left)
	{
		srcSpriteP->horizMoveDelta = -srcSpriteP->horizMoveDelta;
		horizOffset = (srcSpriteP->moveBoundsRect.left - srcSpriteP->destFrameRect.left) * 2;
	}
	else if (srcSpriteP->destFrameRect.right > srcSpriteP->moveBoundsRect.right)
	{
		srcSpriteP->horizMoveDelta = -srcSpriteP->horizMoveDelta;
		horizOffset = (srcSpriteP->moveBoundsRect.right - srcSpriteP->destFrameRect.right) * 2;
	}
	else
		horizOffset = 0;


		// Hit top or bottom
	if (srcSpriteP->destFrameRect.top < srcSpriteP->moveBoundsRect.top)
	{
		srcSpriteP->vertMoveDelta = -srcSpriteP->vertMoveDelta;
		vertOffset = (srcSpriteP->moveBoundsRect.top - srcSpriteP->destFrameRect.top) * 2;
	}
	else if (srcSpriteP->destFrameRect.bottom > srcSpriteP->moveBoundsRect.bottom)
	{
		srcSpriteP->vertMoveDelta = -srcSpriteP->vertMoveDelta;
		vertOffset = (srcSpriteP->moveBoundsRect.bottom - srcSpriteP->destFrameRect.bottom) * 2;
	}
	else
		vertOffset = 0;
	
	
		// Bounce the sprite if it had a collision with the wall
	if ((horizOffset != 0) || (vertOffset != 0))
	{
		srcSpriteP->destFrameRect.right += horizOffset;
		srcSpriteP->destFrameRect.bottom += vertOffset;
		srcSpriteP->destFrameRect.left += horizOffset;
		srcSpriteP->destFrameRect.top += vertOffset;
		srcSpriteP->needsToBeDrawn = true;
		
		return true;
	}
	else
	{
		return false;
	}
}


///--------------------------------------------------------------------------------------
//	SWWrapSprite
///--------------------------------------------------------------------------------------

SW_FUNC Boolean SWWrapSprite(
	SpritePtr srcSpriteP)
{
	short	horizOffset, vertOffset;
	
		// Wrap to left or right
	if (srcSpriteP->oldFrameRect.left > srcSpriteP->moveBoundsRect.right)
	{
		horizOffset = - ((srcSpriteP->moveBoundsRect.right -
						srcSpriteP->moveBoundsRect.left) +
						(srcSpriteP->destFrameRect.right - 
						srcSpriteP->destFrameRect.left) + srcSpriteP->horizMoveDelta);
	}
	else if (srcSpriteP->oldFrameRect.right < srcSpriteP->moveBoundsRect.left)
	{
		horizOffset = (srcSpriteP->moveBoundsRect.right -
						srcSpriteP->moveBoundsRect.left) +
						(srcSpriteP->destFrameRect.right - 
						srcSpriteP->destFrameRect.left) - srcSpriteP->horizMoveDelta;
	}
	else
		horizOffset = 0;
	
	
		// Wrap to top or bottom
	if (srcSpriteP->oldFrameRect.top > srcSpriteP->moveBoundsRect.bottom)
	{
		vertOffset = -((srcSpriteP->moveBoundsRect.bottom -
							srcSpriteP->moveBoundsRect.top) +
							(srcSpriteP->destFrameRect.bottom - 
							srcSpriteP->destFrameRect.top) + srcSpriteP->vertMoveDelta);
	}
	else if (srcSpriteP->oldFrameRect.bottom < srcSpriteP->moveBoundsRect.top)
	{
		vertOffset = (srcSpriteP->moveBoundsRect.bottom -
							srcSpriteP->moveBoundsRect.top) +
							(srcSpriteP->destFrameRect.bottom - 
							srcSpriteP->destFrameRect.top) - srcSpriteP->vertMoveDelta;
	}
	else
		vertOffset = 0;
	
	
		// Wrap sprite, and set delta and old rects to current rect
	if ((horizOffset != 0) || (vertOffset != 0))
	{
		srcSpriteP->destFrameRect.right += horizOffset;
		srcSpriteP->destFrameRect.bottom += vertOffset;
		srcSpriteP->destFrameRect.left += horizOffset;
		srcSpriteP->destFrameRect.top += vertOffset;
		
		srcSpriteP->deltaFrameRect = srcSpriteP->destFrameRect;
		srcSpriteP->oldFrameRect = srcSpriteP->destFrameRect;
		srcSpriteP->needsToBeDrawn = true;
		
		return true;
	}
	else
	{
		return false;
	}
}


///--------------------------------------------------------------------------------------
//	SWSetSpriteCollideProc
///--------------------------------------------------------------------------------------

SW_FUNC void SWSetSpriteCollideProc(
	SpritePtr srcSpriteP,
	CollideProcPtr collideProc)
{
	srcSpriteP->spriteCollideProc = collideProc;
}


///--------------------------------------------------------------------------------------
// SWRegionCollision
///--------------------------------------------------------------------------------------
 
SW_FUNC Boolean SWRegionCollision(
	SpritePtr srcSpriteP,
	SpritePtr dstSpriteP)
{
	Rect 				rgnRect;
 
	/*
		Some minor graphics functions changes for region and point
		differences.
	*/
		// copy one of the regions in case they're the same 
		// (would happen if the sprites are clones using the same frame)
	gCollisionSpareRgn->MakeEmpty();
	gCollisionSpareRgn->Include(dstSpriteP->curFrameP->maskRgn);	

	rgnRect = srcSpriteP->curFrameP->maskRgn->Frame();

		// move the mask region to the new sprite location

	OffsetRgn(srcSpriteP->curFrameP->maskRgn,
				(srcSpriteP->destFrameRect.left - rgnRect.left) +
				srcSpriteP->curFrameP->offsetPoint.x,
				(srcSpriteP->destFrameRect.top - rgnRect.top) +
				srcSpriteP->curFrameP->offsetPoint.y);

	rgnRect = gCollisionSpareRgn->Frame();

		// move the mask region to the new sprite location

	OffsetRgn(gCollisionSpareRgn,
				(dstSpriteP->destFrameRect.left - rgnRect.left) +
				dstSpriteP->curFrameP->offsetPoint.x,
				(dstSpriteP->destFrameRect.top - rgnRect.top) +
				dstSpriteP->curFrameP->offsetPoint.y);

	gCollisionSpareRgn->IntersectWith(srcSpriteP->curFrameP->maskRgn);
	
	return (!EmptyRgn(gCollisionSectRgn));

}


///--------------------------------------------------------------------------------------
// SWRadiusCollision
///--------------------------------------------------------------------------------------
 
SW_FUNC Boolean SWRadiusCollision(
	SpritePtr srcSpriteP,
	SpritePtr dstSpriteP)
{
	Point		 		centerOne, centerTwo;
	register short		radiusOne, radiusTwo;
	register short		horizDistance, vertDistance;

		// get radii - - assumes circular sprites!
	radiusOne = (srcSpriteP->destFrameRect.right-srcSpriteP->destFrameRect.left)/2;
	radiusTwo = (dstSpriteP->destFrameRect.right-dstSpriteP->destFrameRect.left)/2;
	
	/*
		Minor change to reflect point differences
	*/
	
	centerOne.x = srcSpriteP->destFrameRect.left + radiusOne;
	centerOne.y = srcSpriteP->destFrameRect.top + radiusOne;
	centerTwo.x = dstSpriteP->destFrameRect.left + radiusTwo;
	centerTwo.y = dstSpriteP->destFrameRect.top + radiusTwo;
					
	horizDistance = centerOne.x - centerTwo.x;
	vertDistance = centerOne.y - centerTwo.y;

	return ( (horizDistance*horizDistance) + (vertDistance*vertDistance) <
		((radiusOne+radiusTwo)*(radiusOne+radiusTwo)) );
}


///--------------------------------------------------------------------------------------
// SWPixelCollision
///--------------------------------------------------------------------------------------

SW_FUNC Boolean SWPixelCollision(
	SpritePtr srcSpriteP,
	SpritePtr dstSpriteP)
{
	Boolean			collision;
	Rect			sectRect;
	FramePtr		srcFrameP,
					dstFrameP;
	short			numBytesPerRow;
	short			srcPieceTop,
					srcPieceLeft,
					dstPieceTop,
					dstPieceLeft;
	unsigned long 	srcBaseOffset,
					dstBaseOffset;
	unsigned char*	srcPixelPtr;
	unsigned char*	dstPixelPtr;
	short			i, 
					j;
	
		
	collision = false;
		
	sectRect.left =
		SW_MAX(srcSpriteP->destFrameRect.left, dstSpriteP->destFrameRect.left);
	sectRect.top =
		SW_MAX(srcSpriteP->destFrameRect.top, dstSpriteP->destFrameRect.top);
	sectRect.right =
		SW_MIN(srcSpriteP->destFrameRect.right, dstSpriteP->destFrameRect.right);
	sectRect.bottom =
		SW_MIN(srcSpriteP->destFrameRect.bottom, dstSpriteP->destFrameRect.bottom);
	
	if ( sectRect.right > sectRect.left && sectRect.bottom > sectRect.top )
	{
		
		srcFrameP = srcSpriteP->curFrameP;
		dstFrameP = dstSpriteP->curFrameP;
		
			// Make sure we're in 8-bit
//		if (srcFrameP->framePix->pixelSize != 8)
//			return false;
	
			// calculate the top left of the intersecting portions of the frameRects
		srcPieceTop = srcFrameP->frameRect.top + (sectRect.top-srcSpriteP->destFrameRect.top);
		srcPieceLeft = srcFrameP->frameRect.left + (sectRect.left-srcSpriteP->destFrameRect.left);
		dstPieceTop = dstFrameP->frameRect.top + (sectRect.top-dstSpriteP->destFrameRect.top);
		dstPieceLeft = dstFrameP->frameRect.left + (sectRect.left-dstSpriteP->destFrameRect.left);
		
			// calculate the offset to the first byte of the sectRect of srcSprite
//		srcBaseOffset = srcFrameP->scanLinePtrArray[srcPieceTop - srcFrameP->frameRect.top] + 
//			srcPieceLeft;
			
			// ... and for the dstSprite
//		dstBaseOffset = dstFrameP->scanLinePtrArray[dstPieceTop - dstFrameP->frameRect.top] + 
//			dstPieceLeft;
		
			// calculate pointers to the mask pixels within the overlapping sections
		srcPixelPtr = (unsigned char*)(srcFrameP->maskBaseAddr + srcBaseOffset);
		dstPixelPtr = (unsigned char*)(dstFrameP->maskBaseAddr + dstBaseOffset);
		
			// calculate the number of bytes in a row
		numBytesPerRow = (sectRect.right - sectRect.left);
		
		i = 0;
		while ( i < sectRect.bottom-sectRect.top && !collision )
		{
			j = 0;
			while ( j < numBytesPerRow && !collision )
			{
				j++;
				collision = ( (*srcPixelPtr | *dstPixelPtr) == 0 );
				srcPixelPtr++;
				dstPixelPtr++;
			}
			i++;
			srcPixelPtr += (unsigned long)(srcFrameP->frameRowBytes - numBytesPerRow);
			dstPixelPtr += (unsigned long)(dstFrameP->frameRowBytes - numBytesPerRow);
		}
	}
	return collision;
}


///--------------------------------------------------------------------------------------
//	SWSetSpriteVisible
///--------------------------------------------------------------------------------------

SW_FUNC void SWSetSpriteVisible(
	SpritePtr srcSpriteP,
	Boolean isVisible)
{
	srcSpriteP->isVisible = isVisible;
	srcSpriteP->needsToBeDrawn = true;
	srcSpriteP->needsToBeErased = !isVisible;
}


///--------------------------------------------------------------------------------------
//	SWIsSpriteInRect
///--------------------------------------------------------------------------------------

SW_FUNC Boolean SWIsSpriteInRect(
	SpritePtr srcSpriteP,
	Rect* testRect)
{
	return ((srcSpriteP->destFrameRect.top < testRect->bottom) &&
			(srcSpriteP->destFrameRect.bottom > testRect->top) &&
		 	(srcSpriteP->destFrameRect.left < testRect->right) &&
		 	(srcSpriteP->destFrameRect.right > testRect->left));
}

///--------------------------------------------------------------------------------------
//	SWIsSpriteFullyInRect
///--------------------------------------------------------------------------------------

SW_FUNC Boolean SWIsSpriteFullyInRect(
	SpritePtr srcSpriteP,
	Rect* testRect)
{
	return ((srcSpriteP->destFrameRect.top >= testRect->top) &&
			(srcSpriteP->destFrameRect.bottom <= testRect->bottom) &&
		 	(srcSpriteP->destFrameRect.left >= testRect->left) &&
		 	(srcSpriteP->destFrameRect.right <= testRect->right));
}

///--------------------------------------------------------------------------------------
//	SWIsPointInSprite
///--------------------------------------------------------------------------------------

SW_FUNC Boolean SWIsPointInSprite(
	SpritePtr srcSpriteP,
	Point testPoint)
{
	/*
		Minor changes for point differences
	*/
	return	(testPoint.x >= srcSpriteP->destFrameRect.left) &&
			(testPoint.x <= srcSpriteP->destFrameRect.right) &&
			(testPoint.y >= srcSpriteP->destFrameRect.top) &&
			(testPoint.y <= srcSpriteP->destFrameRect.bottom);

}

///--------------------------------------------------------------------------------------
//	SWCreateSpriteFromIcon
///--------------------------------------------------------------------------------------

SW_FUNC OSErr SWCreateSpriteFromIcon(
	SpriteWorldPtr destSpriteWorld,
	SpritePtr* newSpriteP,
	void* spriteStorageP,
	type_code	iconType,
	short cIconID,
	short maxFrames,
	MaskType maskType,
	BResources		*resFile)
{
	OSErr 			err;
	SpritePtr 		tempSpriteP;
	FramePtr 		newFrameP;
	short 			frame;

	*newSpriteP = NULL;

	err = SWCreateSprite(&tempSpriteP, spriteStorageP, maxFrames);
	
	if (err == noErr)
	{
		for (frame = 0; frame < maxFrames; frame++)
		{
			err = SWCreateFrameFromIconResource(&newFrameP, resFile, iconType,
					cIconID + frame);

			if (err == noErr)
			{
				err = SWAddFrame(tempSpriteP, newFrameP);
			}

			if (err != noErr)
			{
				SWDisposeFrame(newFrameP);
				SWDisposeSprite(tempSpriteP);
				break;
			}
		}

		if (err == noErr)
		{
			SWSetSpriteFrameRange(tempSpriteP, 0, maxFrames - 1);
			SWSetCurrentFrameIndex(tempSpriteP, 0);
			*newSpriteP = tempSpriteP;
		}
	}

	SWSetStickyIfError( err );
	return err;
}

///--------------------------------------------------------------------------------------
//	SWCreateSpriteFromSingleFileXY
///--------------------------------------------------------------------------------------

OSErr
SWCreateSpriteFromSingleFileXY(SpriteWorldPtr destSpriteWorld,
	SpritePtr	*newSpriteP,
	void		*spriteStorageP,
	const char 	*pictPath,
	const char	*maskPath,
	short		frameWidth,
	short		frameHeight,
	short		borderWidth,
	short		borderHeight,
	short		maxFrames,
	MaskType	maskType)
{
	OSErr			err = noErr;
	BBitmap			*tempPictBitmap, 
					*tempMaskBitmap, 
					*tempTempMaskBitmap;
	Rect		    frameRect;
	SpritePtr		tempSpriteP;
	FramePtr		newFrameP;
	short			frame;


	tempPictBitmap = NULL;
	tempMaskBitmap = NULL;
	*newSpriteP = NULL;
	
	if ( err == noErr )
	{	
		{
			/*
				Replace the CreateGWorld functions with equivalents for creating a
				bitmap from a file.
			*/
			err = SWCreateBitmapFromFile(destSpriteWorld, &tempPictBitmap, pictPath, true);
			
			if ( err == noErr && maskType != kNoMask )
				err = SWCreateBitmapFromFile(destSpriteWorld, &tempMaskBitmap, maskPath, true);

			/*
				Check if self-masking.
			*/
			if (err == noErr  && maskPath != NULL)
			{
				if ( strcmp(pictPath, maskPath) == 0 && maskType != kNoMask)
				{
					/*
						Switch to call SWBlackenBitmap from GWorld
					*/
					err = SWBlackenBitmap( tempMaskBitmap );
				}
			}
			
			if ( maxFrames == 0 )
			{
				/*
					BBitmaps has Bounds() vs. GWorld portRects.
				*/
				maxFrames = ((short)(tempPictBitmap->Bounds().right + 1)/frameWidth) *
							((short)(tempPictBitmap->Bounds().bottom + 1)/frameHeight);
			}
			if (err == noErr)
			{
				err = SWCreateSprite(&tempSpriteP, spriteStorageP, maxFrames);
			}
			if (err == noErr)
			{
				err = SWCreateFrameFromBitmapAndRectStart( tempTempMaskBitmap, frameWidth, 
						frameHeight, maskType );
			}
			if (err == noErr)
			{
				frameRect.left = 0;
				frameRect.top = 0;
				frameRect.right = frameWidth - 1;
				frameRect.bottom = frameHeight - 1;
				
				for (frame = 0; frame < maxFrames; frame++)
				{
					err = SWCreateFrameFromBitmapAndRectPartial( &newFrameP, tempPictBitmap,
							tempMaskBitmap, tempTempMaskBitmap, &frameRect, maskType);

					if (err == noErr)
					{
						err = SWAddFrame(tempSpriteP, newFrameP);
						if ( frame < maxFrames-1 )
						{
							frameRect.left += frameWidth + borderWidth;
							frameRect.right = frameRect.left + frameWidth - 1;
							/*
							BBitmaps has Bounds() vs. GWorld portRects.
							*/
							if ( frameRect.right > tempPictBitmap->Bounds().right )
							{
								frameRect.left = 0;
								frameRect.right = frameWidth - 1;
								frameRect.top += frameHeight + borderHeight;
								frameRect.bottom = frameRect.top + frameHeight - 1;
								
								if ( frameRect.bottom > tempPictBitmap->Bounds().bottom )
								{
									err = kOutOfRangeErr;
								}
							}
						}
					}
					if ( err != noErr )
					{
						SWDisposeFrame(newFrameP);
						SWDisposeSprite(tempSpriteP);
						break;
					}
				} // end for
				SWCreateFrameFromBitmapAndRectFinish( tempTempMaskBitmap );
			}

			if (err == noErr)
			{
					// make a pixel mask
				if ((maskType & kPixelMask) != 0)
				{
					if (destSpriteWorld->pixelDepth <= 8)
					{
//						(void)LockPixels( GetGWorldPixMap(tempMaskGWorldP) );
//						SetGWorld(*tempMaskBitmap, nil);
//						InvertRect(&tempMaskGWorldP->portRect);
//						UnlockPixels( GetGWorldPixMap(tempMaskGWorldP) );
					}
				}		// If no pixel Mask wanted, dispose
				else	// of the GWorld we used to make region
				{
//					DisposeGWorld( tempMaskGWorldP );
					tempMaskBitmap = NULL;
				}
				
				SWSetSpriteFrameRange(tempSpriteP, 0, maxFrames - 1);
				SWSetCurrentFrameIndex(tempSpriteP, 0);
				tempSpriteP->sharedPictGWorld = tempPictBitmap;
				tempSpriteP->sharedMaskGWorld = tempMaskBitmap;

				*newSpriteP = tempSpriteP;
			}
		}
	}
	SWSetStickyIfError( err );
	return err;

}
