///--------------------------------------------------------------------------------------
//	SpriteFrame.c
//
//	Portions are copyright:  1991-94 Tony Myles, All rights reserved worldwide.
//
//	Description:	implementation of the frame stuff
///--------------------------------------------------------------------------------------


#ifndef __SWCOMMON__
#include "SWCommonHeaders.h"
#endif

#ifndef __SPRITEWORLDUTILS__
#include "SpriteWorldUtils.h"
#endif

#ifndef __SPRITECOMPILER__
#include "SpriteCompiler.h"
#endif

#ifndef __SPRITEFRAME__
#include "SpriteFrame.h"
#endif


///--------------------------------------------------------------------------------------
//	SWCreateFrame
///--------------------------------------------------------------------------------------

SW_FUNC OSErr SWCreateFrame(
	GDHandle	theGDH,
	FramePtr*	newFrameP,
	Rect*		frameRect,
	short		depth)
{
	OSErr 			err = noErr;
	GWorldPtr		frameGWorld;
	FramePtr 		tempFrameP;
	long 			numScanLines;

	/****************************************************************************
	We need a BBitmap to replicate the GWorld functionality of the frame.  
	****************************************************************************/	
	BBitmap			*frameBitmap;
	
	
	*newFrameP = NULL;
	frameGWorld = NULL;
	
		
	numScanLines = frameRect->bottom - frameRect->top;

	tempFrameP = (FramePtr)NewPtrClear((Size)sizeof(FrameRec) + 
		(sizeof(unsigned long) * (numScanLines+1)));
	
	if (tempFrameP != NULL)
	{
		/********************************************************************
		Instead of a new GWorld we need the new bitmap at an appropriate
		depth.  Depending on the intent of the frame, we may want a BView
		attached to the bitmap, so we pass true to acceptsViews, but don't
		actually create the view here. 
		********************************************************************/	
		
		frameBitmap = new BBitmap(*frameRect, ColorSpaceFromDepth(depth),
			true, false);
			
		if (frameBitmap == NULL)
			err = kMemoryError;
		
		if (err == noErr)
		{

			tempFrameP->framePix = frameBitmap;

			tempFrameP->framePort = frameGWorld;
			tempFrameP->frameRect = *frameRect;
			tempFrameP->hotSpotH = 0;
			tempFrameP->hotSpotV = 0;
			tempFrameP->tileMaskIsSolid = false;
			tempFrameP->numScanLines = numScanLines;
			tempFrameP->sharesGWorld = false;
			tempFrameP->isFrameLocked = false;
	
			SWInitializeFrame( tempFrameP, depth );
			
			*newFrameP = tempFrameP;
		}
	}
	else
	{
		err = MemError();
	}

	if (err != noErr)
	{
		if (tempFrameP != NULL)
		{
			if (tempFrameP->framePort != NULL)
			{
				tempFrameP->framePix->RemoveChild(tempFrameP->framePort);
				delete tempFrameP->framePort;
			}
			if (tempFrameP->framePix != NULL)
				delete tempFrameP->framePix;
				DisposeGWorld(tempFrameP->framePort);
			DisposePtr((Ptr)tempFrameP);
		}
	}
	
	SWSetStickyIfError( err );
	return err;
}


///--------------------------------------------------------------------------------------
//	SWCreateWindowFrame
///--------------------------------------------------------------------------------------

SW_FUNC OSErr SWCreateWindowFrame(
	WindowPtr	srcWindowP,
	BView*		srcViewP,
	FramePtr* 	newFrameP,
	Rect* 		frameRect,
	short		maxHeight)
{
	OSErr 			err = noErr;
	GDHandle		currentGDH;
	GWorldPtr		windowGWorld;
	FramePtr 		tempFrameP = NULL;
	short			depth;
	long 			numScanLines;
	Point			globalOffset;
	
	*newFrameP = NULL;
	
	GetGWorld( &windowGWorld, &currentGDH );
		// more scanlines than we need, but this saves us from 
		// needing to calculate an offset in BlitPixie, and allows 
		// blitting to outside the worldRect.
		 
	/****************************************************************************
	 We can't globally get the current window like GetGWorld, so we have to pass
	 the window to the function, which we then use for all of the size
	 calculations.  Also, the true number of scan lines for Be is size of the
	 rect + 1.
	****************************************************************************/
	numScanLines = srcWindowP->Bounds().bottom - srcWindowP->Bounds().top + 1;
	
	numScanLines = SW_MAX(numScanLines, maxHeight);
	
		// Clip frameRect with the window's boundaries
	frameRect->top = SW_MAX(frameRect->top, srcWindowP->Bounds().top);
	frameRect->left = SW_MAX(frameRect->left, srcWindowP->Bounds().left);
	frameRect->bottom = SW_MIN(frameRect->bottom, srcWindowP->Bounds().bottom);
	frameRect->right = SW_MIN(frameRect->right, srcWindowP->Bounds().right);
	
	/********************************************************************
	 This may matter for direct to screen blitters but for now it is
	 superfluous.
	********************************************************************/
	globalOffset.Set(0, 0);
	
	tempFrameP = (FramePtr)NewPtrClear((Size)sizeof(FrameRec) + 
			(sizeof(unsigned long) * numScanLines));

	if (tempFrameP != NULL)
	{
		depth = DepthFromWindow(srcWindowP);

		/********************************************************************
		 We have the window - now we need to create a BView to draw into.  We
		 set this into the framePort member.  We don't follow any resizing
		 for now because it would screw up the scanlines, etc...
		********************************************************************/
		if (srcViewP != NULL)
		{
			tempFrameP->framePort = srcViewP;
		}
		else
		{
			tempFrameP->framePort = new BView(*frameRect, "wind_frame", B_FOLLOW_NONE,
				B_WILL_DRAW);
			
			if (tempFrameP->framePort->Window()->Lock()){	
				tempFrameP->framePort->SetViewColor(B_TRANSPARENT_32_BIT);
			}
		}
		
		tempFrameP->frameRect = *frameRect;
		
		// cache frameRowBytes for use by our blitter routine

		/********************************************************************
		 Until we use a BDirectWindow all the frameRowBytes stuff for the
		 window frame is a moot point.
		********************************************************************/
		// tempFrameP->frameRowBytes = (**tempFrameP->framePort->portPixMap).rowBytes & 0x7FFF;

			// this calculation generates a mask value that we use to
			// long word align the rectangle when we draw the frame.
			// note that the expression "sizeof(long) * kBitsPerByte" gives us
			// the number of bits in a long.
	
			// align on long word boundary
		tempFrameP->rightAlignFactor = ((sizeof(long) * kBitsPerByte) / depth) - 1;
		tempFrameP->leftAlignFactor = ~(tempFrameP->rightAlignFactor);

			// useCount not used for window/work/back frames
		tempFrameP->useCount = 0;

			// here we set up an array of offsets to the scan lines of
			// this frame. this allows us to address a particular scan line
			// without doing a costly multiply.
		tempFrameP->numScanLines = numScanLines;
		tempFrameP->scanLinePtrArray = (unsigned long*)(tempFrameP + 1);

		{	
			tempFrameP->worldRectOffset = 0;
				
			if (depth < 8)
			{
				short		roundOff;
				
				roundOff = 8/depth;
				/*******************************************************
		 		 MacOS using h:v and BeOS using x:y is utterly annoying.
				*******************************************************/
				tempFrameP->worldRectOffset = 
					(globalOffset.x - frameRect->left) - 
					(((globalOffset.x - frameRect->left)/roundOff)*roundOff);
			}
				
			/********************************************************************
			 Again, until we use a BDirectWindow all the scanline stuff doesn't
			 really matter, so I haven't even checked if this would work.
			********************************************************************/
			for (numScanLines = 0; numScanLines < tempFrameP->numScanLines; numScanLines++)
			{
				/*******************************************************
		 		 MacOS using h:v and BeOS using x:y is utterly annoying.
				*******************************************************/
				tempFrameP->scanLinePtrArray[numScanLines] = 
					((numScanLines+(globalOffset.y - frameRect->top) ) * 
					tempFrameP->frameRowBytes) +
					(globalOffset.x - frameRect->left)*4/(32/depth);
			}
		}

		tempFrameP->sharesGWorld = false;

		*newFrameP = tempFrameP;
	}
	else
	{
		err = MemError();
	}

	if (err != noErr)
	{
		if (tempFrameP != NULL)
		{
			DisposePtr((Ptr)tempFrameP);
		}
	}
	
	SWSetStickyIfError( err );
	
	return err;
}


///--------------------------------------------------------------------------------------
//	SWDisposeFrame
///--------------------------------------------------------------------------------------

SW_FUNC Boolean SWDisposeFrame(
	FramePtr oldFrameP)
{
	Boolean		frameDisposed;
	
	
	frameDisposed = false;
	if (oldFrameP != NULL)
	{
			// is this frame still in use by another sprite?
		if (oldFrameP->useCount > 1)
		{
				// one less sprite is using it now!
			oldFrameP->useCount--;
		}
		else	// no more sprites are using this frame
		{	
			frameDisposed = true;
			if (oldFrameP->framePort != NULL)
			{
				if ( !oldFrameP->sharesGWorld )
				{
					
					delete oldFrameP->framePort;
				}
					DisposeGWorld(oldFrameP->framePort);

				oldFrameP->framePort = NULL;
			}
			
			if (oldFrameP->maskRgn != NULL)
			{
				DisposeRgn(oldFrameP->maskRgn);

				oldFrameP->maskRgn = NULL;
			}
			
			if (oldFrameP->maskPort != NULL)
			{

				if ( !oldFrameP->sharesGWorld )
				{
					delete oldFrameP->maskPort;
				}
					DisposeGWorld(oldFrameP->maskPort);

				oldFrameP->maskPort = NULL;
			}

			if (oldFrameP->pixCodeH != NULL)
			{
				DisposeHandle((Handle)oldFrameP->pixCodeH);
				oldFrameP->pixCodeH = NULL;
				oldFrameP->frameBlitterP = NULL;
			}

			DisposePtr((Ptr)oldFrameP);
			oldFrameP = NULL;	// This is stupid, because we're just changing our local copy
		}
	}
	return frameDisposed;
}


///--------------------------------------------------------------------------------------
//	SWDisposeWindowFrame
///--------------------------------------------------------------------------------------

SW_FUNC void SWDisposeWindowFrame(
	FramePtr oldFrameP)
{
	if (oldFrameP != NULL)
	{
		DisposePtr((Ptr)oldFrameP);
		oldFrameP = NULL;	// This is stupid, because we're just changing our local copy
	}
}


///--------------------------------------------------------------------------------------
//	SWInitializeFrame
///--------------------------------------------------------------------------------------

SW_FUNC void SWInitializeFrame(
	FramePtr tempFrameP,
	short depth )
{
		long 			numScanLines;
		
	
		// cache frameRowBytes for use by our blitter routine

	/*******************************************************************
	bytes-per-row stored a little differently for a BBitmap vs.a GWorld.
	*******************************************************************/
	tempFrameP->frameRowBytes = tempFrameP->framePix->BytesPerRow();

		// this calculation generates a mask value that we use to
		// long word align the rectangle when we draw the frame.
		// note that the expression "sizeof(long) * kBitsPerByte" gives us
		// the number of bits in a long.


		// align on long word boundary
	tempFrameP->rightAlignFactor = ((sizeof(long) * kBitsPerByte) / depth) - 1;
	tempFrameP->leftAlignFactor = ~(tempFrameP->rightAlignFactor);

		// the useCount keeps track of the number of sprites that
		// are using this frame. we need to know this so we don't
		// dispose this frame twice when we are disposing the 
		// sprites that use it.
	tempFrameP->useCount = 0;
	
	tempFrameP->worldRectOffset = 0;
	
		// here we set up an array of offsets to the scan lines of
		// this frame. this allows us to address a particular scan line
		// without doing a costly multiply.
	tempFrameP->scanLinePtrArray = (unsigned long*)(tempFrameP + 1);
	
	for (numScanLines = 0; numScanLines < tempFrameP->numScanLines; numScanLines++)
	{
		tempFrameP->scanLinePtrArray[numScanLines] = 
			((tempFrameP->frameRect.top+numScanLines) * tempFrameP->frameRowBytes);
	}
}


///--------------------------------------------------------------------------------------
//	SWSetFrameMaskRgn
///--------------------------------------------------------------------------------------

SW_FUNC void SWSetFrameMaskRgn(
	FramePtr srcFrameP,
	RgnHandle maskRgn)
{
	if (maskRgn != NULL)
	{
		srcFrameP->maskRgn = maskRgn;
	
		/****************************
		Change MacOS h-v to BeOS x-y.
		****************************/
		srcFrameP->offsetPoint.x = maskRgn->Frame().left;
		srcFrameP->offsetPoint.y = maskRgn->Frame().top;
	}
}


///--------------------------------------------------------------------------------------
//	SWSetFrameHotSpot
///--------------------------------------------------------------------------------------

SW_FUNC void SWSetFrameHotSpot(
	FramePtr srcFrameP,
	short hotSpotH,
	short hotSpotV)
{
	srcFrameP->hotSpotH = hotSpotH;
	srcFrameP->hotSpotV = hotSpotV;
}


///--------------------------------------------------------------------------------------
// SWCopyFrame
///--------------------------------------------------------------------------------------

SW_FUNC OSErr SWCopyFrame(
	SpriteWorldPtr destSpriteWorldP,
	FramePtr oldFrameP, 
	FramePtr *newFrameP,
	Boolean copyMasks)
{
	OSErr 			err;  
	GWorldPtr 		saveGWorld;
	GDHandle 		saveGDH;
	GDHandle		theGDH;
//	GWorldFlags		pixelState;
	short			depth;
	FramePtr 		tempFrameP;
	Rect 			frameRect;

	*newFrameP = NULL;
	tempFrameP = NULL;
	depth = destSpriteWorldP->pixelDepth;
	theGDH = destSpriteWorldP->mainSWGDH;

	GetGWorld(&saveGWorld, &saveGDH);
	
		// Get size of new frame
	frameRect = oldFrameP->frameRect;

	frameRect.OffsetBy(frameRect.left, -frameRect.top);
		// Create the new frame
	err = SWCreateFrame(theGDH, &tempFrameP, &frameRect, depth);
	
	/***************************
	Try to copy the frame image.
	***************************/
	if (err == noErr)
	{
		if (tempFrameP->framePix && oldFrameP->framePix)
		{
			if (tempFrameP->framePix->Lock())
			{
				if (oldFrameP->framePix->Lock())
				{
					memcpy(tempFrameP->framePix->Bits(), oldFrameP->framePix->Bits(),
						oldFrameP->framePix->BitsLength());
						
					oldFrameP->framePix->Unlock();
				}
				else
				{
					err = kLockError;
				}
				
				tempFrameP->framePix->Unlock();
			}
			else
			{
				err = kLockError;
			}
		}
		else
		{
			err = MemError();
		}
	}
	
		// copy region mask
	if ((oldFrameP->maskRgn != NULL) && (err == noErr))
	{
		/***************************************
		Modify region copying code for BRegions.
		***************************************/

		tempFrameP->maskRgn = new BRegion(*oldFrameP->maskRgn);
		
		if (tempFrameP->maskRgn == NULL)
			err = MemError();
	}

	/******************************************************
	Try to copy the pixel mask image if it exists.
	******************************************************/
	if (oldFrameP->maskPix != NULL && err == noErr)
	{
		tempFrameP->maskPix = new BBitmap(oldFrameP->frameRect, B_GRAY1, true, false);
			
		if (tempFrameP->maskPix != NULL)
		{
			if (tempFrameP->maskPix->Lock())
			{
				if (oldFrameP->maskPix->Lock())
				{
					memcpy(tempFrameP->maskPix->Bits(), oldFrameP->maskPix->Bits(),
						oldFrameP->maskPix->BitsLength());
						
					oldFrameP->maskPix->Unlock();
				}
				else
				{
					err = kLockError;
				}
				
				tempFrameP->maskPix->Unlock();
			}
			else
			{
				err = kLockError;
			}
		}
		else
		{
			err = MemError();
		}
	}

	if (err == noErr)
	{
		*newFrameP = tempFrameP;
	}
	else
	{
			// an error occurred so dispose of anything we managed to create
		if (tempFrameP != NULL)
		{
			SWDisposeFrame(tempFrameP);
		}
	}

	SetGWorld(saveGWorld, nil);

	SWSetStickyIfError( err );
	return err;
}


///--------------------------------------------------------------------------------------
//	SWUpdateFrameMasks - updates the pixel and region masks of a frame
///--------------------------------------------------------------------------------------

SW_FUNC OSErr SWUpdateFrameMasks(
	SpriteWorldPtr destSpriteWorldP, 
	FramePtr srcFrameP)
{
	GWorldPtr			saveGWorld, newGWld;
	GDHandle			saveGDH;
	Rect				gwldRect;
//	ColorSearchUPP		blackenColSearchUPP;
	Boolean				frameWasUnlocked = false;
	OSErr				err = noErr;
/*	
	GetGWorld( &saveGWorld, &saveGDH );
	
	if (!srcFrameP->isFrameLocked)
	{
		SWLockFrame(srcFrameP);
		frameWasUnlocked = true;
	}
	
	gwldRect = srcFrameP->framePort->portRect;
	
		// Create 1-bit GWorld
	err = NewGWorld(&newGWld, 1, &gwldRect, nil, nil, 0 );
	
	if ( err == noErr )
	{
		blackenColSearchUPP = NewColorSearchProc( blackenColorSearch );
		(void)LockPixels( GetGWorldPixMap(newGWld) );
		SetGWorld( srcFrameP->framePort, nil ); ForeColor( blackColor ); BackColor( whiteColor );
		SetGWorld( newGWld, nil );
		AddSearch( blackenColSearchUPP );
		
			// Copy 8-bit image into 1-bit GWorld while turning it into a mask
		CopyBits( (BitMap*)*GetGWorldPixMap( srcFrameP->framePort ),
			(BitMap*)*GetGWorldPixMap( newGWld ),
			&srcFrameP->framePort->portRect, 
			&gwldRect, 
			srcCopy, 
			nil);
		
		DelSearch( blackenColSearchUPP );
		DisposeRoutineDescriptor( blackenColSearchUPP );

		
			// update the region mask if there is one
		if (srcFrameP->maskRgn != NULL)
		{
			err = BitMapToRegion(srcFrameP->maskRgn, (BitMap*)&newGWld->portPixMap);
		}
		
			// update the pixel mask if there is one
		if (srcFrameP->maskPort != NULL)
		{
			SetGWorld( srcFrameP->maskPort, nil ); ForeColor( blackColor ); BackColor( whiteColor );
			CopyBits( (BitMap*)*GetGWorldPixMap( newGWld ),
				(BitMap*)*GetGWorldPixMap( srcFrameP->maskPort ),
				&gwldRect, 
				&srcFrameP->maskPort->portRect, 
				srcCopy,
				nil);
			
			if (destSpriteWorldP->pixelDepth <= 8)
			{
				SetGWorld(srcFrameP->maskPort, nil);
				InvertRect(&srcFrameP->maskPort->portRect);
			}
		}
		
		UnlockPixels( GetGWorldPixMap(newGWld) );
		DisposeGWorld( newGWld );
	}
	
	if (frameWasUnlocked)
		SWUnlockFrame(srcFrameP);
	
	SetGWorld( saveGWorld, saveGDH );
	
	SWSetStickyIfError( err );
*/
	return err;
}


///--------------------------------------------------------------------------------------
//	SWLockFrame
///--------------------------------------------------------------------------------------

SW_FUNC void SWLockFrame(
	FramePtr srcFrameP)
{
	/*
	We can't lock the frame in the BeOS sense expect for when we actually draw
	becuase otherwise things will lock up.  All locking of the frame has to happen
	within the drawproc.
	*/
}


///--------------------------------------------------------------------------------------
//	SWUnlockFrame
///--------------------------------------------------------------------------------------

SW_FUNC void SWUnlockFrame(
	FramePtr srcFrameP)
{
	/*
	See SWLockFrame
	*/
}


///--------------------------------------------------------------------------------------
//	SWLockWindowFrame
///--------------------------------------------------------------------------------------

SW_FUNC bool SWLockWindowFrame(
	FramePtr windowFrameP)
{
	/*
	We need lock the bitmap of a frame before we draw to it.
	*/
	return windowFrameP->framePort->Window()->Lock();
}


///--------------------------------------------------------------------------------------
//	SWUnlockWindowFrame
///--------------------------------------------------------------------------------------

SW_FUNC void SWUnlockWindowFrame(
	FramePtr windowFrameP)
{
	/*
	See SWLockWindowFrame.
	*/
	windowFrameP->framePort->Window()->Unlock();

}

///--------------------------------------------------------------------------------------
//	SWCreateFrameFromIconResource
///--------------------------------------------------------------------------------------
SW_FUNC OSErr SWCreateFrameFromIconResource(
	FramePtr		*newFrameP,
	BResources		*resFile,
	type_code		iconType,
	int32 			iconID)
{
	OSErr 			err = noErr; 
	size_t			resLength;
	FramePtr		tempFrameP;
	
	*newFrameP = NULL;
	tempFrameP = NULL;

 	if(resFile->HasResource(iconType, iconID))
    {
    	char	*iconData;
  		int32	bytesPerRow;
  		char	*bits;
  		uint8	x, y;
		char	*p;  	
		
		iconData = (char *)resFile->FindResource(iconType, iconID, &resLength);
		
		if (iconData == NULL)
			return -1;
    	
    	switch(iconType)
    	{
    		case B_SMALL_ICON:
 
 				SWCreateFrame( 0, &tempFrameP, &BRect(0, 0, 15, 15), 32);

				bits = (char *)tempFrameP->framePix->Bits();
				
				bytesPerRow = tempFrameP->framePix->BytesPerRow();

				for (y = 0; y < 16; y++)
				{
					p = (char *)(bits + (y * bytesPerRow));
					
					for (x = 0; x < 16; x++)
						*p++ = iconData[(16 * y) + x];
				}
   			break;
    		
    		case B_BIG_ICON:
				
				SWCreateFrame( 0, &tempFrameP, &BRect(0, 0, 31, 31), 32 );

				bits = (char *)tempFrameP->framePix->Bits();
				
				bytesPerRow = tempFrameP->framePix->BytesPerRow();

				for (y = 0; y < 32; y++)
				{
					p = (char *)(bits + (y * bytesPerRow));
					
					for (x = 0; x < 32; x++)
						*p++ = iconData[(32 * y) + x];
				}
			break;
		}
		
		tempFrameP->framePix->RemoveChild(tempFrameP->framePort);
		delete tempFrameP->framePort;
		tempFrameP->framePort = NULL;
		
		*newFrameP = tempFrameP;
	}

	SWSetStickyIfError( err );
	
	return err;
}

///--------------------------------------------------------------------------------------
//	SWCreateFrameFromBitmap
///--------------------------------------------------------------------------------------
SW_FUNC OSErr SWCreateFrameFromBitmap(
	SpriteWorldPtr	destSpriteWorld,
	FramePtr 		*newFrameP, 
	BBitmap			*tileBitmap,
	BBitmap			*maskBitmap,
	MaskType 		maskType)
{
	OSErr 			err = noErr; 
	FramePtr		tempFrameP;
	Rect			frameRect;
	RgnHandle		maskRgn = NULL;
	
	*newFrameP = NULL;
	tempFrameP = NULL;
	
	frameRect = tileBitmap->Bounds();

 	err = SWCreateFrame( 0, &tempFrameP, &frameRect, tileBitmap->ColorSpace());

	if (err == noErr)
	{
		if (tempFrameP->framePix != NULL)
			delete tempFrameP->framePix;
		
		tempFrameP->framePix = tileBitmap;		
	}

	if (((maskType & kPixelMask) != 0) && (err == noErr)) 
	{
		tempFrameP->maskPix = maskBitmap;
	}
	
	if (((maskType & kRegionMask) != 0) && (err == noErr))
	{
		err = SWCreateRegionFromBitmapAndRectPartial(&maskRgn, maskBitmap);
				
		if (err == noErr)
		{
			SWSetFrameMaskRgn(tempFrameP, maskRgn);
		}

	}

	if (err == noErr)
	{
		*newFrameP = tempFrameP;
	}

	SWSetStickyIfError( err );
	
	return err;
}

///--------------------------------------------------------------------------------------
//	SWCreateFrameFromBitmapAndRectStart
///--------------------------------------------------------------------------------------

SW_FUNC OSErr SWCreateFrameFromBitmapAndRectStart(
	BBitmap		*tempMaskGWorld,
	short		maxWidth,
	short		maxHeight,
	MaskType	maskType)
{
	OSErr		err = noErr;
	
//	*tempMaskGWorld = NULL;
	if ((maskType & kRegionMask) != 0)
	{
//		err = SWCreateRegionFromGWorldAndRectStart( tempMaskGWorld, maxWidth, maxHeight );
	}
	return err;
}

///--------------------------------------------------------------------------------------
//	SWCreateFrameFromBitmapAndRectFinish
///--------------------------------------------------------------------------------------

SW_FUNC void SWCreateFrameFromBitmapAndRectFinish(
	BBitmap *tempMaskGWorld)
{
//	SWCreateRegionFromGWorldAndRectFinish( tempMaskGWorld );
}

///--------------------------------------------------------------------------------------
//	SWCreateFrameFromBitmapAndRectPartial
///--------------------------------------------------------------------------------------

SW_FUNC OSErr SWCreateFrameFromBitmapAndRectPartial(
	FramePtr*	newFrameP,
	BBitmap		*pictGWorld,
	BBitmap		*maskGWorld,
	BBitmap		*tempMaskGWorld,
	Rect*		frameRect,
	MaskType	maskType)
{						
	OSErr 			err = noErr;
	short			depth;
	long 			numScanLines;
	FramePtr 		tempFrameP;
	RgnHandle 		maskRgn = NULL;


	tempFrameP = NULL;
	*newFrameP = NULL;

	numScanLines = frameRect->bottom - frameRect->top;
//---
//	depth = (**(GetGWorldPixMap( pictGWorld ))).pixelSize;
	depth = DepthFromColorSpace(pictGWorld->ColorSpace());
	
	tempFrameP = (FramePtr)NewPtrClear((Size)sizeof(FrameRec) + 
		(sizeof(unsigned long) * (numScanLines+1)));
	
	if (tempFrameP != NULL)
	{
//		tempFrameP->framePort = pictGWorld;
		tempFrameP->framePix = pictGWorld;
		
			// make a pixel mask
		if ((maskType & kPixelMask) != 0) 
		{
//			tempFrameP->maskPort = maskGWorld;
			tempFrameP->maskPix = maskGWorld;
		}
		tempFrameP->frameRect = *frameRect;
		tempFrameP->sharesGWorld = true;
		tempFrameP->numScanLines = numScanLines;

		SWInitializeFrame( tempFrameP, depth );
	}
	else
	{
		err = MemError();
	}

	if (err == noErr)
	{
			// make a region mask
		if (((maskType & kRegionMask) != 0) && (err == noErr))
		{
			err = SWCreateRegionFromBitmapAndRectPartial(&maskRgn, maskGWorld);
				
			if (err == noErr)
			{
				SWSetFrameMaskRgn(tempFrameP, maskRgn);
			}

		}
	
		if (err == noErr)
		{
			*newFrameP = tempFrameP;
		}
		else
		{
				// an error occurred so dispose of anything we managed to create
				
			if (tempFrameP != NULL)
			{
				SWDisposeFrame(tempFrameP);
			}
		}
	}
	
	SWSetStickyIfError( err );

	return err;
}
