///--------------------------------------------------------------------------------------
//	Tiling.c
//
//	By Vern Jensen. SWLoadTile routines by Karl Bunker.
//	
//	Created: 11/15/95
//
//	Description:	Routines to use tiling with SpriteWorld
///--------------------------------------------------------------------------------------

#ifndef __TILING__
#include "Tiling.h"
#endif

#ifndef __SPRITEWORLDUTILS__
#include "SpriteWorldUtils.h"
#endif

#ifndef __SPRITEWORLD__
#include "SpriteWorld.h"
#endif

#ifndef __BLITPIXIE__
#include "BlitPixie.h"
#endif


	// Placeholder until we get a real ASSERT function
#define	SW_ASSERT(dummyStuff)	



///--------------------------------------------------------------------------------------
//	SWInitTiling
///--------------------------------------------------------------------------------------

SW_FUNC OSErr SWInitTiling(
	SpriteWorldPtr	spriteWorldP,
	short 			tileHeight,
	short			tileWidth,
	short			maxNumTiles)
{
	OSErr 	err = noErr;
	Size	arraySize;
	short	tileIndex, row, col, numOffscreenCols, numOffscreenRows;
	
	
	if (spriteWorldP->tilingIsInitialized)
	{
		err = kTilingAlreadyInitialized;
	}
	
	
	if (err == noErr)
	{
		spriteWorldP->tilingIsOn = true;
		spriteWorldP->tilingIsInitialized = true;
		spriteWorldP->maxNumTiles = maxNumTiles;
		spriteWorldP->tileHeight = tileHeight;
		spriteWorldP->tileWidth = tileWidth;
		spriteWorldP->tileMapStructP = NULL;
////////////////////////////////////////////////////////////////////////////////
//
// A 32x32 tile uses at rect with coordinates (0, 0, 31, 31) with Be, so we need
// to add one to the size of the rect before dividing by the tile sizes to get
// the correct number of rows.  Also, Be coordinates are float so we cast for
// completeness.
////////////////////////////////////////////////////////////////////////////////
//		numOffscreenRows = spriteWorldP->backRect.bottom / tileHeight;
//		numOffscreenCols = spriteWorldP->backRect.right / tileWidth;
		numOffscreenRows = ((int32)spriteWorldP->backRect.bottom + 1 ) / tileHeight;
		numOffscreenCols = ((int32)spriteWorldP->backRect.right + 1 ) / tileWidth;
////////////////////////////////////////////////////////////////////////////////
		spriteWorldP->numOffscreenTileMapRows = numOffscreenRows;
		spriteWorldP->numOffscreenTileMapCols = numOffscreenCols;
	}
	
	
	if (err == noErr)
	{
			// Allocate memory for offscreenTileMap ("Tiling Cache")
		arraySize = (Size)numOffscreenRows * numOffscreenCols * sizeof(short);
		spriteWorldP->offscreenTileMap = (short *)NewPtr(arraySize);
//		err = MemError();
		
		if (err == noErr)
		{
				// Set all elements to -1 (indicating that each tile needs to be drawn)
			for (row = 0; row < numOffscreenRows; row++)
				for (col = 0; col < numOffscreenCols; col++)
					spriteWorldP->offscreenTileMap[(long)row * numOffscreenCols + col] = -1;
		}
	}
	
	
	if (err == noErr)
	{
			// Allocate memory for changedTiles array of rects
		spriteWorldP->changedTilesArraySize = (numOffscreenRows+1) * (numOffscreenCols+1);
		arraySize = (Size)(numOffscreenRows+1) * (numOffscreenCols+1) * sizeof(Rect);
		spriteWorldP->changedTiles = (Rect *)NewPtr(arraySize);
//		err = MemError();
	}
	
	
	if (err == noErr)
	{
			// Allocate memory for tileFrameArray
		arraySize = (Size)maxNumTiles * sizeof(FramePtr);
		spriteWorldP->tileFrameArray = (FramePtr *)NewPtrClear(arraySize);
//		err = MemError();
	}
	
	
	if (err == noErr)
	{
			// Allocate memory for curTileImage array
		arraySize = (Size)maxNumTiles * sizeof(short);
		spriteWorldP->curTileImage = (short *)NewPtr(arraySize);
//		err = MemError();
		
		if (err == noErr)
		{
				// Set up values in curTileImage array
			for (tileIndex = 0; tileIndex < maxNumTiles; tileIndex++)
				spriteWorldP->curTileImage[tileIndex] = tileIndex;
		}
	}
	
	
	SWSetStickyIfError(err);
	return err;
}


///--------------------------------------------------------------------------------------
//	SWExitTiling
///--------------------------------------------------------------------------------------

SW_FUNC void SWExitTiling(
	SpriteWorldPtr	spriteWorldP)
{
	short	tileIndex;
	
		// Was tiling ever initialized?
	if (spriteWorldP->tilingIsInitialized)
	{
		tileIndex = spriteWorldP->maxNumTiles;
		while (tileIndex--)
		{
			SWDisposeTile(spriteWorldP, tileIndex);
		}
		
		DisposePtr((Ptr)spriteWorldP->offscreenTileMap);
		DisposePtr((Ptr)spriteWorldP->tileFrameArray);
		DisposePtr((Ptr)spriteWorldP->changedTiles);
		DisposePtr((Ptr)spriteWorldP->curTileImage);
		
		spriteWorldP->tilingIsInitialized = false;
		spriteWorldP->tilingIsOn = false;
	}
}


///--------------------------------------------------------------------------------------
//	SWCreateTileMap
///--------------------------------------------------------------------------------------

SW_FUNC OSErr SWCreateTileMap(
	TileMapStructPtr	*tileMapStructPP,
	short 				numTileMapRows,
	short				numTileMapCols)
{
	TileMapStructPtr	tileMapStructP;
	Size				arraySize;
	OSErr				err = noErr;
	
	
		// Allocate memory for the TileMapStruct
	tileMapStructP = (TileMapStructPtr)NewPtrClear(sizeof(TileMapStruct));
//	err = MemError();
	
	if (err == noErr)
	{
		tileMapStructP->numRows = numTileMapRows;
		tileMapStructP->numCols = numTileMapCols;
		
			// Allocate the array of pointers that point to the data of the TileMap
		arraySize = (Size)numTileMapRows * sizeof(short*);
////////////////////////////////////////////////////////////////////////////////
//
// Rather than mimic NewHandle(), I'll just replace with a malloc for now.
//
////////////////////////////////////////////////////////////////////////////////
//		tileMapStructP->arrayOfPointersH = NewHandle(arraySize);
		tileMapStructP->arrayOfPointersH = (short **)malloc(arraySize);
////////////////////////////////////////////////////////////////////////////////
//		err = MemError();
	}
	
	if (err == noErr)
	{
			// Allocate memory for the actual data of the TileMap
		arraySize = ((Size)numTileMapRows * (numTileMapCols + 2)) * sizeof(short);
////////////////////////////////////////////////////////////////////////////////
//
// Rather than mimic NewHandle(), I'll just replace with a malloc for now.
//
////////////////////////////////////////////////////////////////////////////////
//		tileMapStructP->tileMapDataH = NewHandleClear(arraySize);
		tileMapStructP->tileMapDataH = (short *)malloc(arraySize);
////////////////////////////////////////////////////////////////////////////////
//		err = MemError();
	}
	
	if (err == noErr)
	{
		SWLockTileMap(tileMapStructP);
	}
	
	
	if (err != noErr)
	{
			// Dispose what we created
		if (tileMapStructP != NULL)
		{
			DisposeHandle((Handle)tileMapStructP->arrayOfPointersH);
			DisposeHandle((Handle)tileMapStructP->tileMapDataH);
			DisposePtr((Ptr)tileMapStructP);
			tileMapStructP = NULL;
		}
	}

		// Return a pointer to the TileMapStruct in the tileMapStructPP variable
	*tileMapStructPP = tileMapStructP;
	
	SWSetStickyIfError(err);
	return err;
}


///--------------------------------------------------------------------------------------
//	SWDisposeTileMap
///--------------------------------------------------------------------------------------

SW_FUNC void SWDisposeTileMap(
	TileMapStructPtr	*tileMapStructPP)
{
	TileMapStructPtr	myTileMapStructP = *tileMapStructPP;
	
	if (myTileMapStructP != NULL)
	{
		DisposeHandle((Handle)myTileMapStructP->arrayOfPointersH);
		DisposeHandle((Handle)myTileMapStructP->tileMapDataH);
		DisposePtr((Ptr)myTileMapStructP);
		*tileMapStructPP = NULL;
	}
}


///--------------------------------------------------------------------------------------
//	SWLockTileMap
///--------------------------------------------------------------------------------------

SW_FUNC void SWLockTileMap(
	TileMapStructPtr	tileMapStructP)
{
	short		**arrayOfPointers;
	short		*tileMapDataP;
	short		row;

////////////////////////////////////////////////////////////////////////////////

// Easier to comment out than reimplement

////////////////////////////////////////////////////////////////////////////////	
//	HLockHi(tileMapStructP->arrayOfPointersH);
//	HLockHi(tileMapStructP->tileMapDataH);
////////////////////////////////////////////////////////////////////////////////	

////////////////////////////////////////////////////////////////////////////////
//
// We've replaced the Handle with a pointer.
//
////////////////////////////////////////////////////////////////////////////////	
//	tileMapDataP = (short*)*tileMapStructP->tileMapDataH;
	tileMapDataP = (short*)tileMapStructP->tileMapDataH;
////////////////////////////////////////////////////////////////////////////////	

////////////////////////////////////////////////////////////////////////////////
//
// We've replaced the Handle with a pointer.
//	
////////////////////////////////////////////////////////////////////////////////	
//	arrayOfPointers = (short**)*tileMapStructP->arrayOfPointersH;
	arrayOfPointers = (short**)tileMapStructP->arrayOfPointersH;
////////////////////////////////////////////////////////////////////////////////	

		// Point each element of the array of pointers to each row of the data
	for (row = 0; row < tileMapStructP->numRows; row++)
	{
		arrayOfPointers[row] = &tileMapDataP[(long)row * (tileMapStructP->numCols + 2)];
	}
	
	tileMapStructP->tileMap = arrayOfPointers;
	tileMapStructP->isLocked = true;
}


///--------------------------------------------------------------------------------------
//	SWUnlockTileMap
///--------------------------------------------------------------------------------------

SW_FUNC void SWUnlockTileMap(
	TileMapStructPtr	tileMapStructP)
{
////////////////////////////////////////////////////////////////////////////////	
//
// Easier to comment out than reimplement for now.
//
////////////////////////////////////////////////////////////////////////////////	
//	HUnlock(tileMapStructP->arrayOfPointersH);
//	HUnlock(tileMapStructP->tileMapDataH);
////////////////////////////////////////////////////////////////////////////////	
	tileMapStructP->tileMap = NULL;
	tileMapStructP->isLocked = false;
}


///--------------------------------------------------------------------------------------
//	SWInstallTileMap
///--------------------------------------------------------------------------------------

SW_FUNC void SWInstallTileMap(
	SpriteWorldPtr		spriteWorldP,
	TileMapStructPtr	tileMapStructP)
{
	spriteWorldP->tileMapStructP = tileMapStructP;
}

////////////////////////////////////////////////////////////////////////////////
//	
// SWLoadTileMap: Not implemented on for Be yet.
//
////////////////////////////////////////////////////////////////////////////////
	
///--------------------------------------------------------------------------------------
//	SWLoadTileMap
///--------------------------------------------------------------------------------------

SW_FUNC OSErr SWLoadTileMap(
	TileMapStructPtr	*tileMapStructPP,
	short 				resourceID)
{
	TileMapStructPtr	tileMapStructP;
	short				*tileMapData;
	Size				arraySize;
	OSErr				err = noErr;
	
	
		// Allocate memory for the TileMapStruct
	tileMapStructP = (TileMapStructPtr)NewPtrClear(sizeof(TileMapStruct));
	err = MemError();
	
	if (err == noErr)
	{	
			// Load the resource
//		tileMapStructP->tileMapDataH = Get1Resource('TMAP', resourceID);
//---		
//		if (tileMapStructP->tileMapDataH == NULL)
//			err = ResError() ? ResError() : resNotFound;
//---
	}
	
	if (err == noErr)
	{
		DetachResource((Handle)tileMapStructP->tileMapDataH);
		
		tileMapData = (short*)*tileMapStructP->tileMapDataH;
		tileMapStructP->numRows = tileMapData[0];
		tileMapStructP->numCols = tileMapData[1];
		
			// Allocate the array of pointers that point to the data of the TileMap
		arraySize = (Size)tileMapStructP->numRows * sizeof(short*);
//		tileMapStructP->arrayOfPointersH = NewHandle(arraySize);
		err = MemError();
	}
	
	if (err == noErr)
	{
		SWLockTileMap(tileMapStructP);
	}
	
	
	if (err != noErr)
	{
			// Dispose what we created
		if (tileMapStructP != NULL)
		{
			DisposeHandle((Handle)tileMapStructP->arrayOfPointersH);
			DisposeHandle((Handle)tileMapStructP->tileMapDataH);
			DisposePtr((Ptr)tileMapStructP);
			tileMapStructP = NULL;
		}
	}
	
		// Return a pointer to the TileMapStruct in the tileMapStructPP variable
	*tileMapStructPP = tileMapStructP;
	
	SWSetStickyIfError(err);
	return err;
}

////////////////////////////////////////////////////////////////////////////////
//	
// SWSaveTileMap: Not implemented on for Be yet.
//
////////////////////////////////////////////////////////////////////////////////	

///--------------------------------------------------------------------------------------
//	SWSaveTileMap
///--------------------------------------------------------------------------------------

SW_FUNC OSErr SWSaveTileMap(
	TileMapStructPtr	tileMapStructP,
	short 				destResID)
{
	short		tempResID;
	short		*tileMapData;
	Handle		tempTileMapH;
	OSErr		err = noErr;
	
		// Make sure there is a tileMap to save
	if (tileMapStructP == NULL)
		err = kNullTileMapErr;

	if (err == noErr)
	{
			// Save numRows & numCols in the first two elements of the tileMapData
		tileMapData = (short*)*tileMapStructP->tileMapDataH;
		tileMapData[0] = tileMapStructP->numRows;
		tileMapData[1] = tileMapStructP->numCols;
		
			// Add the TMAP resource, assigning it a temporary, unused ID
//		tempResID = Unique1ID('TMAP');
//		AddResource(tileMapStructP->tileMapDataH, 'TMAP', tempResID, "\p");
//		err = ResError();
	}
	
	if (err == noErr)
	{
		do		// Remove any old TMAP resources with destResID
		{
//			SetResLoad(false);
			tempTileMapH = Get1Resource('TMAP', destResID);
//			SetResLoad(true);
			
			if (tempTileMapH != NULL)
			{
//				RemoveResource(tempTileMapH);
				
					// This could happen if the resource is protected
//				err = ResError();
//				SW_ASSERT(ResError() == noErr);
			}
//			else if (ResError() != resNotFound && ResError() != noErr)
			{
				// This could happen if Get1Resource didn't have a free master
				// pointer and there isn't enough memory to allocate a new block
//				err = ResError();
			}
		} while (tempTileMapH != NULL);
		
		
			// Change the resource ID to the proper one
//		SetResInfo(tileMapStructP->tileMapDataH, destResID, "\p");
		
			// Update file, writing new resource and deleting old one in one step.
//		UpdateResFile( CurResFile() );
//		DetachResource(tileMapStructP->tileMapDataH);
	}
				
	
	SWSetStickyIfError(err);
	return err;
}


///--------------------------------------------------------------------------------------
//	SWResizeTileMap
///--------------------------------------------------------------------------------------

SW_FUNC OSErr SWResizeTileMap(
	TileMapStructPtr	oldTileMapStructP,
	short 				numNewTileMapRows,
	short				numNewTileMapCols)
{
	TileMapStruct		tempTileMapStruct;
	TileMapStructPtr	newTileMapStructP;
	short				row, numRowsToCopy, numColsToCopy;
	Size				arraySize;
	OSErr				err = noErr;
		
	if (oldTileMapStructP == NULL)
		err = kNullTileMapErr;
	
	
	if (err == noErr)
	{
			// Don't do anything if the TileMap is already the requested size.
		if (oldTileMapStructP->numRows == numNewTileMapRows && 
			oldTileMapStructP->numCols == numNewTileMapCols)
		{
			return noErr;
		}
		
			// Create the new TileMap
		err = SWCreateTileMap(&newTileMapStructP, numNewTileMapRows, numNewTileMapCols);
	}
	
	if (err == noErr)
	{
		if (oldTileMapStructP->isLocked == false)
			SWLockTileMap(oldTileMapStructP);
		
		numRowsToCopy = SW_MIN(oldTileMapStructP->numRows, numNewTileMapRows);
		numColsToCopy = SW_MIN(oldTileMapStructP->numCols, numNewTileMapCols);
		
			// Copy the data from the old TileMap to the new TileMap
		arraySize = (Size)numColsToCopy * sizeof(short);
		for (row = 0; row < numRowsToCopy; row++)
		{
			BlockMoveData(&oldTileMapStructP->tileMap[row][0],
				&newTileMapStructP->tileMap[row][0], arraySize);
		}
		
			// Swap contents of the new and old TileMapStructs
		tempTileMapStruct = *oldTileMapStructP;
		*oldTileMapStructP = *newTileMapStructP;
		*newTileMapStructP = tempTileMapStruct;
		
			// Dispose the newTileMapStruct, which now contains the old TileMap
		SWDisposeTileMap(&newTileMapStructP);
	}
	
	
	SWSetStickyIfError(err);
	return err;
}


///--------------------------------------------------------------------------------------
//	SWLockTiles
///--------------------------------------------------------------------------------------

SW_FUNC void SWLockTiles(
	SpriteWorldPtr	spriteWorldP)
{
	short		tileIndex;
	
	if (spriteWorldP->tilingIsInitialized)
	{
		for (tileIndex = 0; tileIndex < spriteWorldP->maxNumTiles; tileIndex++)
		{
			if (spriteWorldP->tileFrameArray[tileIndex] != NULL)
				SWLockFrame(spriteWorldP->tileFrameArray[tileIndex]);
		}
	}
}


///--------------------------------------------------------------------------------------
//	SWUnlockTiles
///--------------------------------------------------------------------------------------

SW_FUNC void SWUnlockTiles(
	SpriteWorldPtr	spriteWorldP)
{
	short		tileIndex;
	
	if (spriteWorldP->tilingIsInitialized)
	{
		for (tileIndex = 0; tileIndex < spriteWorldP->maxNumTiles; tileIndex++)
		{
			if (spriteWorldP->tileFrameArray[tileIndex] != NULL)
				SWUnlockFrame(spriteWorldP->tileFrameArray[tileIndex]);
		}
	}
}


///--------------------------------------------------------------------------------------
//	SWSetTilingOn
///--------------------------------------------------------------------------------------

SW_FUNC void SWSetTilingOn(
	SpriteWorldPtr	spriteWorldP,
	Boolean			tilingIsOn)
{
	spriteWorldP->tilingIsOn = tilingIsOn;
}


///--------------------------------------------------------------------------------------
//	SWSetSpriteUnderTiles
///--------------------------------------------------------------------------------------

SW_FUNC void SWSetSpriteUnderTiles(
	SpritePtr	srcSpriteP,
	Boolean		isUnder)
{
	srcSpriteP->isUnderTiles = isUnder;
	
		// Redraw sprite over/under tiles
	srcSpriteP->needsToBeDrawn = true;
}


///--------------------------------------------------------------------------------------
//	SWSetTileMaskDrawProc
///--------------------------------------------------------------------------------------

SW_FUNC OSErr SWSetTileMaskDrawProc(
	SpriteWorldPtr	spriteWorldP,
	DrawProcPtr		drawProc)
{
	OSErr	err = noErr;
////////////////////////////////////////////////////////////////////////////////
//
// These drawprocs are not currently implemented for Be.
//
////////////////////////////////////////////////////////////////////////////////	
/*	if (spriteWorldP->pixelDepth != 8)
	{
		if (drawProc == BlitPixie8BitMaskDrawProc || 
			drawProc == BP8BitInterlacedMaskDrawProc ||
			drawProc == BlitPixie8BitPartialMaskDrawProc || 
			drawProc == BP8BitInterlacedPartialMaskDrawProc)
		{
			err = kWrongDepthErr;
		}
	}
*/
	if ( err == noErr )
		spriteWorldP->tileMaskDrawProc = drawProc;
	
	SWSetStickyIfError( err );
	return err;
}

////////////////////////////////////////////////////////////////////////////////
//	
// SWLoadTileFromCicnResource: Not implemented on for Be yet.
//
////////////////////////////////////////////////////////////////////////////////	

///--------------------------------------------------------------------------------------
//	SWLoadTileFromCicnResource
///--------------------------------------------------------------------------------------

SW_FUNC OSErr SWLoadTileFromCicnResource(
	SpriteWorldPtr	spriteWorldP, 
	short			tileID,
	short			cicnID, 
	MaskType		maskType)
{
	OSErr 			err;
	GWorldPtr 		saveGWorld;
	GDHandle 		saveGDH;
	FramePtr 		newFrameP;
	CIconHandle 	cIconH;
	Rect			frameRect;
	Boolean			maskIsSolid = false;
	
	
	err = noErr;
	newFrameP = NULL;

	if ( !spriteWorldP->tilingIsInitialized )
	{
		err = kTilingNotInitialized;
	}
	else if ( tileID < 0 || tileID >= spriteWorldP->maxNumTiles)
	{
		err = kOutOfRangeErr;
	}
	
	if (maskType == kSolidMask)
	{
		maskIsSolid = true;
		maskType = kNoMask;
	}
	
	if ( err == noErr )
	{	
		err = SWCreateFrameFromCicnResource(spriteWorldP, &newFrameP, cicnID, maskType);
		
		if ( err == noErr )
		{	
			newFrameP->tileMaskIsSolid = maskIsSolid;
			
				// "Fix" the size if the CICN is larger than the tiles are
			if (newFrameP->frameRect.right > spriteWorldP->tileWidth)
				newFrameP->frameRect.right = spriteWorldP->tileWidth;
			if (newFrameP->frameRect.bottom > spriteWorldP->tileHeight)
				newFrameP->frameRect.bottom = spriteWorldP->tileHeight;
			
/*			cIconH = GetCIcon( cicnID );

			if (cIconH != NULL)
			{
				GetGWorld(&saveGWorld, &saveGDH);
				
				HLock((Handle)cIconH);
				frameRect = ((**cIconH).iconPMap.bounds);
				(**cIconH).iconPMap.baseAddr = *(**cIconH).iconData;
				(**cIconH).iconBMap.baseAddr = (Ptr)&((**cIconH).iconMaskData ) +
					((**cIconH).iconMask.rowBytes * (frameRect.bottom-frameRect.top));	
								
				(void)LockPixels( GetGWorldPixMap( newFrameP->framePort ) );
				SetGWorld( newFrameP->framePort, NULL );
				if ( spriteWorldP->pixelDepth > 1 )
				{
					CopyBits( (BitMap*)(&((**cIconH).iconPMap)),
						(BitMap*)*GetGWorldPixMap( newFrameP->framePort ), 
						&frameRect, &newFrameP->frameRect, srcCopy, nil);
				}
				else
				{
					CopyBits( (BitMap*)(&((**cIconH).iconBMap)),
						(BitMap*)*GetGWorldPixMap( newFrameP->framePort ), 
						&frameRect,
						&newFrameP->frameRect, srcCopy, nil);
				}
				UnlockPixels( GetGWorldPixMap( newFrameP->framePort ) );
				DisposeCIcon( cIconH );
				SetGWorld( saveGWorld, saveGDH );
			}
			else
			{
				err = MemError();
			}
*/
		}
		
		if ( err == noErr )
		{
				// Are we replacing an old tile?
			if (spriteWorldP->tileFrameArray[tileID] != NULL)
			{
				SWDisposeTile( spriteWorldP, tileID );
			}
			spriteWorldP->tileFrameArray[tileID] = newFrameP;
		}
	}
	
	if ( err != noErr && newFrameP != NULL )
	{
		SWDisposeFrame(newFrameP);
	}
	
	SWSetStickyIfError(err);
	return err;
}

////////////////////////////////////////////////////////////////////////////////
//	
// SWLoadTilesFromPictResource: Not implemented on for Be yet.
//
////////////////////////////////////////////////////////////////////////////////	

///--------------------------------------------------------------------------------------
//	SWLoadTilesFromPictResource
///--------------------------------------------------------------------------------------

SW_FUNC OSErr SWLoadTilesFromPictResource(
	SpriteWorldPtr	spriteWorldP, 
	short			startTileID,
	short			endTileID, 
	short			pictResID, 
	short			maskResID,
	MaskType		maskType,
	short			horizBorderWidth,
	short			vertBorderHeight)
{
	OSErr 			err;
	short			tileIndex;
	FramePtr 		newFrameP;
	GWorldPtr		tempPictGWorldP,
					tempMaskGWorldP,
					tempTempMaskGWorldP,
					currentGWorld;
	GDHandle		currentGDH;
	Rect			currentTileRect;
	short			horizOffset,
					vertOffset;
	Boolean			allTilesDone,
					maskIsSolid = false;
	
	
	err = noErr;
	newFrameP = NULL;
	tempPictGWorldP = NULL;
	tempMaskGWorldP = NULL;
	tempTempMaskGWorldP = NULL;
	
	if ( !spriteWorldP->tilingIsInitialized )
	{
		err = kTilingNotInitialized;
	}
	else if ( startTileID < 0 || endTileID >= spriteWorldP->maxNumTiles)
	{
		err = kOutOfRangeErr;
	}
	
	if (maskType == kSolidMask)
	{
		maskIsSolid = true;
		maskType = kNoMask;
	}
	
	if ( err == noErr )
	{	
		err = SWCreateGWorldFromPictResource( spriteWorldP, &tempPictGWorldP, pictResID );
		
		if ( err == noErr && maskType != kNoMask )
			err = SWCreateGWorldFromPictResource( spriteWorldP, &tempMaskGWorldP, maskResID );
		
		if (err == noErr)
		{
			if ( pictResID == maskResID && tempMaskGWorldP != NULL )
			{
				err = SWBlackenGWorld( tempMaskGWorldP );
			}
		}
		
		if ( err == noErr )
		{
			err = SWCreateFrameFromGWorldAndRectStart( &tempTempMaskGWorldP, 
					spriteWorldP->tileWidth, spriteWorldP->tileHeight, maskType );
		}
	
		if ( err == noErr )
		{
//			SetRect( &currentTileRect, 0,  0, 
//				spriteWorldP->tileWidth, spriteWorldP->tileHeight );
			tileIndex = startTileID;
			horizOffset = spriteWorldP->tileWidth + horizBorderWidth;
			vertOffset = spriteWorldP->tileHeight + vertBorderHeight;
			
			allTilesDone = false;
			while( !allTilesDone && err == noErr )
			{
				err = SWCreateFrameFromGWorldAndRectPartial( &newFrameP, tempPictGWorldP, 
						tempMaskGWorldP, tempTempMaskGWorldP, &currentTileRect, maskType);
				
				if ( newFrameP != NULL && err == noErr )
				{
					newFrameP->tileMaskIsSolid = maskIsSolid;
					
						// Are we replacing an old tile?
					if (spriteWorldP->tileFrameArray[tileIndex] != NULL)
					{
						SWDisposeTile( spriteWorldP, tileIndex );
					}
					spriteWorldP->tileFrameArray[tileIndex] = newFrameP;
					
					tileIndex++;
					if (tileIndex > endTileID )
					{
						allTilesDone = true;
					}
					else
					{
						if (tileIndex >= spriteWorldP->maxNumTiles)
						{
							err = kOutOfRangeErr;
						}
						currentTileRect.left += horizOffset;
						currentTileRect.right += horizOffset;
//						if ( currentTileRect.right > tempPictGWorldP->portRect.right )
						{
							currentTileRect.left = 0;
							currentTileRect.right = spriteWorldP->tileWidth;
							currentTileRect.top += vertOffset;
							currentTileRect.bottom += vertOffset;
//							if ( currentTileRect.bottom > tempPictGWorldP->portRect.bottom )
							{
								err = kOutOfRangeErr;
							}
						}
					}
				}
			}
			
/*			if (err == noErr)
			{
					// make a pixel mask
				if ((maskType & kPixelMask) != 0)
				{
					if (spriteWorldP->pixelDepth <= 8)
					{
						GetGWorld( &currentGWorld, &currentGDH );
						(void)LockPixels( GetGWorldPixMap(tempMaskGWorldP) );
						SetGWorld(tempMaskGWorldP, nil);
						InvertRect(&tempMaskGWorldP->portRect);
						SetGWorld( currentGWorld, currentGDH );
						UnlockPixels( GetGWorldPixMap(tempMaskGWorldP) );
					}
				}
					// If no pixel Mask wanted, dispose
					// of the GWorld we used to make region
				else
				{
					DisposeGWorld( tempMaskGWorldP );
				}
			}*/
		}
		SWCreateFrameFromGWorldAndRectFinish( tempTempMaskGWorldP );
	}
	
	SWSetStickyIfError(err);
	return err;
}


///--------------------------------------------------------------------------------------
//	SWDisposeTile
///--------------------------------------------------------------------------------------

SW_FUNC void SWDisposeTile(
	SpriteWorldPtr	spriteWorldP,
	short			tileID)
{
	short		tileIndex;
	Boolean		gWorldStillInUse;
	
	
		// see if any other tile is using this tile's GWorld
	gWorldStillInUse = false;
	tileIndex = spriteWorldP->maxNumTiles;
	while ( tileIndex-- )
	{
		if ( tileIndex != tileID )
		{
			if ( spriteWorldP->tileFrameArray[tileIndex] != NULL &&
				((spriteWorldP->tileFrameArray[tileIndex])->framePort == 
				(spriteWorldP->tileFrameArray[tileID])->framePort) )
			{
				gWorldStillInUse = true;
			}
		}
	}
		// set flag that tells SWDisposeFrame whether to dispose of GWorld
	(spriteWorldP->tileFrameArray[tileID])->sharesGWorld = gWorldStillInUse;
	
	(void)SWDisposeFrame( spriteWorldP->tileFrameArray[tileID] );
}


///--------------------------------------------------------------------------------------
//	SWDrawTilesInBackground
///--------------------------------------------------------------------------------------

SW_FUNC OSErr SWDrawTilesInBackground(
	SpriteWorldPtr	spriteWorldP)
{
	GWorldPtr		holdGWorld;
	GDHandle		holdGDH;
	OSErr			err = noErr;
	
	GetGWorld( &holdGWorld, &holdGDH );
	
	if (spriteWorldP->tileMapStructP == NULL)
	{
		err = kNullTileMapErr;
	}
	else if ( !spriteWorldP->tilingIsInitialized )
	{
		err = kTilingNotInitialized;
	}
	else
	{
		SWDrawTilesInRect(spriteWorldP, &spriteWorldP->visScrollRect, true);
	}
	
	SetGWorld( holdGWorld, holdGDH );
	
	SWSetStickyIfError(err);
	return err;
}


///--------------------------------------------------------------------------------------
//	SWResetTilingCache
///--------------------------------------------------------------------------------------

SW_FUNC void SWResetTilingCache(
	SpriteWorldPtr	spriteWorldP)
{
	short	row, col;
	
		// Set all elements to -1 (indicating that each tile needs to be drawn)
	for (row = 0; row < spriteWorldP->numOffscreenTileMapRows; row++)
	{
		for (col = 0; col < spriteWorldP->numOffscreenTileMapCols; col++)
		{
			spriteWorldP->offscreenTileMap[(long)row * 
				spriteWorldP->numOffscreenTileMapCols + col] = -1;
		}
	}
}


///--------------------------------------------------------------------------------------
//	SWDrawTile - sets value in tileMap and draws tile if visible in visScrollRect.
///--------------------------------------------------------------------------------------

SW_FUNC void SWDrawTile(
	SpriteWorldPtr	spriteWorldP,
	short			tileRow,
	short			tileCol,
	short			tileID)
{
	short		row, col, offscreenTileRow, offscreenTileCol;
	Rect*		visScrollRectP = &spriteWorldP->visScrollRect;
	Rect*		backRectP = &spriteWorldP->backRect;
	Rect		srcRect, dstRect;
	FramePtr	tileFrameP;
	Boolean		tileClipped;
	
	
		// Set value in tileMap and get tileFrameP
	spriteWorldP->tileMapStructP->tileMap[tileRow][tileCol] = tileID;
	tileFrameP = spriteWorldP->tileFrameArray[spriteWorldP->curTileImage[tileID]];
	
	row = tileRow * spriteWorldP->tileHeight;
	col = tileCol * spriteWorldP->tileWidth;
	
	srcRect = tileFrameP->frameRect;
////////////////////////////////////////////////////////////////////////////////
//
// For Be an nxn tile uses a (0, 0, n - 1, n - 1) rect.  Therefore the bottom
// and right of the rect is one less than the tileheight/tilewidth.
//
////////////////////////////////////////////////////////////////////////////////
//	dstRect.bottom = row + spriteWorldP->tileHeight;
//	dstRect.right = col + spriteWorldP->tileWidth;
	dstRect.bottom = row + spriteWorldP->tileHeight - 1;
	dstRect.right = col + spriteWorldP->tileWidth - 1;
////////////////////////////////////////////////////////////////////////////////
	
	
		// Clip tile with visScrollRect
	tileClipped = false;
	if (row < visScrollRectP->top)
	{
		dstRect.top = visScrollRectP->top;
		srcRect.top += visScrollRectP->top - row;
		tileClipped = true;
	}
	else
		dstRect.top = row;
	
	if (col < visScrollRectP->left)
	{
		dstRect.left = visScrollRectP->left;
		srcRect.left += visScrollRectP->left - col;
		tileClipped = true;
	}
	else
		dstRect.left = col;
	
	
	if (dstRect.bottom > visScrollRectP->bottom)
	{
		srcRect.bottom -= dstRect.bottom - visScrollRectP->bottom;
		dstRect.bottom = visScrollRectP->bottom;
		tileClipped = true;
	}
	
	if (dstRect.right > visScrollRectP->right)
	{
		srcRect.right -= dstRect.right - visScrollRectP->right;
		dstRect.right = visScrollRectP->right;
		tileClipped = true;
	}

		// Draw tile if visible on screen (in visScrollRect)
	if (dstRect.left < dstRect.right && dstRect.top < dstRect.bottom)
	{
			// Save rect as having been changed
		SWAddChangedRect(spriteWorldP, &dstRect);

			// Make the tile's dest rect local to the offscreen area
		dstRect.top -= spriteWorldP->vertScrollRectOffset;
		dstRect.bottom -= spriteWorldP->vertScrollRectOffset;
		dstRect.left -= spriteWorldP->horizScrollRectOffset;
		dstRect.right -= spriteWorldP->horizScrollRectOffset;
		
////////////////////////////////////////////////////////////////////////////////
//
// A BRect with n height or width contains n + 1 pixels	in that direction so
// when we wrap we need to add 1 to the rect bottom or right.
//
////////////////////////////////////////////////////////////////////////////////

			// Wrap tile to top or bottom of offscreen area
		if (dstRect.bottom > backRectP->bottom)
		{
			dstRect.top -= (backRectP->bottom + 1);
			dstRect.bottom -= (backRectP->bottom + 1);
		}
		else if (dstRect.top < backRectP->top)
		{
			dstRect.top += (backRectP->bottom + 1);
			dstRect.bottom += (backRectP->bottom + 1);
		}
		
			// Wrap tile to left or right side of offscreen area
		if (dstRect.right > backRectP->right)
		{
			dstRect.left -= (backRectP->right + 1);
			dstRect.right -= (backRectP->right + 1);
		}
		else if (dstRect.left < backRectP->left)
		{
			dstRect.left += (backRectP->right + 1);
			dstRect.right += (backRectP->right + 1);
		}
////////////////////////////////////////////////////////////////////////////////
		
			// Draw tile in the background frame
		SetGWorld(spriteWorldP->backFrameP->framePort, nil);
		(*spriteWorldP->offscreenDrawProc)(tileFrameP, 
				spriteWorldP->backFrameP, &srcRect, &dstRect);
		
		
			// Copy tile to the work area
		SetGWorld(spriteWorldP->workFrameP->framePort, nil);
		(*spriteWorldP->offscreenDrawProc)(spriteWorldP->backFrameP, 
				spriteWorldP->workFrameP, &dstRect, &dstRect);
		
		
		offscreenTileRow = ((int32)dstRect.top) / spriteWorldP->tileHeight;
		offscreenTileCol = ((int32)dstRect.left) / spriteWorldP->tileWidth;
		
			// Set new tile value in offscreenTileMap (the tiling "Cache")
		if (tileClipped)
		{
			spriteWorldP->offscreenTileMap[(long)offscreenTileRow * 
				spriteWorldP->numOffscreenTileMapCols + offscreenTileCol] = -1;
		}
		else
		{
			spriteWorldP->offscreenTileMap[(long)offscreenTileRow * 
				spriteWorldP->numOffscreenTileMapCols + offscreenTileCol] = 
				spriteWorldP->curTileImage[tileID];
		}
	}
}


///--------------------------------------------------------------------------------------
//	SWSetTileChangeProc
///--------------------------------------------------------------------------------------

SW_FUNC void SWSetTileChangeProc(
	SpriteWorldPtr		spriteWorldP,
	TileChangeProcPtr	tileChangeProc)
{
	spriteWorldP->tileChangeProc = tileChangeProc;
}


///--------------------------------------------------------------------------------------
//	SWChangeTileImage
///--------------------------------------------------------------------------------------

SW_FUNC void SWChangeTileImage(
	SpriteWorldPtr	spriteWorldP,
	short			tileID,
	short			newImage)
{
		// Set the current image
	spriteWorldP->curTileImage[tileID] = newImage;
	
		// Update the tile image on screen
	SWUpdateTileOnScreen(spriteWorldP, tileID);
}


///--------------------------------------------------------------------------------------
//	SWUpdateTileOnScreen - render new tile image in offscreen areas
///--------------------------------------------------------------------------------------

SW_FUNC void SWUpdateTileOnScreen(
	SpriteWorldPtr	spriteWorldP,
	short			tileID)
{
	short		tileRow, tileCol;
	short		startRow, startCol, stopRow, stopCol;
	short		tileWidth = spriteWorldP->tileWidth;
	short		tileHeight = spriteWorldP->tileHeight;
	
	
		// Convert pixel row and col into tile row and col
	startRow = spriteWorldP->visScrollRect.top / tileHeight;
	startCol = spriteWorldP->visScrollRect.left / tileWidth;
////////////////////////////////////////////////////////////////////////////////
//
// This changes compensates for the fact that a BRect has a different number of
// pixels than a (Mac) rect, and that Be coordinates are floats, not ints.
//
////////////////////////////////////////////////////////////////////////////////
//	stopRow = (spriteWorldP->visScrollRect.bottom-1) / tileHeight;
//	stopCol = (spriteWorldP->visScrollRect.right-1) / tileWidth;
	stopRow = ((int32)(spriteWorldP->visScrollRect.bottom)) / tileHeight;
	stopCol = ((int32)(spriteWorldP->visScrollRect.right)) / tileWidth;
////////////////////////////////////////////////////////////////////////////////
	
	for (tileRow = startRow; tileRow <= stopRow; tileRow++)
	{
		for (tileCol = startCol; tileCol <= stopCol; tileCol++)
		{
			if (tileID == spriteWorldP->tileMapStructP->tileMap[tileRow][tileCol])
				SWDrawTile(spriteWorldP, tileRow, tileCol, tileID);
		}
	}
}


///--------------------------------------------------------------------------------------
//	SWResetCurrentTileImages
///--------------------------------------------------------------------------------------

SW_FUNC void SWResetCurrentTileImages(
	SpriteWorldPtr	spriteWorldP)
{
	short	tileIndex;
	
	if (spriteWorldP->tilingIsInitialized)
	{
		for (tileIndex = 0; tileIndex < spriteWorldP->maxNumTiles; tileIndex++)
			spriteWorldP->curTileImage[tileIndex] = tileIndex;
	}
}


///--------------------------------------------------------------------------------------
//	SWDrawTilesInRect - draw tiles in updateRect of backFrame
///--------------------------------------------------------------------------------------

SW_FUNC void SWDrawTilesInRect(
	SpriteWorldPtr	spriteWorldP,
	Rect*			updateRectP,
	Boolean			optimizingOn)
{
	short		row, col, tileRow, tileCol, tileID;
	short		startRow, startCol, stopRow, stopCol;
	short		offscreenTileRow, offscreenTileCol;
	Rect		srcRect, dstRect;
	FramePtr	tileFrameP;
	Boolean		tileClipped;
	Rect*		visScrollRectP = &spriteWorldP->visScrollRect;
	Rect*		backRectP = &spriteWorldP->backRect;
	Rect		updateRect = *updateRectP;
	short		tileWidth = spriteWorldP->tileWidth;
	short		tileHeight = spriteWorldP->tileHeight;
	
	SetGWorld(spriteWorldP->backFrameP->framePort, nil);
	
		// Convert pixel row and col into tile row and col
	startRow = ((int32)updateRect.top) / tileHeight;
	startCol = ((int32)updateRect.left) / tileWidth;

////////////////////////////////////////////////////////////////////////////////
//
// This changes compensates for the fact that a BRect has a different number of
// pixels than a (Mac) rect, and that Be coordinates are floats, not ints.
//
////////////////////////////////////////////////////////////////////////////////
//	stopRow = ((int32)updateRect.bottom - 1) / tileHeight;
//	stopCol = ((int32)updateRect.right - 1) / tileWidth;
	stopRow = ((int32)updateRect.bottom) / tileHeight;
	stopCol = ((int32)updateRect.right) / tileWidth;
		
	for (tileRow = startRow; tileRow <= stopRow; tileRow++)
	{
		row = tileRow * tileHeight;
		
		for (tileCol = startCol; tileCol <= stopCol; tileCol++)
		{
			col = tileCol * tileWidth;
			
			// Add ASSERT stuff here to make sure that:
			// 1) The tileRow and tileCol are within bounds (spriteWorldP->tileMapStructP->numRows)
			// 2) The tileID that is returned is less than the SpriteWorld's maxNumTiles.
			// 3) The tileFrameP is not NULL.
			
			tileID = spriteWorldP->tileMapStructP->tileMap[tileRow][tileCol];
			tileFrameP = spriteWorldP->tileFrameArray[spriteWorldP->curTileImage[tileID]];
			
			srcRect = tileFrameP->frameRect;
////////////////////////////////////////////////////////////////////////////////
//
// For Be an nxn tile uses a (0, 0, n - 1, n - 1) rect.  Therefore the bottom
// and right of the rect is one less than the tileheight/tilewidth.
//
////////////////////////////////////////////////////////////////////////////////
//			dstRect.bottom = row + tileHeight;
//			dstRect.right = col + tileWidth;
			dstRect.bottom = row + tileHeight - 1;
			dstRect.right = col + tileWidth - 1;
////////////////////////////////////////////////////////////////////////////////
			
				// Is tile completely visible on screen?
			if (row >= visScrollRectP->top && col >= visScrollRectP->left &&
				dstRect.bottom <= visScrollRectP->bottom && 
				dstRect.right <= visScrollRectP->right)
			{
				tileClipped = false;
			}
			else
			{
				tileClipped = true;
			}
			
				
				// Clip tile dstRect with updateRect //
			if (row < updateRect.top)
			{
				dstRect.top = updateRect.top;
				srcRect.top += updateRect.top - row;
			}
			else
				dstRect.top = row;
			
			if (col < updateRect.left)
			{
				dstRect.left = updateRect.left;
				srcRect.left += updateRect.left - col;
			}
			else
				dstRect.left = col;
			
			
			if (dstRect.bottom > updateRect.bottom)
			{
				srcRect.bottom -= dstRect.bottom - updateRect.bottom;
				dstRect.bottom = updateRect.bottom;
			}
			
			if (dstRect.right > updateRect.right)
			{
				srcRect.right -= dstRect.right - updateRect.right;
				dstRect.right = updateRect.right;
			}
	
				// Make the tile's dest rect local to the offscreen area
			dstRect.top -= (spriteWorldP->vertScrollRectOffset);
			dstRect.bottom -= (spriteWorldP->vertScrollRectOffset);
			dstRect.left -= (spriteWorldP->horizScrollRectOffset);
			dstRect.right -= (spriteWorldP->horizScrollRectOffset);
			
////////////////////////////////////////////////////////////////////////////////
//
// A BRect with n height or width contains n + 1 pixels	in that direction so
// when we wrap we need to add 1 to the rect bottom or right.
//
////////////////////////////////////////////////////////////////////////////////
				// Wrap tile to top or bottom of offscreen area
			if (dstRect.bottom > backRectP->bottom)
			{
				dstRect.top -= (backRectP->bottom + 1);
				dstRect.bottom -= (backRectP->bottom + 1);
			}
			else if (dstRect.top < backRectP->top)
			{
				dstRect.top += (backRectP->bottom + 1);
				dstRect.bottom += (backRectP->bottom + 1);
			}
			
				// Wrap tile to left or right side of offscreen area
			if (dstRect.right > backRectP->right)
			{
				dstRect.left -= (backRectP->right + 1);
				dstRect.right -= (backRectP->right + 1);
			}
			else if (dstRect.left < backRectP->left)
			{
				dstRect.left += (backRectP->right + 1);
				dstRect.right += (backRectP->right + 1);
			}
////////////////////////////////////////////////////////////////////////////////

			offscreenTileRow = ((int32)dstRect.top) / tileHeight;
			offscreenTileCol = ((int32)dstRect.left) / tileWidth;
			
			
				// Save time by not drawing tile if already the same
			if (spriteWorldP->offscreenTileMap[(long)offscreenTileRow * 
				spriteWorldP->numOffscreenTileMapCols + offscreenTileCol] != 
				spriteWorldP->curTileImage[tileID] || !optimizingOn)
			{	
					// Draw the tile
				(*spriteWorldP->offscreenDrawProc)(tileFrameP, 
						spriteWorldP->backFrameP, &srcRect, &dstRect);
				
					// Set new tile value in offscreenTileMap
				if (tileClipped)
				{
					spriteWorldP->offscreenTileMap[(long)offscreenTileRow * 
						spriteWorldP->numOffscreenTileMapCols + offscreenTileCol] = -1;
				}
				else
				{
					spriteWorldP->offscreenTileMap[(long)offscreenTileRow * 
						spriteWorldP->numOffscreenTileMapCols + offscreenTileCol] = 
						spriteWorldP->curTileImage[tileID];
				}
			}
		}
	}
}


///--------------------------------------------------------------------------------------
//	SWAddChangedRect - used by SWDrawTile
///--------------------------------------------------------------------------------------

SW_FUNC static void SWAddChangedRect(
	SpriteWorldPtr	spriteWorldP, 
	Rect			*changedRectP)
{
	short	index;
	Rect	*changedTileP;
	
	changedTileP = spriteWorldP->changedTiles;
	for ( index = 0; index < spriteWorldP->numTilesChanged; index++, changedTileP++ )
	{
			// check for changedRectP entirely contained by changedTileP
		if ( changedTileP->left <= changedRectP->left && 
			 changedTileP->top <= changedRectP->top && 
			 changedTileP->right >= changedRectP->right && 
			 changedTileP->bottom >= changedRectP->bottom ) 
		{
			return;
		}
	}
	
	changedTileP = spriteWorldP->changedTiles;
	for ( index = 0; index < spriteWorldP->numTilesChanged; index++, changedTileP++ )
	{
			// check for changedRectP horizontally adjacent to changedTileP
		if ( changedTileP->top == changedRectP->top && 
			 changedTileP->bottom == changedRectP->bottom )
		{
			if ( changedRectP->left <= changedTileP->left && 		// changedRectP is to the left of changedTileP
				 changedRectP->right >= changedTileP->left ||		// or
				 changedRectP->left <= changedTileP->right && 		// changedRectP is to the right of changedTileP
				 changedRectP->right >= changedTileP->right )
			{
				changedTileP->left = SW_MIN( changedTileP->left, changedRectP->left );
				changedTileP->right = SW_MAX( changedTileP->right, changedRectP->right );
				return;
			}
		}
		
			// check for changedRectP vertically adjacent to changedTileP
		if ( changedTileP->left == changedRectP->left && 
			 changedTileP->right == changedRectP->right )
		{
			if ( changedRectP->top <= changedTileP->top && 			// changedRectP is above changedTileP
				 changedRectP->bottom >= changedTileP->top ||		// or
				 changedRectP->top <= changedTileP->bottom && 		// changedRectP is below changedTileP
				 changedRectP->bottom >= changedTileP->bottom )
			{
				changedTileP->top = SW_MIN( changedTileP->top, changedRectP->top );
				changedTileP->bottom = SW_MAX( changedTileP->bottom, changedRectP->bottom );
				return;
			}
		}
	}
	
	if (spriteWorldP->numTilesChanged < spriteWorldP->changedTilesArraySize)
	{
		spriteWorldP->changedTiles[spriteWorldP->numTilesChanged++] = *changedRectP;
	}
	else
	{
//		SysBeep(1);		// This shouldn't ever happen, but if it does, we'll beep.
		SWFlagRectAsChanged(spriteWorldP, changedRectP);
	}
}


///--------------------------------------------------------------------------------------
//	SWDrawTilesAboveSprite - draw tiles over sprite using the tiles' masks.
//	Assumes that updateRect fits within the bounds of the tileMap.
///--------------------------------------------------------------------------------------

SW_FUNC void SWDrawTilesAboveSprite(
	SpriteWorldPtr	spriteWorldP,
	Rect*			updateRectP)
{
	Rect*		backRectP = &spriteWorldP->backRect;
	Rect		updateRect = *updateRectP;
	short		tileWidth = spriteWorldP->tileWidth;
	short		tileHeight = spriteWorldP->tileHeight;
	short		row, col, tileRow, tileCol, tileID;
	short		startRow, startCol, stopRow, stopCol;
	Rect		srcRect, dstRect;
	FramePtr	tileFrameP;

	
		// Convert pixel row and col into tile row and col
	startRow = ((int32)updateRect.top) / tileHeight;
	startCol = ((int32)updateRect.left) / tileWidth;
////////////////////////////////////////////////////////////////////////////////
//
// This changes compensates for the fact that a BRect has a different number of
// pixels than a (Mac) rect, and that Be coordinates are floats, not ints.
//
////////////////////////////////////////////////////////////////////////////////
//	stopRow = (updateRect.bottom-1) / tileHeight;
//	stopCol = (updateRect.right-1) / tileWidth;
	stopRow = ((int32)updateRect.bottom) / tileHeight;
	stopCol = ((int32)updateRect.right) / tileWidth;
////////////////////////////////////////////////////////////////////////////////
	
	
	for (tileRow = startRow; tileRow <= stopRow; tileRow++)
	{
		row = tileRow * tileHeight;
		
		for (tileCol = startCol; tileCol <= stopCol; tileCol++)
		{
			col = tileCol * tileWidth;
			
			tileID = spriteWorldP->tileMapStructP->tileMap[tileRow][tileCol];
			tileFrameP = spriteWorldP->tileFrameArray[spriteWorldP->curTileImage[tileID]];
			
				// Skip tiles that have no mask, and therefore aren't above the sprites
//---
//			if (tileFrameP->maskPort == NULL && tileFrameP->maskRgn == NULL &&
			if (tileFrameP->maskPix == NULL && tileFrameP->maskRgn == NULL &&
				tileFrameP->tileMaskIsSolid == false)
			{
				continue;
			}
			
			srcRect = tileFrameP->frameRect;
////////////////////////////////////////////////////////////////////////////////
//
// For Be an nxn tile uses a (0, 0, n - 1, n - 1) rect.  Therefore the bottom
// and right of the rect is one less than the tileheight/tilewidth.
//
////////////////////////////////////////////////////////////////////////////////
//			dstRect.bottom = row + tileHeight;
//			dstRect.right = col + tileWidth;
			dstRect.bottom = row + tileHeight - 1;
			dstRect.right = col + tileWidth - 1;
////////////////////////////////////////////////////////////////////////////////			
				
				// Clip tile dstRect with updateRect //
			if (row < updateRect.top)
			{
				dstRect.top = updateRect.top;
				srcRect.top += updateRect.top - row;
			}
			else
				dstRect.top = row;
			
			if (col < updateRect.left)
			{
				dstRect.left = updateRect.left;
				srcRect.left += updateRect.left - col;
			}
			else
				dstRect.left = col;
			
			
			if (dstRect.bottom > updateRect.bottom)
			{
				srcRect.bottom -= dstRect.bottom - updateRect.bottom;
				dstRect.bottom = updateRect.bottom;
			}
			
			if (dstRect.right > updateRect.right)
			{
				srcRect.right -= dstRect.right - updateRect.right;
				dstRect.right = updateRect.right;
			}
	

			
				// Make the tile's dest rect local to the offscreen area
			dstRect.top -= (spriteWorldP->vertScrollRectOffset );
			dstRect.bottom -= (spriteWorldP->vertScrollRectOffset );
			dstRect.left -= (spriteWorldP->horizScrollRectOffset );
			dstRect.right -= (spriteWorldP->horizScrollRectOffset );
			
////////////////////////////////////////////////////////////////////////////////
//
// A BRect with n height or width contains n + 1 pixels	in that direction so
// when we wrap we need to add 1 to the rect bottom or right.
//
////////////////////////////////////////////////////////////////////////////////
				// Wrap tile to top or bottom of offscreen area
			if (dstRect.bottom > backRectP->bottom)
			{
				dstRect.top -= (backRectP->bottom + 1);
				dstRect.bottom -= (backRectP->bottom + 1);
			}
			else if (dstRect.top < backRectP->top)
			{
				dstRect.top += (backRectP->bottom + 1);
				dstRect.bottom += (backRectP->bottom + 1);
			}
			
				// Wrap tile to left or right side of offscreen area
			if (dstRect.right > backRectP->right)
			{
				dstRect.left -= (backRectP->right + 1);
				dstRect.right -= (backRectP->right + 1);
			}
			else if (dstRect.left < backRectP->left)
			{
				dstRect.left += (backRectP->right + 1);
				dstRect.right += (backRectP->right + 1);
			}
			
			if (tileFrameP->tileMaskIsSolid)
			{
					// Draw the tile without using a mask
				(*spriteWorldP->offscreenDrawProc)(tileFrameP, 
					spriteWorldP->workFrameP, &srcRect, &dstRect);
			}
			else
			{
					// Draw the masked part of the tile
				(*spriteWorldP->tileMaskDrawProc)(tileFrameP, 
						spriteWorldP->workFrameP, &srcRect, &dstRect);
			}
		}
	}
}


///--------------------------------------------------------------------------------------
//   SWReturnTileUnderPixel
///--------------------------------------------------------------------------------------

SW_FUNC short SWReturnTileUnderPixel(
	SpriteWorldPtr	spriteWorldP,
	short pixelCol,
	short pixelRow)
{
	short 	row, col;
	
	SW_ASSERT(spriteWorldP->tileMapStructP != NULL);
	
	row = pixelRow / spriteWorldP->tileHeight;
	col = pixelCol / spriteWorldP->tileWidth;
	
	if (row < 0 || row >= spriteWorldP->tileMapStructP->numRows ||
		col < 0 || col >= spriteWorldP->tileMapStructP->numCols )
	{
		return -1;	// Pixel location is outside TileMap bounds!
	}
	else
	{
		return spriteWorldP->tileMapStructP->tileMap[row][col];
	}
}


///--------------------------------------------------------------------------------------
//   SWCheckSpriteWithTiles
///--------------------------------------------------------------------------------------

SW_FUNC Boolean SWCheckSpriteWithTiles(
	SpriteWorldPtr	spriteWorldP,
	SpritePtr		srcSpriteP, 
	SWTileSearchType searchType,
	Rect			*insetRectP,
	short			firstTileID,
	short			lastTileID,
	Boolean			fixPosition)
{
	short 			row, col, startRow, stopRow, startCol, stopCol;
	TileMapPtr		tileMap = spriteWorldP->tileMapStructP->tileMap;
	Rect			testRect;
	Boolean			foundTile = false;
	short			temp, tileID;
	
	testRect = srcSpriteP->destFrameRect;
	
	if (insetRectP != NULL)
	{
		testRect.left += insetRectP->left;
		testRect.top += insetRectP->top;
		testRect.right -= insetRectP->right;
		testRect.bottom -= insetRectP->bottom;
	}

	if (searchType == kSWTopSide)
	{
		startRow = testRect.top / spriteWorldP->tileHeight;
		stopRow = startRow;
		startCol = testRect.left / spriteWorldP->tileWidth;
		stopCol = (testRect.right-1) / spriteWorldP->tileWidth;
	}
	else if (searchType == kSWBottomSide)
	{
		startRow = (testRect.bottom-1) / spriteWorldP->tileHeight;
		stopRow = startRow;
		startCol = testRect.left / spriteWorldP->tileWidth;
		stopCol = (testRect.right-1) / spriteWorldP->tileWidth;
	}
	else if (searchType == kSWRightSide)
	{
		startRow = testRect.top / spriteWorldP->tileHeight;
		stopRow = (testRect.bottom-1) / spriteWorldP->tileHeight;
		startCol = (testRect.right-1) / spriteWorldP->tileWidth;
		stopCol = startCol;
	}
	else if (searchType == kSWLeftSide)
	{
		startRow = testRect.top / spriteWorldP->tileHeight;
		stopRow = (testRect.bottom-1) / spriteWorldP->tileHeight;
		startCol = testRect.left / spriteWorldP->tileWidth;
		stopCol = startCol;
	}
	else
	{
		startRow = testRect.top / spriteWorldP->tileHeight;
		stopRow = (testRect.bottom-1) / spriteWorldP->tileHeight;
		startCol = testRect.left / spriteWorldP->tileWidth;
		stopCol = (testRect.right-1) / spriteWorldP->tileWidth;
	}
	
		// Make sure things are within bounds (in case Sprite is hanging off edge of TileMap)
	if (startRow < 0)
		startRow = 0;
	if (stopRow >= spriteWorldP->tileMapStructP->numRows)
		stopRow = spriteWorldP->tileMapStructP->numRows-1;
	if (startCol < 0)
		startCol = 0;
	if (stopCol >= spriteWorldP->tileMapStructP->numCols)
		stopCol = spriteWorldP->tileMapStructP->numCols-1;
	
	
		// Look for the tiles
	for (row = startRow; row <= stopRow; row++)
	{
		for (col = startCol; col <= stopCol; col++)
		{
			tileID = tileMap[row][col];
			if (tileID >= firstTileID && tileID <= lastTileID)
			{
				foundTile = true;
				goto exit;
			}
		}
	}
	
exit:

	
	if (foundTile && fixPosition)
	{
		if (searchType == kSWTopSide)
		{
			temp = (startRow+1) * spriteWorldP->tileHeight - testRect.top;
			srcSpriteP->destFrameRect.top += temp;
			srcSpriteP->destFrameRect.bottom += temp;
			srcSpriteP->needsToBeDrawn = true;
		}
		else if (searchType == kSWBottomSide)
		{
			temp = testRect.bottom - startRow * spriteWorldP->tileHeight;
			srcSpriteP->destFrameRect.top -= temp;
			srcSpriteP->destFrameRect.bottom -= temp;
			srcSpriteP->needsToBeDrawn = true;
		}
		else if (searchType == kSWRightSide)
		{
			temp = testRect.right - startCol * spriteWorldP->tileWidth;
			srcSpriteP->destFrameRect.left -= temp;
			srcSpriteP->destFrameRect.right -= temp;
			srcSpriteP->needsToBeDrawn = true;
		}
		else if (searchType == kSWLeftSide)
		{
			temp = (startCol+1) * spriteWorldP->tileWidth - testRect.left;
			srcSpriteP->destFrameRect.left += temp;
			srcSpriteP->destFrameRect.right += temp;
			srcSpriteP->needsToBeDrawn = true;
		}
	}
	
	
	return foundTile;
}


///--------------------------------------------------------------------------------------
//	SWWrapRectToWorkArea - I think this is identical to SWEraseWrappedSprite, and is here
//	simply so Scrolling.c isn't required during compiles.
///--------------------------------------------------------------------------------------

SW_FUNC void SWWrapRectToWorkArea(
	SpriteWorldPtr	spriteWorldP,
	Rect*			destRectP)
{
	Rect		destRect = *destRectP;
	Rect		tempDestRect;
	FramePtr	srcFrameP = spriteWorldP->backFrameP;
	FramePtr	dstFrameP = spriteWorldP->workFrameP;
	
	
	SetGWorld(spriteWorldP->workFrameP->framePort, nil);
	
		// Make destRect local to the offscreen area
	destRect.top -= (spriteWorldP->vertScrollRectOffset );
	destRect.bottom -= (spriteWorldP->vertScrollRectOffset );
	destRect.left -= (spriteWorldP->horizScrollRectOffset );
	destRect.right -= (spriteWorldP->horizScrollRectOffset );
	
	
		// Draw main image
	(*spriteWorldP->offscreenDrawProc)(srcFrameP, dstFrameP, &destRect, &destRect);
	
	
		// Wrap to top //
	if (destRect.bottom > dstFrameP->frameRect.bottom)
	{
		tempDestRect.top = destRect.top - dstFrameP->frameRect.bottom ;
		tempDestRect.bottom = destRect.bottom - dstFrameP->frameRect.bottom ;
		tempDestRect.left = destRect.left;
		tempDestRect.right = destRect.right;
		
		(*spriteWorldP->offscreenDrawProc)(srcFrameP, dstFrameP, 
			&tempDestRect, &tempDestRect);
		
			// Wrap to upper left or right corner //
		if (destRect.right > dstFrameP->frameRect.right)
		{
			tempDestRect.left -= dstFrameP->frameRect.right;
			tempDestRect.right -= dstFrameP->frameRect.right;
			
			(*spriteWorldP->offscreenDrawProc)(srcFrameP, dstFrameP, 
				&tempDestRect, &tempDestRect);
		}
		else if (destRect.left < dstFrameP->frameRect.left)
		{
			tempDestRect.left += dstFrameP->frameRect.right;
			tempDestRect.right += dstFrameP->frameRect.right;
			
			(*spriteWorldP->offscreenDrawProc)(srcFrameP, dstFrameP, 
				&tempDestRect, &tempDestRect);
		}
	}
	
			// Wrap to left or right side //
	if (destRect.right > dstFrameP->frameRect.right)
	{
		tempDestRect.top = destRect.top;
		tempDestRect.bottom = destRect.bottom;
		tempDestRect.left = destRect.left - dstFrameP->frameRect.right;
		tempDestRect.right = destRect.right - dstFrameP->frameRect.right;
		
		(*spriteWorldP->offscreenDrawProc)(srcFrameP, dstFrameP, 
			&tempDestRect, &tempDestRect);
	}
	else if (destRect.left < dstFrameP->frameRect.left)
	{
		tempDestRect.top = destRect.top;
		tempDestRect.bottom = destRect.bottom;
		tempDestRect.left = destRect.left + dstFrameP->frameRect.right;
		tempDestRect.right = destRect.right + dstFrameP->frameRect.right;
		
		(*spriteWorldP->offscreenDrawProc)(srcFrameP, dstFrameP, 
			&tempDestRect, &tempDestRect);
	}
	
	
			// Wrap to bottom //
	if (destRect.top < dstFrameP->frameRect.top)
	{
		tempDestRect.top = destRect.top + dstFrameP->frameRect.bottom;
		tempDestRect.bottom = destRect.bottom + dstFrameP->frameRect.bottom ;
		tempDestRect.left = destRect.left;
		tempDestRect.right = destRect.right;
		
		(*spriteWorldP->offscreenDrawProc)(srcFrameP, dstFrameP, 
			&tempDestRect, &tempDestRect);
		
			// Wrap to lower left or right corner //
		if (destRect.right > dstFrameP->frameRect.right)
		{
			tempDestRect.left -= dstFrameP->frameRect.right;
			tempDestRect.right -= dstFrameP->frameRect.right;
			
			(*spriteWorldP->offscreenDrawProc)(srcFrameP, dstFrameP, 
				&tempDestRect, &tempDestRect);
		}
		else if (destRect.left < dstFrameP->frameRect.left)
		{
			tempDestRect.left += dstFrameP->frameRect.right;
			tempDestRect.right += dstFrameP->frameRect.right;
			
			(*spriteWorldP->offscreenDrawProc)(srcFrameP, dstFrameP, 
				&tempDestRect, &tempDestRect);
		}
	}
}

///--------------------------------------------------------------------------------------
//	SWStdTileMaskDrawProc
///--------------------------------------------------------------------------------------
SW_FUNC void SWStdTileMaskDrawProc(
	FramePtr srcFrameP,
	FramePtr dstFrameP,
	Rect* srcRect,
	Rect* dstRect)
{
	if (!srcFrameP->maskPix)
		return;
			
	if (dstFrameP->framePort->Window()->Lock())
	{
		dstFrameP->framePort->DrawBitmapAsync(srcFrameP->maskPix, *srcRect, *dstRect);	
		dstFrameP->framePort->Sync();	
		dstFrameP->framePort->Window()->Unlock();
	}
}

///--------------------------------------------------------------------------------------
//	SWLoadTileFromFile
///--------------------------------------------------------------------------------------

SW_FUNC OSErr SWLoadTileFromFile(
	SpriteWorldPtr	spriteWorldP, 
	short			tileID,
	const char 		*tilePath,
	const char 		*maskPath, 
	MaskType		maskType)
{
	OSErr 			err;
	BBitmap 		*tempTileBitmap,
					*tempMaskBitmap;
	FramePtr 		newFrameP;
	Rect			frameRect;
	Boolean			maskIsSolid = false;
	
	
	err = noErr;
	newFrameP = NULL;

	if ( !spriteWorldP->tilingIsInitialized )
	{
		err = kTilingNotInitialized;
	}
	else if ( tileID < 0 || tileID >= spriteWorldP->maxNumTiles)
	{
		err = kOutOfRangeErr;
	}
	
	if (maskType == kSolidMask)
	{
		maskIsSolid = true;
		maskType = kNoMask;
	}
	
	if ( err == noErr )
	{	
		/*
			Load the tile and mask (if required) bitmaps.
		*/
		err = SWCreateBitmapFromFile(spriteWorldP, &tempTileBitmap, tilePath, true);
			
		if ( err == noErr && maskType != kNoMask )
			err = SWCreateBitmapFromFile(spriteWorldP, &tempMaskBitmap, maskPath, true);

		/*
			Check if self-masking.
		*/
		if (err == noErr && maskPath != NULL)
		{
			if ( strcmp(tilePath, maskPath) == 0 && maskType != kNoMask)
			{
				err = SWBlackenBitmap( tempMaskBitmap );
			}
		}

		/*
			Create the frame for the tile.
		*/
		if (err == noErr)
		{
			err = SWCreateFrameFromBitmap(spriteWorldP, &newFrameP, tempTileBitmap,
				tempMaskBitmap, maskType);
		}
		
		if ( err == noErr )
		{	
			newFrameP->tileMaskIsSolid = maskIsSolid;
			
				// "Fix" the size if the bitmap is larger than the tiles are
			if (newFrameP->frameRect.right > spriteWorldP->tileWidth)
				newFrameP->frameRect.right = spriteWorldP->tileWidth;
			if (newFrameP->frameRect.bottom > spriteWorldP->tileHeight)
				newFrameP->frameRect.bottom = spriteWorldP->tileHeight;
			
		}
		
		if (err == noErr)
		{
			newFrameP->maskPix = tempMaskBitmap;
		}
		
		if ( err == noErr )
		{
				// Are we replacing an old tile?
			if (spriteWorldP->tileFrameArray[tileID] != NULL)
			{
				SWDisposeTile( spriteWorldP, tileID );
			}
			spriteWorldP->tileFrameArray[tileID] = newFrameP;
		}
	}
	
	if ( err != noErr && newFrameP != NULL )
	{
		SWDisposeFrame(newFrameP);
	}
	
	SWSetStickyIfError(err);
	return err;
}

///--------------------------------------------------------------------------------------
//	SWLoadTilesFromFileXY
///--------------------------------------------------------------------------------------

SW_FUNC OSErr SWLoadTilesFromFileXY(
	SpriteWorldPtr	spriteWorldP, 
	short			startTileID,
	short			endTileID, 
	const char 		*tilePath, 
	const char 		*maskPath,
	short			borderWidth,
	short			borderHeight,
	MaskType		maskType)
{
	OSErr 			err;
	short			tileIndex;
	FramePtr 		newFrameP;
	BBitmap			*tempTileBitmap,
					*tempMaskBitmap,
					*tempTempMaskBitmap;
	Rect			currentTileRect;
	short			horizOffset,
					vertOffset;
	Boolean			allTilesDone,
					maskIsSolid = false;
	
	
	err = noErr;
	newFrameP = NULL;
	tempTileBitmap = NULL;
	tempMaskBitmap = NULL;
	tempTempMaskBitmap = NULL;
	
	if ( !spriteWorldP->tilingIsInitialized )
	{
		err = kTilingNotInitialized;
	}
	else if ( startTileID < 0 || endTileID >= spriteWorldP->maxNumTiles)
	{
		err = kOutOfRangeErr;
	}
	
	if (maskType == kSolidMask)
	{
		maskIsSolid = true;
		maskType = kNoMask;
	}
	
	if ( err == noErr )
	{	
		err = SWCreateBitmapFromFile(spriteWorldP, &tempTileBitmap, tilePath, true);
		
		if ( err == noErr && maskType != kNoMask )
			err = SWCreateBitmapFromFile(spriteWorldP, &tempMaskBitmap, maskPath, true);
		
		if (err == noErr  && maskPath != NULL)
		{
			if ( strcmp(tilePath, maskPath) == 0 && maskType != kNoMask)
			{
				/*
					Switch to call SWBlackenBitmap from GWorld
				*/
				err = SWBlackenBitmap( tempMaskBitmap );
			}
		}
		
		if ( err == noErr )
		{
			err = SWCreateFrameFromBitmapAndRectStart( tempTempMaskBitmap,
				spriteWorldP->tileWidth, spriteWorldP->tileHeight, maskType );
		}
	
		if ( err == noErr )
		{
			currentTileRect.Set(0,  0, spriteWorldP->tileWidth - 1,
				spriteWorldP->tileHeight - 1);
			tileIndex = startTileID;
			horizOffset = spriteWorldP->tileWidth + borderWidth;
			vertOffset = spriteWorldP->tileHeight + borderHeight;
			
			allTilesDone = false;
			while( !allTilesDone && err == noErr )
			{
				err = SWCreateFrameFromBitmapAndRectPartial( &newFrameP, tempTileBitmap, 
						tempMaskBitmap, NULL, &currentTileRect, maskType);
				
				if ( newFrameP != NULL && err == noErr )
				{
					newFrameP->tileMaskIsSolid = maskIsSolid;
					
						// Are we replacing an old tile?
					if (spriteWorldP->tileFrameArray[tileIndex] != NULL)
					{
						SWDisposeTile( spriteWorldP, tileIndex );
					}
					spriteWorldP->tileFrameArray[tileIndex] = newFrameP;
					
					tileIndex++;
					if (tileIndex > endTileID )
					{
						allTilesDone = true;
					}
					else
					{
						if (tileIndex >= spriteWorldP->maxNumTiles)
						{
							err = kOutOfRangeErr;
						}
						currentTileRect.left += horizOffset;
						currentTileRect.right += horizOffset;

						if ( currentTileRect.right > tempTileBitmap->Bounds().right )
						{
							currentTileRect.left = 0;
							currentTileRect.right = spriteWorldP->tileWidth - 1;
							currentTileRect.top += vertOffset;
							currentTileRect.bottom += vertOffset;
							if ( currentTileRect.bottom > tempTileBitmap->Bounds().bottom )
							{
								err = kOutOfRangeErr;
							}
						}
					}
				}
			}
			
			if (err == noErr)
			{
					// make a pixel mask
				if ((maskType & kPixelMask) != 0)
				{
					if (spriteWorldP->pixelDepth <= 8)
					{
//						GetGWorld( &currentGWorld, &currentGDH );
//						(void)LockPixels( GetGWorldPixMap(tempMaskGWorldP) );
//						SetGWorld(tempMaskGWorldP, nil);
//						InvertRect(&tempMaskGWorldP->portRect);
//						SetGWorld( currentGWorld, currentGDH );
//						UnlockPixels( GetGWorldPixMap(tempMaskGWorldP) );
					}
				}
					// If no pixel Mask wanted, dispose
					// of the GWorld we used to make region
				else
				{
//					DisposeGWorld( tempMaskGWorldP );
					tempMaskBitmap = NULL;
				}
			}
		}
		
		SWCreateFrameFromBitmapAndRectFinish( tempTempMaskBitmap );
	}
	
	SWSetStickyIfError(err);
	return err;
}
