///--------------------------------------------------------------------------------------
//	Blit8BitScaledPixie.h
//
//	Description: This is the include file for all the BlitPixie source files.
///--------------------------------------------------------------------------------------


#ifndef __BLITSCALEDPIXIE__
#define __BLITSCALEDPIXIE__

#ifndef __SWCOMMON__
#include "SWCommonHeaders.h"
#endif

#ifndef __SPRITEWORLDUTILS__
#include "SpriteWorldUtils.h"
#endif


#ifndef __BLITPIXIE__
#include "BlitPixie.h"
#endif

#define CLIP_RECT(r, r1, r2) \
	/* clip off the top so we don't write into random memory */ \
	if (r2.top < r.top) { \
		r1.top += r.top - r2.top; \
		r2.top =  r.top; \
	} \
	/* clip off the bottom */ \
	if (r2.bottom > r.bottom) {	 \
		r2.bottom = r.bottom;	 \
	}							 \
	/* clip off the left */ \
	if (r2.left < r.left) { \
		r1.left += r.left - r2.left; \
		r2.left = r.left; \
	} \
	/* clip off the right */ \
	if (r2.right > r.right) { \
		r2.right = r.right; \
	} \
	/* Make sure height is valid */ \
	if (r2.bottom <= r2.top) \
		return; \
	/* Make sure width is valid */ \
	if (r2.right <= r2.left) \
		return;

#define START_MMU_MODE() \
	if (gSWmmuMode != true32b) \
	{ \
		mmuMode = true32b; \
		SwapMMUMode(&mmuMode); \
	}
#define STOP_MMU_MODE() \
	if (gSWmmuMode != true32b) \
	{ \
		SwapMMUMode(&mmuMode); \
	}


#ifdef __cplusplus
extern "C" {
#endif

///--------------------------------------------------------------------------------------
/// Public API
///--------------------------------------------------------------------------------------
SW_FUNC OSErr SWSetSpriteScaledSize(
	SpritePtr srcSpriteP,
	long width,
	long height);
SW_FUNC int SWGetSpriteScaledWidth(
	SpritePtr srcSpriteP);
SW_FUNC int SWGetSpriteScaledHeight(
	SpritePtr srcSpriteP);

///--------------------------------------------------------------------------------------
//	Function prototypes for 8-bit and 16-bit scaled blitters
///--------------------------------------------------------------------------------------

SW_FUNC void BlitPixie8BitScaledRectDrawProc(
	FramePtr srcFrameP,
	FramePtr dstFrameP,
	Rect *srcRect,
	Rect *dstRect);

SW_FUNC void BlitPixie8BitScaledMaskDrawProc(
	FramePtr srcFrameP,
	FramePtr dstFrameP,
	Rect *srcRect,
	Rect *dstRect);

SW_FUNC void BlitPixie8BitScaledPartialMaskDrawProc(
	FramePtr srcFrameP,
	FramePtr dstFrameP,
	Rect *srcRect,
	Rect *dstRect);

void BlitPixie8BitScaled(
	unsigned char *srcPixelP,
	unsigned char *dstPixelP,
	register long srcRowsToCopy,
	register long dstRowsToCopy,
	register long srcNumBytesPerRow,
	register long dstNumBytesPerRow,
	register unsigned long srcOffset,
	register unsigned long dstOffset);

void BlitPixieMask8BitScaled(
	unsigned char *srcPixelP,
	unsigned char *dstPixelP,
	unsigned char *maskPixelP,
	register long srcRowsToCopy,
	register long dstRowsToCopy,
	register long srcNumBytesPerRow,
	register long dstNumBytesPerRow,
	register unsigned long srcOffset,
	register unsigned long dstOffset);

SW_FUNC void BlitPixie16BitScaledRectDrawProc(
	FramePtr srcFrameP,
	FramePtr dstFrameP,
	Rect *srcRect,
	Rect *dstRect);

SW_FUNC void BlitPixie16BitScaledMaskDrawProc(
	FramePtr srcFrameP,
	FramePtr dstFrameP,
	Rect *srcRect,
	Rect *dstRect);

SW_FUNC void BlitPixie16BitScaledPartialMaskDrawProc(
	FramePtr srcFrameP,
	FramePtr dstFrameP,
	Rect *srcRect,
	Rect *dstRect);

void BlitPixie16BitScaled(
	unsigned short *srcPixelP,
	unsigned short *dstPixelP,
	register long srcRowsToCopy,
	register long dstRowsToCopy,
	register long srcNumBytesPerRow,
	register long dstNumBytesPerRow,
	register unsigned long srcOffset,
	register unsigned long dstOffset);

void BlitPixieMask16BitScaled(
	unsigned short *srcPixelP,
	unsigned short *dstPixelP,
	unsigned short *maskPixelP,
	register long srcRowsToCopy,
	register long dstRowsToCopy,
	register long srcNumBytesPerRow,
	register long dstNumBytesPerRow,
	register unsigned long srcOffset,
	register unsigned long dstOffset);

#ifdef __cplusplus
}
#endif

#endif /* __BLITSCALEDPIXIE__ */
