///--------------------------------------------------------------------------------------
//	Circular Scrolling.c
//
//	By:	Vern Jensen
//	
//	Created: 7/20/96
//
//	Description:	Routines that enable you to wrap around while scrolling
///--------------------------------------------------------------------------------------



#ifndef __QUICKDRAW__
#include <QuickDraw.h>
#endif

#ifndef __MEMORY__
#include <Memory.h>
#endif

#ifndef __GESTALT__
#include <Gestalt.h>
#endif

#ifndef __SPRITEWORLD__
#include "SpriteWorld.h"
#endif

#ifndef __SPRITEWORLDUTILS__
#include "SpriteWorldUtils.h"
#endif

#ifndef __BLITPIXIE__
#include "BlitPixie.h"
#endif

#ifndef __SCROLLING__
#include "Scrolling.h"
#endif

#ifndef __TILING__
#include "Tiling.h"
#endif

#ifndef __CIRCULARSCROLLING__
#include "Circular Scrolling.h"
#endif


///--------------------------------------------------------------------------------------
//	SWUpdateCircularSpriteWorld
///--------------------------------------------------------------------------------------

SW_FUNC void SWUpdateCircularSpriteWorld(
	SpriteWorldPtr spriteWorldP)
{
	GWorldPtr	saveGWorld;
	GDHandle	saveGDH;
	
	GetGWorld( &saveGWorld, &saveGDH );

		// Copy the background into the work area
	SetGWorld(spriteWorldP->workFrameP->framePort, nil);
	(*spriteWorldP->offscreenDrawProc)(spriteWorldP->backFrameP,
						spriteWorldP->workFrameP,
						&spriteWorldP->backFrameP->frameRect,
						&spriteWorldP->workFrameP->frameRect);
	
	
	spriteWorldP->frameHasOccurred = true;
	SWAnimateCircularSpriteWorld(spriteWorldP);
	
	SetGWorld( saveGWorld, saveGDH );
}


///--------------------------------------------------------------------------------------
//	SWProcessCircularSpriteWorld
///--------------------------------------------------------------------------------------

SW_FUNC void SWProcessCircularSpriteWorld(
	SpriteWorldPtr spriteWorldP)
{
	register SpriteLayerPtr		curSpriteLayerP;
	register SpritePtr			curSpriteP;
	Rect						backRect = spriteWorldP->backRect;
	
		// Process the sprites
	SWProcessSpriteWorld(spriteWorldP);
	if ( !spriteWorldP->frameHasOccurred )
	{
		return;
	}


		// Wrap any sprites that have moved completely outside the backRect
	curSpriteLayerP = spriteWorldP->headSpriteLayerP;
	while (curSpriteLayerP != NULL)
	{
		curSpriteP = curSpriteLayerP->headSpriteP;

		while (curSpriteP != NULL)
		{			
				// Wrap left or right
			if (curSpriteP->destFrameRect.left >= backRect.right)
			{
				curSpriteP->destFrameRect.left -= backRect.right;
				curSpriteP->destFrameRect.right -= backRect.right;
			}
			else if (curSpriteP->destFrameRect.right <= backRect.left)
			{
				curSpriteP->destFrameRect.left += backRect.right;
				curSpriteP->destFrameRect.right += backRect.right;
			}
			
				// Wrap up or down
			if (curSpriteP->destFrameRect.top >= backRect.bottom)
			{
				curSpriteP->destFrameRect.top -= backRect.bottom;
				curSpriteP->destFrameRect.bottom -= backRect.bottom;
			}
			else if (curSpriteP->destFrameRect.bottom <= backRect.top)
			{
				curSpriteP->destFrameRect.top += backRect.bottom;
				curSpriteP->destFrameRect.bottom += backRect.bottom;
			}
			
			curSpriteP = curSpriteP->nextSpriteP;
		}
		
		curSpriteLayerP = curSpriteLayerP->nextSpriteLayerP;
	}
	
	
		// Call the scrolling world move proc
	if (spriteWorldP->worldMoveProc != NULL)
	{
		(*spriteWorldP->worldMoveProc)(spriteWorldP, spriteWorldP->followSpriteP);
	}
	
	
		// Move visScrollRect
	if (spriteWorldP->horizScrollDelta || spriteWorldP->vertScrollDelta)
	{
		SWOffsetCircularVisScrollRect(spriteWorldP, 
			spriteWorldP->horizScrollDelta, 
			spriteWorldP->vertScrollDelta);
	}
}


///--------------------------------------------------------------------------------------
//	SWAnimateCircularSpriteWorld
///--------------------------------------------------------------------------------------

SW_FUNC void SWAnimateCircularSpriteWorld(
	SpriteWorldPtr spriteWorldP)
{
	register SpriteLayerPtr 	curSpriteLayerP;
	register SpritePtr 			curSpriteP;
	Rect*						backRectP = &spriteWorldP->backRect;
	Rect						rectA, rectB, rectC, rectD;
	Boolean						horizClip, vertClip;
	
	if (!spriteWorldP->frameHasOccurred)
		return;
	
		// Set the port to the work area so we can draw in it
	SetGWorld(spriteWorldP->workFrameP->framePort, nil);

	

	//-----------------erase the sprites--------------------
	
	rectA = spriteWorldP->oldVisScrollRect;
	
		// Clip and wrap left or right side
	if (rectA.left < backRectP->left)
	{
			// wrap left side to the right
		rectB.top = rectA.top;
		rectB.bottom = rectA.bottom;
		rectB.left = rectA.left + backRectP->right;
		rectB.right = backRectP->right;
		
		rectA.left = backRectP->left;
		horizClip = true;
	}
	else if (rectA.right > backRectP->right)
	{
			// wrap right side to the left
		rectB.top = rectA.top;
		rectB.bottom = rectA.bottom;
		rectB.left = backRectP->left;
		rectB.right = rectA.right - backRectP->right;
		
		rectA.right = backRectP->right;
		horizClip = true;
	}
	else
		horizClip = false;
	
	
		// Clip and wrap top or bottom
	if (rectA.top < backRectP->top)
	{
			// wrap top side to the bottom
		rectC.left = rectA.left;
		rectC.right = rectA.right;
		rectC.top = rectA.top + backRectP->bottom;
		rectC.bottom = backRectP->bottom;
		
		rectA.top = backRectP->top;
		vertClip = true;
	}
	else if (rectA.bottom > backRectP->bottom)
	{
			// wrap bottom side to the top
		rectC.left = rectA.left;
		rectC.right = rectA.right;
		rectC.top = backRectP->top;
		rectC.bottom = rectA.bottom - backRectP->bottom;
		
		rectA.bottom = backRectP->bottom;
		vertClip = true;
	}
	else
		vertClip = false;
	
	
	SWEraseSpritesInRect(spriteWorldP, &rectA);
	
	if (horizClip)
		SWEraseSpritesInRect(spriteWorldP, &rectB);
	
	if (vertClip)
		SWEraseSpritesInRect(spriteWorldP, &rectC);
	
		// Wrap the corner piece
	if (vertClip && horizClip)
	{
		rectD.left = rectB.left;
		rectD.right = rectB.right;
		rectD.top = rectC.top;
		rectD.bottom = rectC.bottom;
		SWEraseSpritesInRect(spriteWorldP, &rectD);
	}

	
	
	
	//-----------------draw the sprites-------------------
	
	rectA = spriteWorldP->visScrollRect;
	
		// Clip and wrap left or right side
	if (rectA.left < backRectP->left)
	{
			// wrap left side to the right
		rectB.top = rectA.top;
		rectB.bottom = rectA.bottom;
		rectB.left = rectA.left + backRectP->right;
		rectB.right = backRectP->right;
		
		rectA.left = backRectP->left;
		horizClip = true;
	}
	else if (rectA.right > backRectP->right)
	{
			// wrap right side to the left
		rectB.top = rectA.top;
		rectB.bottom = rectA.bottom;
		rectB.left = backRectP->left;
		rectB.right = rectA.right - backRectP->right;
		
		rectA.right = backRectP->right;
		horizClip = true;
	}
	else
		horizClip = false;
	
	
		// Clip and wrap top or bottom
	if (rectA.top < backRectP->top)
	{
			// wrap top side to the bottom
		rectC.left = rectA.left;
		rectC.right = rectA.right;
		rectC.top = rectA.top + backRectP->bottom;
		rectC.bottom = backRectP->bottom;
		
		rectA.top = backRectP->top;
		vertClip = true;
	}
	else if (rectA.bottom > backRectP->bottom)
	{
			// wrap bottom side to the top
		rectC.left = rectA.left;
		rectC.right = rectA.right;
		rectC.top = backRectP->top;
		rectC.bottom = rectA.bottom - backRectP->bottom;
		
		rectA.bottom = backRectP->bottom;
		vertClip = true;
	}
	else
		vertClip = false;
	
	
	SWDrawSpritesInRect(spriteWorldP, &rectA);
	
	if (horizClip)
		SWDrawSpritesInRect(spriteWorldP, &rectB);
	
	if (vertClip)
		SWDrawSpritesInRect(spriteWorldP, &rectC);
	
		// Wrap the corner piece
	if (vertClip && horizClip)
	{
		rectD.left = rectB.left;
		rectD.right = rectB.right;
		rectD.top = rectC.top;
		rectD.bottom = rectC.bottom;
		SWDrawSpritesInRect(spriteWorldP, &rectD);
	}
	
	
		// Clean up //
		
	curSpriteLayerP = spriteWorldP->headSpriteLayerP;

		// iterate through the layers in this world
	while (curSpriteLayerP != NULL)
	{
		curSpriteP = curSpriteLayerP->headSpriteP;

			// iterate through the sprites in this layer
		while (curSpriteP != NULL)
		{
				// Set last rect to current rect
			curSpriteP->oldFrameRect = curSpriteP->destFrameRect;
			
			curSpriteP->needsToBeDrawn = false;
			curSpriteP->needsToBeErased = false;

			curSpriteP = curSpriteP->nextSpriteP;
		}
		
		curSpriteLayerP = curSpriteLayerP->nextSpriteLayerP;
	}
	
	spriteWorldP->oldVisScrollRect = spriteWorldP->visScrollRect;
	
	
	
	//-----------------update the screen--------------------
	
		// Set the port to the window
	SetGWorld(spriteWorldP->windowFrameP->framePort, nil);
	
	if (spriteWorldP->usingVBL)
	{  
		spriteWorldP->vblTaskRec.hasVBLFired = false;
		while ( !spriteWorldP->vblTaskRec.hasVBLFired )
		{}
	}
	
		// Copy offscreen area to screen while wrapping
	SWWrapWorldToScreen(spriteWorldP);
}


///--------------------------------------------------------------------------------------
//  SWDrawSpritesInRect
///--------------------------------------------------------------------------------------

void SWDrawSpritesInRect(
	SpriteWorldPtr	spriteWorldP,
	Rect*			updateRect)
{
	register SpriteLayerPtr 	curSpriteLayerP;
	register SpritePtr 			curSpriteP;
	Rect						sourceRect, destRect;
	
	
	curSpriteLayerP = spriteWorldP->headSpriteLayerP;

		// iterate through the layers in this world
	while (curSpriteLayerP != NULL)
	{
		curSpriteP = curSpriteLayerP->headSpriteP;

			// iterate through the sprites in this layer
		while (curSpriteP != NULL)
		{
			if (curSpriteP->isVisible)
			{
				destRect = curSpriteP->destFrameRect;
				sourceRect = curSpriteP->curFrameP->frameRect;
				
					// Clip the sprite only if it is not being wrapped.
					// (that is, if it is entirely inside the backRect)
				if (destRect.top >= spriteWorldP->backRect.top &&
					destRect.bottom <= spriteWorldP->backRect.bottom &&
					destRect.left >= spriteWorldP->backRect.left &&
					destRect.right <= spriteWorldP->backRect.right)
				{
					if (destRect.top < updateRect->top)
					{
						sourceRect.top += updateRect->top - curSpriteP->destFrameRect.top;
						destRect.top = updateRect->top;
					}
					
					if (destRect.bottom > updateRect->bottom)
					{
						sourceRect.bottom += updateRect->bottom - curSpriteP->destFrameRect.bottom;
						destRect.bottom = updateRect->bottom;
					}
					
					if (destRect.left < updateRect->left)
					{
						sourceRect.left += updateRect->left - curSpriteP->destFrameRect.left;
						destRect.left = updateRect->left;
					}
					
					if (destRect.right > updateRect->right)
					{
						sourceRect.right += updateRect->right - curSpriteP->destFrameRect.right;
						destRect.right = updateRect->right;
					}
				}
				
					// Is the sprite visible in this rect?
				if (destRect.right > destRect.left && destRect.bottom > destRect.top)
				{
					SWDrawWrappedSprite(curSpriteP, spriteWorldP->workFrameP,
								&sourceRect, &destRect);
					
						// Draw tiles above sprite
					if (curSpriteP->isUnderTiles && spriteWorldP->tilingIsOn)
						SWDrawTilesAboveSprite(spriteWorldP, &destRect);
				}
			}
			
			curSpriteP = curSpriteP->nextSpriteP;
		}
		
		curSpriteLayerP = curSpriteLayerP->nextSpriteLayerP;
	}
}


///--------------------------------------------------------------------------------------
//  SWEraseSpritesInRect
///--------------------------------------------------------------------------------------

void SWEraseSpritesInRect(
	SpriteWorldPtr	spriteWorldP,
	Rect*			updateRect)
{
	register SpriteLayerPtr 	curSpriteLayerP;
	register SpritePtr 			curSpriteP;
	Rect						destRect;
	short						temp;
	
	
	curSpriteLayerP = spriteWorldP->headSpriteLayerP;

		// iterate through the layers in this world
	while (curSpriteLayerP != NULL)
	{
		curSpriteP = curSpriteLayerP->headSpriteP;

			// iterate through the sprites in this layer
		while (curSpriteP != NULL)
		{
			if (curSpriteP->needsToBeDrawn && curSpriteP->isVisible ||
				curSpriteP->needsToBeErased && !curSpriteP->isVisible)
			{
				destRect = curSpriteP->oldFrameRect;
				
					// Clip the sprite only if it is not being wrapped.
					// (that is, if it is entirely inside the backRect)
				if (destRect.top >= spriteWorldP->backRect.top &&
					destRect.bottom <= spriteWorldP->backRect.bottom &&
					destRect.left >= spriteWorldP->backRect.left &&
					destRect.right <= spriteWorldP->backRect.right)
				{
					if (destRect.top < updateRect->top)
						destRect.top = updateRect->top;
					
					if (destRect.bottom > updateRect->bottom)
						destRect.bottom = updateRect->bottom;
					
					if (destRect.left < updateRect->left)
						destRect.left = updateRect->left;
					
					if (destRect.right > updateRect->right)
						destRect.right = updateRect->right;
				}
				
				
					// Is the sprite visible in this rect?
				if (destRect.right > destRect.left && destRect.bottom > destRect.top)
				{
						// align the left edge to long word boundary
					destRect.left &= (spriteWorldP->workFrameP->leftAlignFactor);
					
						// align the right edge to long word boundary
					temp = destRect.right & spriteWorldP->workFrameP->rightAlignFactor;
						
					if (temp != 0)
						destRect.right += (spriteWorldP->workFrameP->rightAlignFactor + 1) - temp;
					
					SWEraseWrappedSprite(spriteWorldP, &destRect);
				}
			}
			
			curSpriteP = curSpriteP->nextSpriteP;
		}
		
		curSpriteLayerP = curSpriteLayerP->nextSpriteLayerP;
	}
}


///--------------------------------------------------------------------------------------
//  SWMoveCircularVisScrollRect
///--------------------------------------------------------------------------------------

SW_FUNC void SWMoveCircularVisScrollRect(
	SpriteWorldPtr	spriteWorldP,
	short			horizPos,
	short			vertPos)
{
	short	width, height;
	
	height = spriteWorldP->visScrollRect.bottom - spriteWorldP->visScrollRect.top;
	width = spriteWorldP->visScrollRect.right - spriteWorldP->visScrollRect.left;

	spriteWorldP->visScrollRect.top = vertPos;
	spriteWorldP->visScrollRect.bottom = vertPos + height;
	spriteWorldP->visScrollRect.left = horizPos;
	spriteWorldP->visScrollRect.right = horizPos + width;
	
		// Wrap visScrollRect
	if (spriteWorldP->visScrollRect.top >= spriteWorldP->backRect.bottom)
	{
		spriteWorldP->visScrollRect.top -= spriteWorldP->backRect.bottom;
		spriteWorldP->visScrollRect.bottom -= spriteWorldP->backRect.bottom;
	}
	else if (spriteWorldP->visScrollRect.bottom <= spriteWorldP->backRect.top)
	{
		spriteWorldP->visScrollRect.top += spriteWorldP->backRect.bottom;
		spriteWorldP->visScrollRect.bottom += spriteWorldP->backRect.bottom;
	}
	
	if (spriteWorldP->visScrollRect.left >= spriteWorldP->backRect.right)
	{
		spriteWorldP->visScrollRect.left -= spriteWorldP->backRect.right;
		spriteWorldP->visScrollRect.right -= spriteWorldP->backRect.right;
	}
	else if (spriteWorldP->visScrollRect.right <= spriteWorldP->backRect.left)
	{
		spriteWorldP->visScrollRect.left += spriteWorldP->backRect.right;
		spriteWorldP->visScrollRect.right += spriteWorldP->backRect.right;
	}

	spriteWorldP->offscreenScrollRect = spriteWorldP->visScrollRect;
}


///--------------------------------------------------------------------------------------
//  SWOffsetCircularVisScrollRect
///--------------------------------------------------------------------------------------

SW_FUNC void SWOffsetCircularVisScrollRect(
	SpriteWorldPtr	spriteWorldP,
	short			horizOffset,
	short			vertOffset)
{
	spriteWorldP->visScrollRect.top += vertOffset;
	spriteWorldP->visScrollRect.bottom += vertOffset;
	spriteWorldP->visScrollRect.left += horizOffset;
	spriteWorldP->visScrollRect.right += horizOffset;

		// Wrap visScrollRect
	if (spriteWorldP->visScrollRect.top >= spriteWorldP->backRect.bottom)
	{
		spriteWorldP->visScrollRect.top -= spriteWorldP->backRect.bottom;
		spriteWorldP->visScrollRect.bottom -= spriteWorldP->backRect.bottom;
	}
	else if (spriteWorldP->visScrollRect.bottom <= spriteWorldP->backRect.top)
	{
		spriteWorldP->visScrollRect.top += spriteWorldP->backRect.bottom;
		spriteWorldP->visScrollRect.bottom += spriteWorldP->backRect.bottom;
	}
	
	if (spriteWorldP->visScrollRect.left >= spriteWorldP->backRect.right)
	{
		spriteWorldP->visScrollRect.left -= spriteWorldP->backRect.right;
		spriteWorldP->visScrollRect.right -= spriteWorldP->backRect.right;
	}
	else if (spriteWorldP->visScrollRect.right <= spriteWorldP->backRect.left)
	{
		spriteWorldP->visScrollRect.left += spriteWorldP->backRect.right;
		spriteWorldP->visScrollRect.right += spriteWorldP->backRect.right;
	}

	spriteWorldP->offscreenScrollRect = spriteWorldP->visScrollRect;
}


///--------------------------------------------------------------------------------------
//	SWCollideCircularSpriteLayer - this routine is about as optimized as you can get!
///--------------------------------------------------------------------------------------

SW_FUNC void SWCollideCircularSpriteLayer(
	SpriteWorldPtr spriteWorldP,
	SpriteLayerPtr srcSpriteLayerP,
	SpriteLayerPtr dstSpriteLayerP)
{
	SpritePtr	srcSpriteP, nextSrcSpriteP;
	SpritePtr	dstSpriteP, nextDstSpriteP;
	Rect		sectRect, srcRectA, srcRectB, srcRectC, srcRectD, dstRect;
	Boolean		rectBWrapped, rectCWrapped, rectDWrapped;
	
		// Don't check for collisions unless the frame has been processed!
	if (!spriteWorldP->frameHasOccurred)
		return;

	srcSpriteP = srcSpriteLayerP->headSpriteP;

	while (srcSpriteP != NULL)
	{
		dstSpriteP = dstSpriteLayerP->headSpriteP;
		nextSrcSpriteP = srcSpriteP->nextSpriteP;
		srcRectA = srcSpriteP->destFrameRect;
		
			// Keeps track of which rects were given values
		rectBWrapped = rectCWrapped = rectDWrapped = 0;

			// Wrap to top //
		if (srcRectA.bottom > spriteWorldP->backRect.bottom)
		{
			srcRectB.top = srcRectA.top - spriteWorldP->backRect.bottom;
			srcRectB.bottom = srcRectA.bottom - spriteWorldP->backRect.bottom;
			srcRectB.left = srcRectA.left;
			srcRectB.right = srcRectA.right;
			rectBWrapped = true;
			
				// Wrap to upper left or right corner //
			if (srcRectA.right > spriteWorldP->backRect.right)
			{
				srcRectD = srcRectB;
				srcRectD.left -= spriteWorldP->backRect.right;
				srcRectD.right -= spriteWorldP->backRect.right;
				rectDWrapped = true;
			}
			else if (srcRectA.left < spriteWorldP->backRect.left)
			{
				srcRectD = srcRectB;
				srcRectD.left += spriteWorldP->backRect.right;
				srcRectD.right += spriteWorldP->backRect.right;
				rectDWrapped = true;
			}
		}		// Wrap to bottom //	
		else if (srcRectA.top < spriteWorldP->backRect.top)
		{
			srcRectB.top = srcRectA.top + spriteWorldP->backRect.bottom;
			srcRectB.bottom = srcRectA.bottom + spriteWorldP->backRect.bottom;
			srcRectB.left = srcRectA.left;
			srcRectB.right = srcRectA.right;
			rectBWrapped = true;
			
				// Wrap to lower left or right corner //
			if (srcRectA.right > spriteWorldP->backRect.right)
			{
				srcRectD = srcRectB;
				srcRectD.left -= spriteWorldP->backRect.right;
				srcRectD.right -= spriteWorldP->backRect.right;
				rectDWrapped = true;
			}
			else if (srcRectA.left < spriteWorldP->backRect.left)
			{
				srcRectD = srcRectB;
				srcRectD.left += spriteWorldP->backRect.right;
				srcRectD.right += spriteWorldP->backRect.right;
				rectDWrapped = true;
			}
		}
		
				// Wrap to left or right side //
		if (srcRectA.right > spriteWorldP->backRect.right)
		{
			srcRectC.top = srcRectA.top;
			srcRectC.bottom = srcRectA.bottom;
			srcRectC.left = srcRectA.left - spriteWorldP->backRect.right;
			srcRectC.right = srcRectA.right - spriteWorldP->backRect.right;
			rectCWrapped = true;
		}
		else if (srcRectA.left < spriteWorldP->backRect.left)
		{
			srcRectC.top = srcRectA.top;
			srcRectC.bottom = srcRectA.bottom;
			srcRectC.left = srcRectA.left + spriteWorldP->backRect.right;
			srcRectC.right = srcRectA.right + spriteWorldP->backRect.right;
			rectCWrapped = true;
		}
	
		
			// Check for overlapping rects (dstRect with each srcRect)
		while (dstSpriteP != NULL)
		{
			nextDstSpriteP = dstSpriteP->nextSpriteP;
			dstRect = dstSpriteP->destFrameRect;

			if (srcSpriteP != dstSpriteP)
			{
					// are the sprites main rectangles overlapping?
				if ((srcRectA.top < dstRect.bottom) && (srcRectA.bottom > dstRect.top) &&
					(srcRectA.left < dstRect.right) && (srcRectA.right > dstRect.left))
				{
						// call the source sprites collision routine
					if (srcSpriteP->spriteCollideProc != NULL)
					{
						sectRect.left = SW_MAX(srcRectA.left, dstRect.left);
						sectRect.top = SW_MAX(srcRectA.top, dstRect.top);
						sectRect.right = SW_MIN(srcRectA.right, dstRect.right);
						sectRect.bottom = SW_MIN(srcRectA.bottom, dstRect.bottom);

						(*srcSpriteP->spriteCollideProc)(srcSpriteP, dstSpriteP, &sectRect);
					}
				}	// If the source sprite is wrapped, check wrapped part B
				else if ( rectBWrapped &&
					(srcRectB.top < dstRect.bottom) && (srcRectB.bottom > dstRect.top) &&
					(srcRectB.left < dstRect.right) && (srcRectB.right > dstRect.left))
				{
						// call the source sprites collision routine
					if (srcSpriteP->spriteCollideProc != NULL)
					{
						sectRect.left = SW_MAX(srcRectB.left, dstRect.left);
						sectRect.top = SW_MAX(srcRectB.top, dstRect.top);
						sectRect.right = SW_MIN(srcRectB.right, dstRect.right);
						sectRect.bottom = SW_MIN(srcRectB.bottom, dstRect.bottom);

						(*srcSpriteP->spriteCollideProc)(srcSpriteP, dstSpriteP, &sectRect);
					}
				}	// If the source sprite is wrapped, check wrapped part C
				else if ( rectCWrapped &&
					(srcRectC.top < dstRect.bottom) && (srcRectC.bottom > dstRect.top) &&
					(srcRectC.left < dstRect.right) && (srcRectC.right > dstRect.left))
				{
						// call the source sprites collision routine
					if (srcSpriteP->spriteCollideProc != NULL)
					{
						sectRect.left = SW_MAX(srcRectC.left, dstRect.left);
						sectRect.top = SW_MAX(srcRectC.top, dstRect.top);
						sectRect.right = SW_MIN(srcRectC.right, dstRect.right);
						sectRect.bottom = SW_MIN(srcRectC.bottom, dstRect.bottom);

						(*srcSpriteP->spriteCollideProc)(srcSpriteP, dstSpriteP, &sectRect);
					}
				}	// If the source sprite is wrapped, check wrapped part D
				else if ( rectDWrapped &&
					(srcRectD.top < dstRect.bottom) && (srcRectD.bottom > dstRect.top) &&
					(srcRectD.left < dstRect.right) && (srcRectD.right > dstRect.left))
				{
						// call the source sprites collision routine
					if (srcSpriteP->spriteCollideProc != NULL)
					{
						sectRect.left = SW_MAX(srcRectD.left, dstRect.left);
						sectRect.top = SW_MAX(srcRectD.top, dstRect.top);
						sectRect.right = SW_MIN(srcRectD.right, dstRect.right);
						sectRect.bottom = SW_MIN(srcRectD.bottom, dstRect.bottom);

						(*srcSpriteP->spriteCollideProc)(srcSpriteP, dstSpriteP, &sectRect);
					}
				}
			}
			
			dstSpriteP = nextDstSpriteP;
		}
		
		srcSpriteP = nextSrcSpriteP;
	}
}


///--------------------------------------------------------------------------------------
//	SWDrawTilesInCircularBackground
///--------------------------------------------------------------------------------------

SW_FUNC OSErr SWDrawTilesInCircularBackground(
	SpriteWorldPtr	spriteWorldP)
{
	GWorldPtr		holdGWorld;
	GDHandle		holdGDH;
	OSErr			err = noErr;
	
	GetGWorld( &holdGWorld, &holdGDH );
	
	if (spriteWorldP->tileMapStructP == NULL)
	{
		err = kNullTileMapErr;
	}
	else if ( !spriteWorldP->tilingIsInitialized )
	{
		err = kTilingNotInitialized;
	}
	else
	{
		SWDrawTilesInRect(spriteWorldP, &spriteWorldP->backRect, true);
	}
	
	SetGWorld( holdGWorld, holdGDH );
	
	SWSetStickyIfError(err);
	return err;
}


///--------------------------------------------------------------------------------------
//	SWChangeCircularTileImage
///--------------------------------------------------------------------------------------

SW_FUNC void SWChangeCircularTileImage(
	SpriteWorldPtr	spriteWorldP,
	short			tileID,
	short			newImage)
{
		// Set the current image
	spriteWorldP->curTileImage[tileID] = newImage;
	
		// Update the tile image on screen
	SWUpdateCircularTileOnScreen(spriteWorldP, tileID);
}


///--------------------------------------------------------------------------------------
//	SWUpdateCircularTileOnScreen - render new tile image in offscreen areas
///--------------------------------------------------------------------------------------

SW_FUNC void SWUpdateCircularTileOnScreen(
	SpriteWorldPtr	spriteWorldP,
	short			tileID)
{
	short		tileRow, tileCol;
	short		startRow, startCol, stopRow, stopCol;
	short		tileWidth = spriteWorldP->tileWidth;
	short		tileHeight = spriteWorldP->tileHeight;
	
	
		// Convert pixel row and col into tile row and col
	startRow = spriteWorldP->backRect.top / tileHeight;
	startCol = spriteWorldP->backRect.left / tileWidth;
	stopRow = (spriteWorldP->backRect.bottom-1) / tileHeight;
	stopCol = (spriteWorldP->backRect.right-1) / tileWidth;
	
	
	for (tileRow = startRow; tileRow <= stopRow; tileRow++)
	{
		for (tileCol = startCol; tileCol <= stopCol; tileCol++)
		{
			if (tileID == spriteWorldP->tileMapStructP->tileMap[tileRow][tileCol])
				SWDrawCircularTile(spriteWorldP, tileRow, tileCol, tileID);
		}
	}
}


///--------------------------------------------------------------------------------------
//	SWDrawCircularTile
///--------------------------------------------------------------------------------------

SW_FUNC void SWDrawCircularTile(
	SpriteWorldPtr	spriteWorldP,
	short			tileRow,
	short			tileCol,
	short			tileID)
{
	short		row, col, offscreenTileRow, offscreenTileCol;
	Rect*		visScrollRectP = &spriteWorldP->visScrollRect;
	Rect*		backRectP = &spriteWorldP->backRect;
	Rect		srcRect, dstRect;
	FramePtr	tileFrameP;
	Boolean		tileClipped;
	
	
		// Set value in tileMap and get tileFrameP
	spriteWorldP->tileMapStructP->tileMap[tileRow][tileCol] = tileID;
	tileFrameP = spriteWorldP->tileFrameArray[spriteWorldP->curTileImage[tileID]];
	
	row = tileRow * spriteWorldP->tileHeight;
	col = tileCol * spriteWorldP->tileWidth;
	
	srcRect = tileFrameP->frameRect;
	dstRect.bottom = row + spriteWorldP->tileHeight;
	dstRect.right = col + spriteWorldP->tileWidth;
	
	
		// Clip tile with backRect
	tileClipped = false;
	if (row < backRectP->top)
	{
		dstRect.top = backRectP->top;
		srcRect.top += backRectP->top - row;
		tileClipped = true;
	}
	else
		dstRect.top = row;
	
	if (col < backRectP->left)
	{
		dstRect.left = backRectP->left;
		srcRect.left += backRectP->left - col;
		tileClipped = true;
	}
	else
		dstRect.left = col;
	
	
	if (dstRect.bottom > backRectP->bottom)
	{
		srcRect.bottom -= dstRect.bottom - backRectP->bottom;
		dstRect.bottom = backRectP->bottom;
		tileClipped = true;
	}
	
	if (dstRect.right > backRectP->right)
	{
		srcRect.right -= dstRect.right - backRectP->right;
		dstRect.right = backRectP->right;
		tileClipped = true;
	}

		// Draw tile if visible on screen (in visScrollRect)
	if (dstRect.left < dstRect.right && dstRect.top < dstRect.bottom)
	{
			// Draw tile in the background frame
		SetGWorld(spriteWorldP->backFrameP->framePort, nil);
		(*spriteWorldP->offscreenDrawProc)(tileFrameP, 
				spriteWorldP->backFrameP, &srcRect, &dstRect);
		
		
			// Copy tile to the work area
		SetGWorld(spriteWorldP->workFrameP->framePort, nil);
		(*spriteWorldP->offscreenDrawProc)(spriteWorldP->backFrameP, 
				spriteWorldP->workFrameP, &dstRect, &dstRect);
		
		
		offscreenTileRow = dstRect.top / spriteWorldP->tileHeight;
		offscreenTileCol = dstRect.left / spriteWorldP->tileWidth;
		
			// Set new tile value in offscreenTileMap (the tiling "Cache")
		if (tileClipped)
		{
			spriteWorldP->offscreenTileMap[(long)offscreenTileRow * 
				spriteWorldP->numOffscreenTileMapCols + offscreenTileCol] = -1;
		}
		else
		{
			spriteWorldP->offscreenTileMap[(long)offscreenTileRow * 
				spriteWorldP->numOffscreenTileMapCols + offscreenTileCol] = 
				spriteWorldP->curTileImage[tileID];
		}
	}
}
